#
# CNIDR Isite
#
# Make sure these settings are correct for you platform.  In particular,
# check the CFLAGS value.
#
# This makefile results in:
#
#	libsapi.a	- CNIDR search api
#	libz3950.a	- Z39.50 version 2/3 library
#	libIsearch.a	- CNIDR Isearch library
#	izclient	- Simple, interactive Z39.50 client
#	zserver		- Z39.50 server
#	zcon		- Z39.50 stateful gateway utility
#	zgate		- Z39.50 stateful gateway utility
#	Iindex		- Isearch Text indexer
#	Isearch		- Search tool for Iindexed databases
#	Iutil		- Tool for manipulating Iindexed databases
#

DIST=Isite
VER=2.01

RM=del

#
# Binary directory
#
# Where should I store binaries?
#
BIN_DIR=exe

#
# STOP!!
#
# For the default Isite operation, that should be all you need to change!
# Save this file and type 'make'
#

#
# Search Engines
#
# Isite comes with the following engines precompiled into the Search API:
#
#	SAPI_SCRIPT	- simple shell script search engine
#	SAPI_ISEARCH	- CNIDR Isearch engine.
#
# To link in other databases, refer to the Search API (sapi) documentation.
#

#
# Library directory
#
# Where should I store libraries (*.a)?
#
LIB_DIR=$(BIN_DIR)

#
# Isearch
#
# Where is Isearch code?
#
ISEARCH_DIR=Isearch
ISEARCH_CGI_DIR=$(ISEARCH_DIR)\Isearch-cgi

#
# Doctype-specific libs (e.g., -lgdbm)
#
DOCLIB=

#
# libsapi
#
# Where is libsapi code?
#
SAPI_DIR=sapi2

#
# libz3950
#
# Where is libz3950 code?
#
ZDIST_DIR=zdist2

#
# That should be all you need to change
#
#

all: isearch libsapi zdist done

isearch::
	@echo *** ISEARCH ***
	cd $(ISEARCH_DIR)
	$(MAKE) -$(MAKEFLAGS) -f Isearch.mak \
		"BIN_DIR=..\$(BIN_DIR)" "LIB_DIR=..\$(LIB_DIR)" \
		"DOCLIB=$(DOCLIB)"
	cd ..

libsapi::
	@echo *** LIBSAPI ***
	cd $(SAPI_DIR)
	$(MAKE) -$(MAKEFLAGS) -f sapi.mak \
		"LIB_DIR=..\$(LIB_DIR)" \
		"DBASE_OBJ=$(DBASE_OBJ)" \
		"ZDIST_DIR=../$(ZDIST_DIR)" \
		"ISEARCH_INC=-I../$(ISEARCH_DIR)/src -I../$(ISEARCH_DIR)/doctype"
	cd ..

zdist::
	@echo *** ZDIST ***
	cd $(ZDIST_DIR)
	$(MAKE) -$(MAKEFLAGS) -f zdist.mak \
		"BIN_DIR=..\$(BIN_DIR)" \
		"LIB_DIR=..\$(LIB_DIR)" \
		"ISEARCH_DIR=..\$(ISEARCH_DIR)" \
		"ISEARCH_LIB_DIR=..\$(LIB_DIR)" \
		"SAPI_DIR=..\$(SAPI_DIR)" \
		"DOCLIB=$(DOCLIB)"
	cd ..

done:
	@echo. 
	@echo Welcome to CNIDR $(DIST)!
	@echo.
	@echo All executables have been placed in the $(BIN_DIR) directory.
	@echo.
	@echo Please read all docs (doc/*) for usage instructions.
	@echo.

clean:
	$(RM) *~
	$(RM) *.bak
	cd $(ISEARCH_DIR)
	$(MAKE) -$(MAKEFLAGS) -i -f Isearch.mak clean
	cd ..
	cd $(SAPI_DIR)
	$(MAKE) -$(MAKEFLAGS) -i -f sapi.mak clean
	cd ..
	cd $(ZDIST_DIR)
	$(MAKE) -$(MAKEFLAGS) -i -f zdist.mak clean
	cd ..
	cd $(BIN_DIR)
	$(RM) *~
	$(RM) *.bak
	$(RM) *.pdb
	$(RM) *.pck
	$(RM) *.lib
	$(RM) TESTHTML*
	$(RM) *.log
	$(RM) Iindex.exe
	$(RM) Isearch.exe
	$(RM) Iutil.exe
	$(RM) Isearch.lib
	cd ..

realclean:
	$(RM) *~
	$(RM) *.bak
	cd $(ISEARCH_DIR)
	$(MAKE) -$(MAKEFLAGS) -i -f Isearch.mak realclean
	cd ..
	cd $(SAPI_DIR)
	$(MAKE) -$(MAKEFLAGS) -i -f sapi.mak realclean
	cd ..
	cd $(ZDIST_DIR)
	$(MAKE) -$(MAKEFLAGS) -i -f zdist.mak realclean
	cd ..
	cd $(BIN_DIR)
	$(RM) *~
	$(RM) *.bak
	$(RM) *.pdb
	$(RM) *.pck
	$(RM) *.lib
	$(RM) TESTHTML*
	$(RM) *.log
	$(RM) Iindex.exe
	$(RM) Isearch.exe
	$(RM) Iutil.exe
	$(RM) Isearch.lib

