/* $Id: db_Isearch.cxx,v 1.5 1997/05/10 00:43:36 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		db_Isearch.cxx
Version:	$Revision: 1.5 $
Description:	SAPI Isearch database object
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#include <sys/types.h>
#include <fcntl.h>
#include <fstream.h>
#include <sys/stat.h>

#if defined(_MSDOS) || defined(_WIN32)
#include <limits.h>
#include <sys/locking.h>
#include <direct.h>
#include <io.h>
#endif

#include "isearch.hxx"
#include "common.hxx"
/*
#include "defs.hxx"      
#include "string.hxx"    
#include "vlist.hxx"                                                     
#include "attr.hxx"                                                
#include "attrlist.hxx"  
#include "strlist.hxx"                                                 
#include "strstack.hxx"                                                  
#include "nfield.hxx" 
#include "mdtrec.hxx" 
#include "mdt.hxx"                                                    
#include "dfd.hxx"                                                    
#include "dfdt.hxx"                                                     
#include "fc.hxx"                    
#include "fct.hxx"    
#include "df.hxx"                                                       
#include "dft.hxx"                                                      
#include "record.hxx"                                                     
#include "result.hxx"  
#include "idbobj.hxx"  
#include "iresult.hxx"                                                    
#include "opobj.hxx"                                                     
#include "operand.hxx"                                                    
#include "termobj.hxx" 
#include "sterm.hxx"                                              
#include "rset.hxx"                                                      
#include "irset.hxx"                                   
#include "opstack.hxx" 
#include "squery.hxx"                                              
*/
#include "dtreg.hxx"                                                    
//#include "rcache.hxx"                                                     
//#include "index.hxx"                                                    
#include "fprec.hxx"                                       
#include "fpt.hxx"                                                       
#include "rcache.hxx"                                                     
#include "registry.hxx"
#include "index.hxx"
#include "idb.hxx"   
/*
#include "dfd.hxx"
#include "dfdt.hxx"
*/

#include "db_Isearch.hxx"

SAPI_DB_ISEARCH::SAPI_DB_ISEARCH(const STRING & Name, REGISTRY *Registry) 
	: SAPI_DB(Name, Registry)
{
  i_idb = (IDB *)NULL;
}


SAPI_DB_ISEARCH::~SAPI_DB_ISEARCH()
{
  if (i_idb)
    delete i_idb;
}


GDT_BOOLEAN 
SAPI_DB_ISEARCH::Initialize(STRING pass)
{
  STRLIST Pos, ResultList;

  Pos.Clear();
  Pos.AddEntry(d_name);
  Pos.AddEntry("Location");
  d_registry->GetData(Pos, &ResultList);
  if(ResultList.GetTotalEntries() <= 0) {
    d_errorMessage = "No LOCATION specified";
    return GDT_FALSE;
  }
  ResultList.GetEntry(1, &d_location);

  cerr << "Location is " << d_location << endl;
 
  /*
   * Create the record directory.
   */
  
  struct stat statbuf;
 
  i_records_directory = d_location;
  i_records_directory += "/";
  i_records_directory += d_name;
  i_records_directory += ".records";
 
  if (stat(i_records_directory, &statbuf) < 0) {
#if defined(MSDOS) || defined(_WIN32)
    if (mkdir(i_records_directory) < 0)
#else
    if (mkdir(i_records_directory, 0755) < 0)
#endif
      {
	d_errorMessage = "Unable to create records directory.";
	return GDT_FALSE;
      }
  }
  else if ((statbuf.st_mode & S_IFMT) != S_IFDIR) {
    d_errorMessage = "Records directory is not a valid directory.";
    return GDT_FALSE;
  }
  
  /*
   * Create the IDB object.
   */
 
  if((i_idb = new IDB(d_location,d_name)) == NULL) {
    d_errorMessage = "Failed to create IDB object";
    return GDT_FALSE;
  }
  
  // "Validate" by checking number of records 
  if(i_idb->GetTotalRecords() <= 0) {
    d_errorMessage = "0 records available";
    return GDT_FALSE;
  }
  d_type="ISEARCH";
  return GDT_TRUE;
}


SAPI_RSET *
SAPI_DB_ISEARCH::Search(const ZSQUERY & Query)
{
  SAPI_RSET *RSet;
  RSET *rset;
  IRSET *pirset;
  INT4 Count;
  
  
  pirset = i_idb->Search(Query);
  Count=pirset->GetTotalEntries();
  pirset->SortByScore(); // Sort the scores
  rset=pirset->GetRset(0,Count);

  pirset->Fill(0,Count,rset); // Fill up the result set
  delete pirset; // now we're done with the internal result set

  // Create the right kind of rset to send back to the server
  if (rset) {
    RSet = new SAPI_RSET_ISEARCH(this, rset);
    RSet->SetHitCount(rset->GetTotalEntries());
  } else 
    RSet = (SAPI_RSET *)NULL;
  
  return RSet;
}



GDT_BOOLEAN
SAPI_DB_ISEARCH::Add_Record(const STRING &Key, const STRING &Data)
{
  STRING sequenceFilename;

  /*
   * Open the sequence file.
   */
 
  sequenceFilename = i_records_directory;
  sequenceFilename += "/Sequence";
 
  fstream sequenceFile(sequenceFilename, ios::in | ios::out, 0644);
 
  if ( !sequenceFile ) {
    d_errorMessage = "Unable to open the sequence file.";
    return GDT_FALSE;
  }
 
  /*
   * Lock the sequence file.
   */
 
#if defined(_MSDOS) || defined(_WIN32)
  lseek(sequenceFile.rdbuf()->fd(), SEEK_SET, 0);
  if (locking(sequenceFile.rdbuf()->fd(), LK_LOCK, LONG_MAX) == -1)
#else
    struct flock lock;
 
  lock.l_type = F_WRLCK;
  lock.l_whence = 0;
  lock.l_start = 0;
  lock.l_len = 0;
 
  if (fcntl(sequenceFile.rdbuf()->fd(), F_SETLKW, &lock) < 0)
#endif
    {
      d_errorMessage = "fcntl failed to lock the sequence file.";
      return GDT_FALSE;
    }
 
  /*
   * Read and update the sequence number.
   */
 
  UINT4 sequence;
 
  if ( !(sequenceFile >> sequence) ) {
    sequence = 0;
    sequenceFile.clear();
  }
 
  sequenceFile.seekp(0);
 
  sequenceFile << (sequence + 1) << endl;
   
  sequenceFile.flush();
 
  /*
   * Unlock the sequence file.
   */
 
#if defined(_MSDOS) || defined(_WIN32)
  lseek(sequenceFile.rdbuf()->fd(), SEEK_SET, 0);
  if (locking(sequenceFile.rdbuf()->fd(), LK_UNLCK, LONG_MAX) == -1)
#else
    lock.l_type = F_UNLCK;
 
  if (fcntl(sequenceFile.rdbuf()->fd(), F_SETLKW, &lock) < 0)
#endif
    {
      d_errorMessage = "fcntl failed to unlock the sequence file.";
      return GDT_FALSE;
    }
 
  sequenceFile.close();
 
  /*
   * Ensure that the proper directory structure exists.
   */
 
  STRING dataFilename;
  char buf[16];
  struct stat statbuf;
 
  dataFilename = i_records_directory;
  dataFilename += "/0x";
 
  for (int i = 24; i; i -= 8) {
    sprintf(buf, "%.2lx", (unsigned long)(sequence >> i) & 0xff);
    dataFilename += buf;
 
    if (stat(dataFilename, &statbuf) < 0) {
#if defined(MSDOS) || defined(_WIN32)
      if (mkdir(dataFilename) < 0)
#else
	if (mkdir(dataFilename, 0755) < 0)
#endif
	  {
	    d_errorMessage = "Unable to create directory structure for record.";
	    return GDT_FALSE;
	  }
    }
    else if ((statbuf.st_mode & S_IFMT) != S_IFDIR) {
      d_errorMessage = "Record directory stucture is not valid.";
      return GDT_FALSE;
    }
 
    dataFilename += '/';
  }
 
  /*
   * Store the record.
   */
 
  sprintf(buf, "%.2lx.t", (unsigned long)sequence & 0xff);
  dataFilename += buf;
 
  ofstream dataFile(dataFilename, ios::out, 0600);
 
  if ( !dataFile || !(dataFile << Data) ) {
    d_errorMessage = "Unable to store the record.";
    return GDT_FALSE;
  }
 
  dataFile.close();
 
  /*
   * Index the record.
   */
 
  RECORD Record;
  STRING DocType;
  STRING S;
 
  Record.SetKey(Key);
 
  i_idb->GetGlobalDocType(&DocType);
  if (DocType == "")
    DocType = "SIMPLE";
  Record.SetDocumentType(DocType);
 
  S = dataFilename;
  RemoveFileName(&S);
  Record.SetPathName(S);
 
  S = dataFilename;
  RemovePath(&S);
  Record.SetFileName(S);
 
  Record.SetRecordStart(0);
  Record.SetRecordEnd(0);
 
  i_idb->AddRecord(Record);
 
  i_idb->Index();
 
  return GDT_TRUE;
}


GDT_BOOLEAN
SAPI_DB_ISEARCH::Delete_Record(const STRING &Key)
{
  if (i_idb->DeleteByKey(Key) != 1)
    return GDT_FALSE;
  
  return GDT_TRUE;
}


//void SAPI_DB_ISEARCH::GetFieldNames(SAPI_DB_ISEARCH *d, STRLIST *FieldNames)
void 
SAPI_DB_ISEARCH::GetFieldNames(STRLIST *FieldNames)
{
  STRING DbType;
  FieldNames->Clear();
  GetType(&DbType);
  if(!(DbType.Equals("ISEARCH")))
    return;

  DFDT dfdt;
  DFD dfd;
  i_idb->GetDfdt(&dfdt);
  INT count, i;
  STRING FieldName;
  count = dfdt.GetTotalEntries();
  for(i=1;i <= count;i++) {
    dfdt.GetEntry(i, &dfd);
    dfd.GetFieldName(&FieldName);
    FieldNames->AddEntry(FieldName);
  }
}
