/* $id$ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		db_script.cxx
Version:	$Revision: 1.4 $
Description:	SAPI script database object (for external DBs)
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef UNIX
#include <unistd.h>
#else
#include <process.h>
#endif

#include "isearch.hxx"
#include "registry.hxx"

#include "db_script.hxx"

SAPI_DB_SCRIPT::SAPI_DB_SCRIPT(const STRING & Name, REGISTRY *Registry) 
	: SAPI_DB(Name, Registry)
{
  STRING Pass;
  Pass="";
//  Initialize(Pass);
}


SAPI_DB_SCRIPT::~SAPI_DB_SCRIPT()
{
  StrUnlink(sc_Bresults);
}


GDT_BOOLEAN 
SAPI_DB_SCRIPT::Initialize(STRING pass)
{
  STRLIST Pos, ResultList;
  STRING PID;

  Pos.Clear();
  Pos.AddEntry(d_name);
  Pos.AddEntry("Location");
  d_registry->GetData(Pos, &ResultList);
  if(ResultList.GetTotalEntries() <= 0) {
    d_errorMessage = "No LOCATION specified";
    return GDT_FALSE;
  }
  ResultList.GetEntry(1, &d_location);

  cerr << "Script name is " << d_location << endl;

  Pos.Clear();
  Pos.AddEntry(d_name);
  Pos.AddEntry("Results");
  d_registry->GetData(Pos, &ResultList);
  if(ResultList.GetTotalEntries() <= 0) {
    d_errorMessage = "No result file specified";
    return GDT_FALSE;
  }
  ResultList.GetEntry(1, &sc_Bresults);
  PID = getpid();
  sc_Bresults.Cat(".");
  sc_Bresults.Cat(PID);

  cerr << "Results will go to " << sc_Bresults << endl;

  Pos.Clear();
  Pos.AddEntry(d_name);
  Pos.AddEntry("GetFull");
  d_registry->GetData(Pos, &ResultList);
  if(ResultList.GetTotalEntries() <= 0) {
    d_errorMessage = "No script specified to retrieve full records";
    return GDT_FALSE;
  }
  ResultList.GetEntry(1, &sc_Fresults);

  cerr << "Records will be retrieved using " << sc_Fresults << endl;

  d_type="SCRIPT";
  return GDT_TRUE;
}


SAPI_RSET *
SAPI_DB_SCRIPT::Search(const ZSQUERY & Query)
{
  SAPI_RSET *RSet;
  STRING QueryString;
  PCHR Script;
  PCHR ResultsFile;
  PCHR cQueryString;
  INT Diag;
  INT Hits;

  Query.GetKWAQSTerm(&QueryString);

  Hits = script_Search(d_location, QueryString, &Diag);

  RSet = new SAPI_RSET_SCRIPT(this);
  RSet->SetHitCount(Hits);
  return RSet;
}


INT4 
SAPI_DB_SCRIPT::script_Search(const STRING & Script, 
			      //   const STRING & ResultsFile, 
			      const STRING & Query, INT *Diag)
{
  PCHR cmd;
  STRLIST Position;
  STRING Hold;
  PCHR Separator;
  INT HitCount, Diagnostic;

  if (Script.GetLength() == 0) {
    fprintf(stderr, "script_Search: NULL Script value\n");
    return -1;
  }

//  if (ResultsFile.GetLength() == 0) {
  if (sc_Bresults.GetLength() == 0) {
    fprintf(stderr, "script_Search: NULL ResultsFile value\n");
    return -1;
  }
  if (Query.GetLength() == 0) {
    fprintf(stderr, "script_Search: NULL Query value\n");
    return -1;
  }

  //  Call the command line script
  Hold = Script;
  Hold.Cat(" ");
  Hold.Cat(sc_Bresults);
  Hold.Cat(" ");
  Hold.Cat(" \""); // Add leading quotes
  Hold.Cat(Query);
  Hold.Cat("\"");  // Add trailing quotes

  cmd = Hold.NewCString();

#ifdef SAPI_DEBUG
  cerr << "\tTempFile:" << sc_Bresults << endl;
  cerr << "\tQuery:" << Query << endl;
  cerr << "\tCmd:" << cmd << endl;
#endif

  system(cmd);
        
  /* Now use SCRIPT API to read results */
  Defaults = new REGISTRY("Results");
  Defaults->ProfileLoadFromFile(sc_Bresults, Position);

  Defaults->ProfileGetString("Default", "HitCount", "", &Hold);
  HitCount = Hold.GetInt();
  Defaults->ProfileGetString("Default", "Diagnostic", "", &Hold);
  Diagnostic = Hold.GetInt();

  *Diag = Diagnostic;
  delete cmd;

  return HitCount;
}

