/* $Id: db_zclient.cxx,v 1.2 1997/05/05 21:46:30 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		db_zclient.cxx
Version:	$Revision: 1.2 $
Description:	SAPI object to contact other Z39.50 servers
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#include "defs.hxx"      
#include "string.hxx"    
#include "vlist.hxx"                                                     
#include "attr.hxx"                                                
#include "attrlist.hxx"  
#include "strlist.hxx"                                                 
#include "strstack.hxx"                                                  
#include "nfield.hxx" 
#include "mdtrec.hxx" 
#include "mdt.hxx"                                                    
#include "dfd.hxx"                                                    
#include "dfdt.hxx"                                                     
#include "fc.hxx"                    
#include "fct.hxx"    
#include "df.hxx"                                                       
#include "dft.hxx"                                                      
#include "record.hxx"                                                     
#include "result.hxx"  
#include "idbobj.hxx"  
#include "iresult.hxx"                                                    
#include "opobj.hxx"                                                     
#include "operand.hxx"                                                    
#include "termobj.hxx" 
#include "sterm.hxx"                                              
#include "rset.hxx"                                                      
#include "irset.hxx"                                   
#include "opstack.hxx" 
#include "squery.hxx"                                              
#include "dtreg.hxx"                                                    
#include "rcache.hxx"                                                     
#include "index.hxx"                                                    
#include "fprec.hxx"                                       
#include "fpt.hxx"                                                       
#include "registry.hxx"
#include "idb.hxx"   
#include "dfd.hxx"
#include "dfdt.hxx"

#include "db_zclient.hxx"
#include "rset_zclient.hxx"

SAPI_DB_ZCLIENT::SAPI_DB_ZCLIENT(const STRING & Name, REGISTRY *Registry) 
	: SAPI_DB(Name, Registry)
{
  z_clientCount = 0;
}


SAPI_DB_ZCLIENT::~SAPI_DB_ZCLIENT()
{
  INT i;
  for(i=0;i < z_clientCount;i++)
    z_clientList[i].Close();
}


GDT_BOOLEAN 
SAPI_DB_ZCLIENT::Initialize(STRING pass)
{

  //
  // The LOCATION directive is a comma-delimited list of 
  // Z39.50 session URLs.
  // The TimeOut is a value in seconds to wait for a server to respond.
  //
  // [Metabase]
  // Type=ZCLIENT
  // Location=vinca.cnidr.org/TEST,blah.com:666/BLAH
  // TimeOut=20
  //
  STRLIST Pos, UrlList;
  Pos.Clear();
  Pos.AddEntry(d_name);
  Pos.AddEntry("Location");
  d_registry->GetData(Pos, &UrlList);
  INT HostCount = UrlList.GetTotalEntries();
  if(HostCount <= 0) {
    d_errorMessage = "No LOCATION specified";
    return GDT_FALSE;
  }
  STRING url, temp_host, temp_db;
  STRINGINDEX p;
  INT i;
  for(i = 1;i <= HostCount;i++) {
    UrlList.GetEntry(i, &url);
    if((p = url.Search(':')) == 0) {
      z_portList[i-1] = 210;
      temp_host = url;
      temp_db = url;
    } else {
      temp_host = url;
      temp_host.EraseAfter(p-1);

      url.EraseBefore(p+1);
      p = url.Search('/');
      
      STRING temp_port = url;
      temp_port.EraseAfter(p-1);
      char *port_as_string = temp_port.NewCString();
      z_portList[i-1] = atoi(port_as_string);
      delete [] port_as_string;	
      
      url.EraseBefore(p+1);
      temp_db = url;
    }

      p = temp_host.Search('/');
      if(p) {
              temp_host.EraseAfter(p-1);
              temp_db.EraseBefore(p+1);
              }

    z_hostList.AddEntry(temp_host);
    z_databaseList.AddEntry(temp_db);
    z_clientCount++;
	
  }
  
  STRING host;
  for(i=1;i <= z_clientCount;i++) {
    z_hostList.GetEntry(i, &host);
    z_databaseList.GetEntry(i, &temp_db);
    
    if(d_debugLevel > 1) {
      cerr << "Trying to connect to " << host << ", port ";
      cerr << z_portList[i-1] << " and search database ";
      cerr << temp_db << "...";
    }
    // add password support here

    if(!z_clientList[i-1].Initialize(host, z_portList[i-1], z_password)) {
      if(d_debugLevel > 1) {
	cerr << "Failed [";
	cerr << z_clientList[i-1].GetLastError();
	cerr << "]" << endl;
      }
      continue;
    }
    if(d_debugLevel > 1) 
      cerr << "Success" << endl;
  }
  d_type="ZCLIENT";
  return GDT_TRUE;
}


SAPI_RSET *
SAPI_DB_ZCLIENT::Search(const ZSQUERY & Query)
{
  SAPI_RSET *RSet = (SAPI_RSET *)NULL;
  STRING RSN="Default";
  ZRECORDLIST Rec;
  STRING kwaqs_query,FakeUser="Fake";
  INT4 HitCount, FetchCount, i, TotalHits=0;
  STRING database;
  INT N_Requested;
  ZSQUERY tQuery;
  STRING host;

  tQuery = Query;

  Query.GetKWAQSTerm(&kwaqs_query);
  if(d_debugLevel > 5) 
    cout << "sapi_db_zclient: " << kwaqs_query << endl;

  N_Requested = tQuery.GetHitsRequested();

  // Split up the requested number of hits between the hosts
  if (z_clientCount <= N_Requested)
    N_Requested = N_Requested / z_clientCount;

  for(i=1;i <= z_clientCount;i++) {
    z_databaseList.GetEntry(i, &database);
    z_hostList.GetEntry(i, &host);
    if (!z_clientList[i-1].Search(database,kwaqs_query,RSN, 
				 &HitCount,FakeUser,1,
				 N_Requested,&Rec)) {
      STRING temp_db;
      cerr << "Search failed against " << database << endl;
      z_databaseList.GetEntry(i, &temp_db);
    
      if(d_debugLevel > 1) {
	cerr << "Trying to reconnect to " << host << ", port ";
	cerr << z_portList[i-1] << " and search database ";
	cerr << temp_db << "...";
      }
    
      if(!z_clientList[i-1].Initialize(host, z_portList[i-1])) {
	if(d_debugLevel > 1) {
	  cerr << "Failed [";
	  cerr << z_clientList[i-1].GetLastError();
	  cerr << "]" << endl;
	}
	continue;		// search failed twice - go to next db
      }				// reinit succeeded
      if(d_debugLevel > 1) 
	cerr << "Successfuly restablished connection" << endl;
      if(!z_clientList[i-1].Search(database,kwaqs_query,RSN, 
				   &HitCount,FakeUser,1,N_Requested,&Rec)) {
	cerr << "Search failed against " << database << endl;
	continue;
      } else {
	if(d_debugLevel > 5) 
	  cout << "Got " << HitCount << " hits from" << host << endl;
      }

    } else {
      if(d_debugLevel > 5) 
	cout << "Got " << HitCount << " hits from " << host << endl;
    }
    TotalHits += HitCount;
  }
  
  RSet = new SAPI_RSET_ZCLIENT(this);
  RSet->SetHitCount(TotalHits);
  
  return RSet;
}

