#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#ifdef UNIX
#include <unistd.h>
#else
#include <process.h>
#endif

#include "isearch.hxx"
#include "registry.hxx"
#include "rset_script.hxx"
#include "db_script.hxx"
#include "mime_record.hxx"

SAPI_RSET_SCRIPT::SAPI_RSET_SCRIPT(SAPI_DB_SCRIPT *Database) : SAPI_RSET(Database)
{
}

SAPI_RSET_SCRIPT::~SAPI_RSET_SCRIPT()
{
}

GDT_BOOLEAN SAPI_RSET_SCRIPT::Present(const INT4 StartPoint, 
				      const INT4 NumberRequested, 
				      const STRING & ElementSetName,
				      const STRING & RecordSyntax, 
				      PLIST *RecordList)
{
  MIME_RECORD *mr;
  SAPI_DB_SCRIPT *db;
  GDT_BOOLEAN Status=GDT_FALSE;
  STRING Buffer, PID, Hold;
  PCHR TempFile;
  PCHR Separator;
  STRLIST Position;
  FILE *fp;
  PCHR b;
  INT4 RecStart,RecEnd,ActualLength,len;

  db = (SAPI_DB_SCRIPT *)r_database;
  Defaults = new REGISTRY("Results");

  PID = getpid();
  Hold = db->sc_Bresults;
  TempFile = Hold.NewCString();

  Defaults->ProfileLoadFromFile(TempFile, Position);

  /* Get the separator string */
  Hold = "";
  Defaults->ProfileGetString("Default", "Separator", "", &Hold);
  if(Hold.GetLength()==0) {
    fprintf(stderr,
	    "script_GetRecord:Bad Separator[%s]\n", TempFile);
    return GDT_FALSE;
  }
  Separator = Hold.NewCString();

  fp=fopen(TempFile,"r");
  if(fp==NULL) {
    fprintf(stderr, "script_GetRecord:Can't open %s\n",
	    TempFile);
    return GDT_FALSE;
  }

  fseek(fp, 0L, SEEK_END);
  RecStart = 0;
  RecEnd = ftell(fp);

  fseek(fp, (long)RecStart, SEEK_SET);
  len = RecEnd - RecStart;
  b = new CHR[len + 1];
  ActualLength = fread(b, 1, len, fp);
  b[ActualLength] = '\0';
  fclose(fp);

  PCHR pBuf;
  PCHR cmd;
  STRING sTempFile;
  INT4 i;

  // Loop over the lines in the file, breaking on newlines
  pBuf = strtok(b,"\n"); // [Default]
  pBuf = strtok((CHR*)NULL,"\n"); // HitCount
  pBuf = strtok((CHR*)NULL,"\n"); // Diagnostic
  pBuf = strtok((CHR*)NULL,"\n"); // Separator
  pBuf = strtok((CHR*)NULL,"\n"); // [Data]

  // Skip to the starting point, if necessary
  for (i=0;i<StartPoint-1;i++) {
    pBuf = strtok((CHR*)NULL,"\n"); // result record key
    pBuf = strtok((CHR*)NULL,"\n"); // result record
    pBuf = strtok((CHR*)NULL,"\n"); // separator string
  }

  for(i=StartPoint;i<=NumberRequested+StartPoint-1;i++) {
		
    if (ElementSetName.CaseEquals("B")) {
      pBuf = strtok((CHR*)NULL,"\n"); // result record key
      pBuf = strtok((CHR*)NULL,"\n"); // result record
      Buffer = pBuf;
      mr = new MIME_RECORD(Buffer);
      RecordList->AddTail(mr);
      Buffer = "";
      Status = GDT_TRUE;
    } else if (ElementSetName.CaseEquals("F")) {

      pBuf = strtok((CHR*)NULL,"\n"); // result record key

      TempFile = tempnam("/tmp","");
      sTempFile = TempFile;
      Hold = db->sc_Fresults;
      Hold.Cat(" ");
      Hold.Cat(sTempFile);
      Hold.Cat(" ");
      Hold.Cat(pBuf);
      cmd = Hold.NewCString();

      system(cmd);

      Buffer.ReadFile(sTempFile);
      StrUnlink(sTempFile);

      mr = new MIME_RECORD(Buffer);
      RecordList->AddTail(mr);

      pBuf = strtok((CHR*)NULL,"\n"); // result record
      delete cmd;

      Status = GDT_TRUE;
      pBuf = strtok((CHR*)NULL,"\n"); // separator string
      
      if (pBuf == NULL) // end of results file
	break;
    } else
      Status = GDT_FALSE;
  }
  
  delete Defaults;
  delete TempFile;
  delete b;

  return Status;
}

