#include "isearch.hxx"

#include "registry.hxx"
#include "sapi_db.hxx"
#include "plist.hxx"
#include "mime_record.hxx"

#include "sapi.hxx"
#include "db_conf.hxx"

SAPI::SAPI(const STRING &ConfigFile)
{
  // Parse input file, create SAPI_DB object for each database
  // and add to internal pointer list
  s_registry = new REGISTRY("sapi");
  s_databaseCount = 0;

  STRLIST Pos, ResultList;
  s_registry->ProfileLoadFromFile(ConfigFile, Pos);

  Pos.Clear();
  Pos.AddEntry("Default");
  Pos.AddEntry("DbList");
  s_registry->GetData(Pos, &ResultList);
  INT4 i, dbcount = ResultList.GetTotalEntries();

  STRING DbName, Value;
  STRLIST ValueList;
  SAPI_DB *NewDbDriver;//db->s_clientList[i].GetHitCount()

  for(i=1;i<=dbcount;i++) {
    ResultList.GetEntry(i, &DbName);
    Pos.Clear();
    Pos.AddEntry(DbName);
    Pos.AddEntry("TYPE");
    s_registry->GetData(Pos, &ValueList);
    if(ValueList.GetTotalEntries() <= 0) {
      cerr << "SAPI ERROR: No TYPE provided for database ";
      cerr << DbName << ".  Skipping..." << endl;
      continue;
    }
    ValueList.GetEntry(1, &Value);
    NewDbDriver = NewSAPI_DB(Value, DbName, s_registry);
    /*
      if(NewDbDriver) {
      // Attempt to initialize the driver
      if(NewDbDriver->Initialize() == GDT_FALSE) {
      cerr << "SAPI ERROR: Driver initialization ";
      cerr << "failed [" << DbName << "]: ";
      STRING Reason;
      NewDbDriver->GetLastErrorString(&Reason);
      cerr << Reason << endl;
      delete NewDbDriver;
      } else {
      s_databaseList.AddTail(NewDbDriver);
      s_databaseCount++;
      }
      } else {
      cerr << "SAPI ERROR: Failed to create db driver for ";
      cerr << DbName << ".  Skipping..." << endl;
      }
    */
    s_databaseList.AddTail(NewDbDriver);
    s_databaseCount++;

  }
}


SAPI::SAPI(const STRING &ConfigFile, const STRLIST& ResultList)
{
  // Parse input file, create SAPI_DB object for each database
  // and add to internal pointer list
  s_registry = new REGISTRY("sapi");
  s_databaseCount = 0;

  //  STRLIST Pos, ResultList;
  STRLIST Pos;
  s_registry->ProfileLoadFromFile(ConfigFile, Pos);

  Pos.Clear();
  Pos.AddEntry("Default");
  //  Pos.AddEntry("DbList");
  //  s_registry->GetData(Pos, &ResultList);
  INT4 i, dbcount = ResultList.GetTotalEntries();

  STRING DbName, Value;
  STRLIST ValueList;
  SAPI_DB *NewDbDriver;//db->s_clientList[i].GetHitCount()

  for(i=1;i<=dbcount;i++) {
    ResultList.GetEntry(i, &DbName);
    Pos.Clear();
    Pos.AddEntry(DbName);
    Pos.AddEntry("TYPE");
    s_registry->GetData(Pos, &ValueList);
    if(ValueList.GetTotalEntries() <= 0) {
      cerr << "SAPI ERROR: No TYPE provided for database ";
      cerr << DbName << ".  Skipping..." << endl;
      continue;
    }
    ValueList.GetEntry(1, &Value);
    NewDbDriver = NewSAPI_DB(Value, DbName, s_registry);
    s_databaseList.AddTail(NewDbDriver);
    s_databaseCount++;

  }
}


SAPI::~SAPI()
{
  if(s_registry)
    delete s_registry;

  // Walk through list of db pointers and call destroy each
  IPOSITION p;
  SAPI_DB *db;

  p = s_databaseList.GetHeadPosition();
  while((db=(SAPI_DB *)s_databaseList.GetNext(&p))) {
    if(db)
      delete db;
  }
}


SAPI_DB	*
SAPI::GetDatabasePtr(const INT4 DbNum)
{
  IPOSITION Pos = s_databaseList.GetHeadPosition();
  SAPI_DB *db;

  db = (SAPI_DB *)s_databaseList.GetEntry(DbNum, &Pos);
  if(db == (SAPI_DB *)NULL) {
    cerr << "Failed to find db num " << DbNum << " in list";
    cerr << endl;
  }
  return db;
}


SAPI_DB	*
SAPI::GetDatabasePtr(const STRING &DbName)
{
  // Walk through list of db pointers, check DbName against
  // call to method SAPI_DB::GetName().  Return pointer or NULL
  IPOSITION p;
  SAPI_DB *db;
  STRING Name;

  p = s_databaseList.GetHeadPosition();
  while((db=(SAPI_DB *)s_databaseList.GetNext(&p))) {
    db->GetName(&Name);
    if(DbName.CaseEquals(Name))
      return db;
  }

  return (SAPI_DB *)NULL;
}

