#
# CNIDR Search API
#
# This makefile results in:
#
#       libsapi.lib     - Common API to various search/database systems
#
#  To use this library in an application, you must link with libcnidr.a
#  as well as any libraries specific to the database systems you've chosen
#  to compile in.
#
#  Compile only those search engines you wish (are able) to support
#

RM=del

#
# Compiler and Compiler Flags
#
!IFNDEF WIN16
!include <ntwin32.mak>
#CFLAGS=$(cflags) $(cvarsmt) $(cdebug) -DVERS=\"2.01\"
CFLAGS=$(cflags) $(cvarsmt) -DVERS=\"2.01\"
#CFLAGS=$(cflags) $(cvars) $(cdebug) -DVERS=\"2.01\"
!ELSE
CFLAGS=-Alfw -F 4000 -G3 -Oceglot -W3 -Zi -DVERS=\"2.01\"
!ENDIF

#
# Library Directory
#
# Where should I store the library?
#
LIB_DIR=..\exe

#
# libz3950
#
# Where are the libz3950 include files?
#
ZDIST_DIR=..\zdist2
ZDIST_INC=-I$(ZDIST_DIR)

#
# Isearch
#
# Where is the Isearch distribution?
#
ISEARCH_DIR=..\Isearch
ISEARCH_INC=-I$(ISEARCH_DIR)\src -I$(ISEARCH_DIR)\doctype
ISEARCH_CFLAGS=
ISEARCH_DB_DRIVER=db_Isearch.obj rset_Isearch.obj

#
# Script Search Engine
#
#SCRIPT_OBJ=script.o
SCRIPT_INC=.

#
# Search Engines
#
#DBASE_OBJ - Any object files that need to be linked into libsapi
#DBASE_INC - Location of include files for all database engines
#DBASE_CFLAGS - Any compiler flags required by particular database engines
#
DBASE_OBJ=
DBASE_INC=$(ISEARCH_INC)
DBASE_CFLAGS=
DBASE_H=db_Isearch.hxx rset_Isearch.hxx db_zclient.hxx rset_zclient.hxx

#
# PGIF (Postgres interface)
#
# Where are the PGIF files?
#
PGIF_DIR=
PGIF_OBJ=
PGIF_INC=
#
# Where is the Postgres distribution? (not PGIF)
#
POSTGRES_DIR=
#

#
# Resulting Library
#
SAPI_LIB=$(LIB_DIR)\libsapi.lib

#
# Librarian
#
# Select your platform and comment/uncomment the appropriate ARFLAGS
#
AR=lib
!IFNDEF WIN16
ARFLAGS=-out:$(SAPI_LIB)
!ELSE
ARFLAGS=$(SAPI_LIB) +
!ENDIF

#
# That should be all you need to configure
#
INC=$(ZDIST_INC) $(ISEARCH_INC)

H=mime_record.hxx sapi.hxx sapi_db.hxx db_conf.hxx plist.hxx \
        sapi_rset.hxx db_Isearch.hxx db_script.hxx $(DBASE_H)

OBJ=sapi.obj sapi_db.obj db_conf.obj plist.obj sapi_rset.obj db_zclient.obj \
        rset_zclient.obj mime_record.obj db_Isearch.obj \
        rset_Isearch.obj db_script.obj rset_script.obj

all:$(SAPI_LIB)

mime_record.obj:$(H) mime_record.cxx
        $(CC) $(CFLAGS) $(INC) -c mime_record.cxx

db_zclient.obj:$(H) db_zclient.cxx
        $(CC) $(CFLAGS) $(INC) -c db_zclient.cxx

rset_zclient.obj:$(H) rset_zclient.cxx
        $(CC) $(CFLAGS) $(INC) -c rset_zclient.cxx

rset_Isearch.obj:$(H) rset_Isearch.cxx
        $(CC) $(CFLAGS) $(INC) -c rset_Isearch.cxx

db_Isearch.obj:$(H) db_Isearch.cxx
        $(CC) $(CFLAGS) $(INC) -c db_Isearch.cxx

rset_script.obj:$(H) rset_script.cxx
        $(CC) $(CFLAGS) $(INC) -c rset_script.cxx

db_script.obj:$(H) db_script.cxx
        $(CC) $(CFLAGS) $(INC) -c db_script.cxx

db_conf.obj:$(H) db_conf.cxx
        $(CC) $(CFLAGS) $(INC) -c db_conf.cxx

plist.obj:$(H) plist.cxx
        $(CC) $(CFLAGS) $(INC) -c plist.cxx

sapi_rset.obj:$(H) sapi_rset.cxx
        $(CC) $(CFLAGS) $(INC) -c sapi_rset.cxx

sapi_db.obj:$(H) sapi_db.cxx
        $(CC) $(CFLAGS) $(INC) -c sapi_db.cxx

sapi.obj:$(H) sapi.cxx $(DBASE_OBJ)
        $(CC) $(CFLAGS) $(INC) -c sapi.cxx

$(SAPI_LIB):$(OBJ)
	$(RM) $(SAPI_LIB)
        $(AR) $(ARFLAGS) $(OBJ) $(DBASE_OBJ)

clean:
	$(RM) *~
	$(RM) *.pdb
	$(RM) *.bak
	$(RM) *.pck
	$(RM) *.obj
	$(RM) $(SAPI_LIB)

realclean:
	$(RM) *~
	$(RM) *.pdb
	$(RM) *.bak
	$(RM) *.pck
	$(RM) *.obj
	$(RM) $(SAPI_LIB)

