// $Id: berutil.cxx,v 1.5 1997/05/31 21:08:01 cnidr Exp $
/*
(c)1992 OCLC Online Computer Library Center, Inc., 6565 Frantz Road,
Dublin, Ohio 43017-0702.  OCLC is a registered trademark of OCLC Online
Computer Library Center, Inc.

NOTICE TO USERS: The BER Utilities ("Software") has been developed by
OCLC Online Computer Library Center, Inc.  Subject to the terms and
conditions set forth below, OCLC grants to user a perpetual,
non-exclusive, royalty-free license to use, reproduce, alter, modify,
and create derivative works from Software, and to sublicense Software
subject to the following terms and conditions:

SOFTWARE IS PROVIDED AS IS.  OCLC MAKES NO WARRANTIES, REPRESENTATIONS,
OR GUARANTEES WHETHER EXPRESS OR IMPLIED REGARDING SOFTWARE, ITS FITNESS
FOR ANY PARTICULAR PURPOSE, OR THE ACCURACY OF THE INFORMATION CONTAINED
THEREIN.

User agrees that :1) OCLC shall have no liability to user arising
therefrom, regardless of the basis of the action, including liability
for special, consequential, exemplary, or incidental damages, including
lost profits, even if it has been advised of the possibility thereof;
and :2) user will indemnify and hold OCLC harmless from any claims
arising from the use of the Software by user's sublicensees.

User shall cause the copyright notice of OCLC to appear on all copies of
Software, including derivative works made therefrom.
*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(MSDOS) || defined(_WIN32)
#include <strstrea.h>
#else
#include <strstream.h>
#endif

#include <iostream.h>

#define PORTINFO 1
#ifdef PORTINFO
#include "portinfo.hxx"
#endif

#include "berutil.hxx"

INT berutil_ByteOrder;	// 0 for LOW_TO_HIGH, 1 for HIGH_TO_LOW

#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif

#ifdef DEBUG
static INT level;
#endif

static INT bld_nod(UCHR *s, DATA_DIR *dir, INT4 *tot_len,
	DATA_DIR *parent);


//  builds a directory to an asn.1 record pointed to by s.  calls bld_nod
//  to do all the real work.
INT 
bld_dir(UCHR *s, DATA_DIR *dir)
{
  INT4     len=0;

  if(!dir)
    {

#ifdef DEBUG
      cerr << "in bld_dir: null directory passed in" << endl;
#endif

      return FALSE;
    }

#ifdef DEBUG
  cerr << "in bld_dir: dir=" << dir << ", dir->block=" << dir->block 
       << ", dir->block->dir=" << dir->block->dir << endl;
#endif

  dinit(dir, 0, 0);
  dir=dir->block->dir;
  dir->block->next_dir=dir;

#ifdef DEBUG
  level=0;
#endif

  return bld_nod(s, dir, &len, NULL_DIR);
}


/*
    bld_nod is called with a pointer to an asn.1 field, a pointer to a INT4
    so that it can return the total length of the asn.1 field, and a pointer
    to the node that is to be this node's parent.  if build_node discovers
    that the field has subfields, then it calls itself to handle the subfields,
    passing the node that it created as the parent for the new_node nodes
*/
static INT 
bld_nod(UCHR *ptr, DATA_DIR *dir, INT4 *tot_len, DATA_DIR *parent)
{
  UCHR      c, *s=ptr;
  DATA_DIR *child, *last, *node;
  INT       lenlen, taglen;
  INT4      field_len, sub_len, tlen;

#ifdef DEBUG
  UCHR     *t;
  INT4      i;
  cerr << "in bld_nod: dir=" << dir << ", dir->block=" << dir->block 
       << ", dir->block->next_dir=" << dir->block->next_dir << endl;
  cerr << "    dir->block->last_dir=" << dir->block->last_dir << endl;
#endif

  if(dir->block->next_dir > dir->block->last_dir)
    if(!dmake_bigger(dir))
      {

#ifdef DEBUG
	cerr << "in bld_nod: 1) dir filled up" << endl;
	level--;
#endif

	return FALSE;
      }
  node=dir->block->next_dir++;
  if(dir->block->next_dir <= dir->block->last_dir)
    dir->block->next_dir->block=dir->block;

#ifdef DEBUG
  cerr << "node=" << node;
  if(parent)
    cerr << " last_dir=" << parent->block->last_dir;
  cerr << " level=" << ++level << endl;
#endif

  if(parent)
    parent->count++;

  node->parent=parent;
  node->next = NULL_DIR;
  node->Class = *s>>6;   // the first byte of the asn.1 field
  node->form=(*s>>5)&1;  // contains the class and ASN1_CONSTRUCTED
                         // information

  c=(*s++)&0x1f;         // the first byte of the tag

  *tot_len=1;
  if(c<0x1f)             // if the tag is less than 31, then it
    node->fldid=c;       // is fully contained in the one byte

  else                   // otherwise we keep looking at bytes
    {                    // until we find 1 with the sign bit off
      node->fldid=0;     // catenating the last 7 bits of each byte
      c= *s++;
      while(c>0x80)
        {
	  *tot_len+=1;
	  node->fldid+=c&0x7F;
	  node->fldid<<=7;
	  c= *s++;
        }
      *tot_len+=1;
      node->fldid+=c;
    }
  taglen= (INT)*tot_len;

  lenlen=get_len(&field_len, s, 0x7fffffff);
  if(field_len == -1)  // indeterminate length
    if(IsCompleteBER(ptr, 0x7fffffff, &field_len))
      {

#ifdef DEBUG
	cerr << "indeterminate length determined to be " <<
	  field_len << endl;
#endif

	field_len-=(taglen+lenlen);
	*tot_len+=lenlen+field_len;
	field_len-=2;  // don't include trailing nulls in loop later
      }
    else
      {

#ifdef DEBUG
	cerr << "we don't think we have a complete record" << endl;
#endif

	return FALSE;
      }
  else
    *tot_len+=lenlen+field_len;
  s+=lenlen;

#ifdef DEBUG
  cerr << "class=" << node->Class << ", form=" << node->form
       << ", tag=" << node->fldid << ", len=" << field_len << endl;
#endif

  if(node->form==ASN1_CONSTRUCTED)
    {                    // field has subfields.  set up and recurse
      node->count=0;
      if(node->block->next_dir > node->block->last_dir)
	if(!dmake_bigger(node))
	  {
	    node->ptr.child=NULL_DIR;

#ifdef DEBUG
	    cerr << "in bld_nod: 2) dir filled up" << endl;
	    level--;
#endif

	    return FALSE;
	  }
      node->ptr.child=node->block->next_dir;
      tlen=field_len;

#ifdef DEBUG
      cerr << "in blddir: tlen=" << tlen << endl;
#endif

      last=NULL_DIR;
      while(tlen>0)    // we know how long the field is.  loop until
        {              // the total length of the subfields equals
	               // the field length.  don't exceed dir size
	  if(node->block->next_dir > node->block->last_dir)
	    if(!dmake_bigger(node))
	      {
		node->ptr.child=NULL_DIR;

#ifdef DEBUG
		cerr << "in bld_nod: 3) dir filled up" << endl;
		level--;
#endif

		return FALSE;
	      }
	  child=node->block->next_dir;
	  child->prev=last;
	  if(last)
	    last->next=child;
	  if(!bld_nod(s, dir, &sub_len, node))
            {

#ifdef DEBUG
	      level--;
#endif

	      return FALSE;
            }
	  last=child;
	  tlen-=sub_len;

#ifdef DEBUG
	  cerr << "in blddir: tlen=" << tlen << ", sub_len=" 
	       << sub_len << endl;
#endif

	  s+=sub_len;
        }
      if(last)  // this should always be the case
	last->next=NULL_DIR;
      else
        {

#ifdef DEBUG
	  cerr << "in bld_dir: a ASN1_CONSTRUCTED field of length "
	       << "<=0 detected" << endl;
	  cerr << "    field_len=" << field_len << endl;
#endif

	  node->ptr.child=NULL_DIR;
        }

      if(tlen)  // exceeded dir size
        {

#ifdef DEBUG
	  cerr << "in blddir: error, non-zero tlen=" << tlen << endl;
	  level--;
#endif

	  return FALSE;
        }
    }

  else  // no subfields, so just point at the data
    {
      node->count=field_len;
      node->ptr.data=s;

#ifdef DEBUG
      cerr << "data='";
      for(i=0, t=s; i<field_len; i++, t++)
	cerr << *t;
      cerr << endl;
#endif

    }

#ifdef DEBUG
  level--;
#endif

  return TRUE;
}


UCHR *
bld_rec(DATA_DIR *dir, INT4 *len)
{
  UCHR *record;

  *len=rec_len(dir);  // calculate the length of the record

#ifdef DEBUG
  cerr << "in bld_rec: about to allocate " << *len << " bytes" << endl;
  
#endif

  record = new UCHR[(unsigned)*len];
  // allocate space for the record
  if(record)
    asm_rec(dir, record);  // pull the pieces of the record together

  return record;
}


INT4 
asm_rec(DATA_DIR *dir, UCHR *record)
{
  INT       i;
  UCHR     *t;
  DATA_DIR *child;
  INT4      len, tlen, tot_len;

#ifdef DEBUG
  cerr << "in asm_rec: adding field " << (INT)dir->fldid << endl;
#endif

  if(dir->fldid<31)
    *record++ = (UCHR)(dir->fldid + dir->Class*64 + dir->form*32);
  else
    {
      *record++ = (UCHR)(31 + dir->Class*64 + dir->form*32);
      if(dir->fldid<128)
	*record++ = (UCHR)dir->fldid;
      else
        {
	  *record++ = (UCHR)(128 + dir->fldid/128);
	  *record++ = (UCHR)(dir->fldid%128);
        }
    }
  tot_len=TAG_LEN(dir);

  len=dir->length;
  if(len<128)
    {
      *record++ =(UCHR)len;
      tot_len+=1;
    }
  else
    {
      *record=(UCHR)(LEN_LEN(dir)-1);   // number of bytes in length
      tot_len+= *record+1;
      t=record+*record;  // go to end of length field and load backwards
      for(i=0; i<(INT)*record; i++)
        {
	  *t-- =(UCHR)(len & 0xFF);
	  len>>=8;
        }
      record+= *record+1;
      *t+=128;  // t points at length byte.  turn on bit 8
    }

  if(dir->form==ASN1_CONSTRUCTED)
    {
      child=dir->ptr.child;
      while(child)
        {
	  tlen=asm_rec(child, record);
	  record+=tlen;
	  tot_len+=tlen;
	  child=child->next;
        }
    }
  else
    {

#ifdef DEBUG
      cerr << "    which is " << dir->count << " bytes long" << endl;
      
#endif

      memcpy((char *)record, (char *)dir->ptr.data, (unsigned)dir->count);
      tot_len+=dir->count;
    }

  return tot_len;
}


DATA_DIR *
daddnum(DATA_DIR *msg, unsigned fldid, UCHR Class, UCHR *ptr, INT len)
{
  DATA_DIR *dir, *next_dir;
  UCHR     *t;
  INT4      number;
  int       longlen = sizeof(INT4);

  if(!msg || msg->form==ASN1_PRIMITIVE)
    return NULL_DIR;

  if(msg->block->next_dir > msg->block->last_dir)
    if(!dmake_bigger(msg))
      return NULL_DIR;
  next_dir=msg->block->next_dir++;
  if(msg->block->next_dir <= msg->block->last_dir)
    next_dir->block->next_dir->block=next_dir->block;

  msg->count++;
  if(msg->ptr.child)
    {
      dir=msg->ptr.child;
      while(dir->next)
	dir=dir->next;
      dir->next=next_dir;
      next_dir->prev=dir;
    }
  else
    {
      msg->ptr.child=next_dir;
      next_dir->prev=NULL_DIR;
    }

  next_dir->block=msg->block;
  next_dir->parent=msg;
  next_dir->next = NULL_DIR;
  next_dir->user = 0;

  if(len==1)
    number= *ptr;
  else if (longlen == 4)
    {
      if(len==2)
	number= *(short*)ptr;
      else
	number= *(long*)ptr;
    }
  else if (longlen == 8)
    {
      if (len == 2)
	number = *(short*)ptr;
      else if (len == 4)
	number = *(int*)ptr;
      else
	number = *(long*)ptr;
    }

  //  else
  //    if(len==2)
  //      number= *(short*)ptr;
  //    else
  //      number= *(INT4*)ptr;

  ptr=(UCHR*)&number;
  if(berutil_ByteOrder == 1)
    next_dir->number=number;
  else {
    //    UCHR *t;
    t=(UCHR*)(&next_dir->number);
    t[ 0 ]= ptr[ 3 ];
    t[ 1 ]= ptr[ 2 ];
    t[ 2 ]= ptr[ 1 ];
    t[ 3 ]= ptr[ 0 ];
  }

  //    #if (NATIVE_ORDER&PDP_ORDER)==PDP_ORDER
  //    UCHR *t;
  //    t=(UCHR*)(&next_dir->number);
  //    t[ 0 ]= ptr[ 1 ];
  //    t[ 1 ]= ptr[ 0 ];
  //    t[ 2 ]= ptr[ 3 ];
  //    t[ 3 ]= ptr[ 2 ];
  //    #endif

  next_dir->ptr.data=(UCHR*)(&next_dir->number);
  next_dir->count=longlen; // was =sizeof(INT4);
  if(next_dir->ptr.data[ 0 ]=='\0')
    while(next_dir->count>1 && next_dir->ptr.data[ 0 ]=='\0' &&
	  next_dir->ptr.data[ 1 ]<128)
      {
	next_dir->ptr.data++;
	next_dir->count--;
      }
  else
    if(next_dir->ptr.data[ 0 ]==0xff)
      while(next_dir->count>1 && next_dir->ptr.data[ 0 ]==0xff &&
	    next_dir->ptr.data[ 1 ]>127)
	{
	  next_dir->ptr.data++;
	  next_dir->count--;
	}

  next_dir->fldid=fldid;
  next_dir->Class=Class;
  next_dir->form=ASN1_PRIMITIVE;

  return next_dir;
}


DATA_DIR *
dalloc(INT numdirs)
{
  BLK_DIR  *blk_dir;
  DATA_DIR *dir=NULL_DIR;

  // Added by Mike Gursky
  if(numdirs < 1)
    numdirs = 1;
  // end

  if( (blk_dir=(BLK_DIR *)calloc(1,sizeof(BLK_DIR)))==NULL ) {
  //  blk_dir= new BLK_DIR [sizeof(BLK_DIR)];
  //  if (blk_dir==NULL) {

#ifdef DEBUG
    cerr << "in dalloc: malloc of blk_dir failed" << endl;
#endif

    return NULL_DIR;
  }

  if( (blk_dir->block_array=(DATA_DIR **)calloc(1,sizeof(UCHR *)))==NULL ) {
  //  blk_dir->block_array=new (DATA_DIR *)[sizeof(UCHR *)];
  //  if( (blk_dir->block_array)==NULL ) {

#ifdef DEBUG
    cerr << "in dalloc: malloc of block_array failed" << endl;
#endif

    return NULL_DIR;
  }

  if( (blk_dir->next_dir = blk_dir->dir = blk_dir->block_array[ 0 ] =
       dir = (DATA_DIR *)
       calloc(1,(unsigned)(numdirs*sizeof(DATA_DIR))))==NULL )
    {
      free((UCHR*)blk_dir);

#ifdef DEBUG
      cerr << "in dalloc: malloc of " << numdirs << " nodes failed" << endl;
#endif

      return NULL_DIR;
    }

  blk_dir->last_dir = &dir[ numdirs-1 ];
  blk_dir->num_blocks=1;
  blk_dir->num_nodes=numdirs;
  dir->block=blk_dir;

#ifdef DEBUG
  cerr << "in dalloc: blk_dir=" << blk_dir << ", dir=" << dir
       << ", next_dir=" << dir->block->next_dir << ", last_dir=" 
       << dir->block->last_dir << endl;
#endif

  return dir;
}

INT 
dfree(DATA_DIR *dir)
{
  BLK_DIR  *blk_dir;

  if(!dir)
    return FALSE;
  blk_dir=dir->block;
  while(blk_dir->num_blocks-- > 0) {
    free((UCHR*)blk_dir->block_array[ blk_dir->num_blocks ]);
  }
  free((UCHR*)blk_dir->block_array);
  free((UCHR*)blk_dir);
  return TRUE;
}


DATA_DIR *
dmake(unsigned fldid, UCHR Class, INT numdirs)
{
  DATA_DIR *dir;

  dir=dalloc(numdirs);
  if(dir)
    dinit(dir, fldid, Class);

  return dir;
}


DATA_DIR *
dinit(DATA_DIR *dir, unsigned fldid, UCHR Class)
{
  BLK_DIR *blk_dir;

#ifdef DEBUG
  cerr << "dir=" << dir << endl;
  cerr << "dir->block=" << dir->block << endl;
  cerr << "dir->block->dir=" << dir->block->dir << endl;
#endif

  blk_dir=dir->block;
  dir=blk_dir->dir;
  // fix by Mike Gursky
  blk_dir->next_dir = &dir->block->dir[ 1 ];
  blk_dir->last_dir = &dir[ blk_dir->num_nodes-1 ];
  if(blk_dir->next_dir <= blk_dir->last_dir)
    blk_dir->next_dir->block = blk_dir;
  // end

  // old
  // blk_dir->next_dir = &dir->block->dir[ 1 ];
  // blk_dir->next_dir->block=blk_dir;
  // blk_dir->last_dir = &dir[ blk_dir->num_nodes-1 ];

  while(blk_dir->num_blocks > 1)
    {
      blk_dir->num_blocks--;
      free((UCHR*)blk_dir->block_array[ blk_dir->num_blocks ]);
    }

  dir->parent = dir->prev = dir->next = dir->ptr.child = NULL_DIR;
  dir->count = dir->user = 0;

  dir->fldid=fldid;
  dir->Class=Class;
  dir->form=ASN1_CONSTRUCTED;

  return dir;
}


DATA_DIR *
daddchar(DATA_DIR *msg, unsigned fldid, UCHR Class, UCHR *ptr, INT4 len)
{
  DATA_DIR *dir, *next_dir;

  if(!msg || msg->form==ASN1_PRIMITIVE)
    return NULL_DIR;

  if(msg->block->next_dir > msg->block->last_dir)
    if(!dmake_bigger(msg))
      return NULL_DIR;
  next_dir=msg->block->next_dir++;
  if(msg->block->next_dir <= msg->block->last_dir)
    next_dir->block->next_dir->block=next_dir->block;

  msg->count++;
  if(msg->ptr.child)
    {
      dir=msg->ptr.child;
      while(dir->next)
	dir=dir->next;
      dir->next=next_dir;
      next_dir->prev=dir;
    }
  else
    {
      msg->ptr.child=next_dir;
      next_dir->prev=NULL_DIR;
    }

  next_dir->block=msg->block;
  next_dir->parent=msg;
  next_dir->next = NULL_DIR;
  next_dir->user = 0;

  next_dir->ptr.data=ptr;
  next_dir->count = len;

  next_dir->fldid=fldid;
  next_dir->Class=Class;
  next_dir->form=ASN1_PRIMITIVE;

  return next_dir;
}


DATA_DIR *
daddtag(DATA_DIR *msg, unsigned fldid, UCHR Class)
{
  DATA_DIR *dir, *next_dir;

  if(!msg || msg->form==ASN1_PRIMITIVE)
    return NULL_DIR;

  if(msg->block->next_dir > msg->block->last_dir)
    if(!dmake_bigger(msg))
      return NULL_DIR;
  next_dir=msg->block->next_dir++;
  if(msg->block->next_dir <= msg->block->last_dir)
    next_dir->block->next_dir->block=next_dir->block;

  msg->count++;
  if(msg->ptr.child)
    {
      dir=msg->ptr.child;
      while(dir->next)
	dir=dir->next;
      dir->next=next_dir;
      next_dir->prev=dir;
    }
  else
    {
      msg->ptr.child=next_dir;
      next_dir->prev=NULL_DIR;
    }

  next_dir->block=msg->block;
  next_dir->parent=msg;
  next_dir->ptr.child = next_dir->next = NULL_DIR;
  next_dir->count = next_dir->user = 0;

  next_dir->fldid=fldid;
  next_dir->Class=Class;
  next_dir->form=ASN1_CONSTRUCTED;

  return next_dir;
}


DATA_DIR *
dadddir(DATA_DIR *msg, DATA_DIR *new_nodedir)
{
  DATA_DIR *dir;

  if(!msg || msg->form==ASN1_PRIMITIVE)
    return NULL_DIR;

  msg->count++;
  if(msg->ptr.child)
    {
      dir=msg->ptr.child;
      while(dir->next)
	dir=dir->next;
      dir->next=new_nodedir;
      new_nodedir->prev=dir;
    }
  else
    {
      msg->ptr.child=new_nodedir;
      new_nodedir->prev=NULL_DIR;
    }

  new_nodedir->parent=msg;

  return new_nodedir;
}


INT 
ddeldir(DATA_DIR *msg)
{
  DATA_DIR *prev;

  if(!msg || !msg->parent)
    return FALSE;

  msg->parent->count--;

  // find prev
  prev=msg->parent->ptr.child;
  if(prev==msg)
    msg->parent->ptr.child=msg->next;
  else
    {
      while(prev->next!=msg)
	prev=prev->next;
      prev->next=msg->next;
    }
  return TRUE;
}

/* Taken from newer version of berutils from OCLC
int ddeldir(DATA_DIR *dir)
{
    dir->parent->count--;
    if (dir->prev)
      dir->prev->next=dir->next;
    if (dir->next)
      dir->next->prev=dir->prev;
    if (dir->parent->ptr.child==dir)
      dir->parent->ptr.child=dir->next;
    return(TRUE);
}
*/


INT 
dmake_bigger(DATA_DIR *dir)
{
  BLK_DIR  *blk_dir;

  if(!dir)
    return FALSE;
  blk_dir=dir->block;
  blk_dir->num_blocks++;

#ifdef DEBUG
  printf("in dmbigger1: dir = %lx blk_dir = %lx ", dir,blk_dir);
  printf("block_array = %lx num_blocks = %d\n",
	 blk_dir->block_array,blk_dir->num_blocks);
#endif

  if( (blk_dir->block_array=
       (DATA_DIR **)realloc((UCHR*)blk_dir->block_array,
			    blk_dir->num_blocks*sizeof(UCHR*)))==NULL )
    {

#ifdef DEBUG
      cerr << "in dmbigger: realloc of block_array failed" << endl;
#endif

      return FALSE;
    }

  if( (blk_dir->block_array[ blk_dir->num_blocks-1 ]=
       (DATA_DIR *)malloc(blk_dir->num_nodes*sizeof(DATA_DIR)))==NULL )
    {

#ifdef DEBUG
      printf("in dmbigger: malloc of %d new_node nodes failed\n",
	     blk_dir->num_nodes);
#endif

      return FALSE;
    }

  blk_dir->next_dir=blk_dir->block_array[ blk_dir->num_blocks-1 ];
  blk_dir->next_dir->block=blk_dir;
  blk_dir->last_dir=blk_dir->next_dir+blk_dir->num_nodes-1;

#ifdef DEBUG
  printf("in dmbigger2: dir = %lx blk_dir = %lx ", dir,blk_dir);
  printf("block_array = %lx num_blocks = %d\n",
	 blk_dir->block_array,blk_dir->num_blocks);
#endif

  return TRUE;
}


INT 
dtag_found(DATA_DIR *dir, unsigned fldid, UCHR Class)
{
  DATA_DIR *child;

  if( !dir || !(child=dir->ptr.child) )
    return FALSE;

  for(child=dir->ptr.child; child; child=child->next)
    if(child->fldid==fldid && child->Class==Class)
      return TRUE;
  return FALSE;
}


INT4 
dgetnum(DATA_DIR *dir)
{
  INT4     count;
  UINT4    i=0;
  int      longlen = sizeof(LONG);

  // make sure that input data is not bigger than a 'long'
  if (dir->count > longlen)
    return i;

  if(dir->ptr.data[ 0 ]>127)
    for(count=0; count<longlen-dir->count; count++)
      {
        i|=0xff;
        i<<=8;
      }
  for(count=0; count<dir->count; count++)
    {
      i|=dir->ptr.data[ count ];
      if(count<dir->count-1)
	i<<=8;
    }
  return (INT4)i;
}


INT4 
asn1len(UCHR *s)
{
  UCHR c;
  INT  len_len, tag_len=1;
  INT4 field_len;

  c=(UCHR)((*s++)&0x1f);        // the first byte of the tag

  if(c>=0x1f)                   // if the tag is greater than 31, then we
    {                           // keep looking at bytes until we find 1
      c= *s++;                  // with it's sign bit off
      tag_len++;
      while(c>0x80)
        {
	  c= *s++;
	  tag_len++;
        }
    }

  len_len=get_len(&field_len, s, 0x7fffffff);
  return (INT4)(tag_len + len_len + field_len);
}


void 
hex_dir(DATA_DIR *dir, INT level)
{
  hex_dirs(dir, level, cout);
}


void 
hex_dirs(DATA_DIR *dir, INT level, ostream &os)
{
  UCHR     *data;
  DATA_DIR *child;
  INT       i, linelen, tlen;

  static CHR *tabs=NULL;

  if(!tabs)
    {
      tabs=new CHR[201];
      memset(tabs, ' ', 200);
      tabs[200]='\0';
    }
  
  if(!dir)
    {
      os << "hex_dir passed a NULL directory pointer" << endl;
      return;
    }

  if(level>99)
    level=99;
  tabs[ level*2 ]='\0';
    //  tabs[ level*4 ]='\0';

  os << tabs << "tag=" << dir->fldid << ", Class=" << (int)dir->Class;
  os << ", form=" << (int)dir->form << ", count=" << dir->count << endl;

  tabs[ level*2 ]=' ';  // was tabs[ level*4 ]=' ';

  if(dir->form==ASN1_CONSTRUCTED)
    for(child=dir->ptr.child; child; child=child->next)
      hex_dirs(child, level+1, os);
  else
    {
      tabs[ (level+1)*2 ]='\0'; // was tabs[ (level+1)*4 ]='\0';
      
      if(dir->ptr.data)
        {
	  os << tabs << "data=";
	  tlen=(INT)dir->count;
	  data=dir->ptr.data;
	  linelen=79-9-(level*2); // was  linelen=79-strlen(tabs)-5;
            if(linelen<32)
                linelen=64;
	  while(tlen>0)
            {
	      for(i=0; i<tlen && i<linelen; i++)
		os << (isprint(data[i]) ? data[i] : '.');
	      os << endl;
	      os << tabs << "     " << hex;
	      for(i=0; i<tlen && i<linelen; i++)
		os << ((unsigned)data[i] / 16);
	      os << dec << endl;
	      os << tabs << "     " << hex;
	      for(i=0; i<tlen && i<linelen; i++)
		os << ((unsigned)data[i] % 16);
	      os << dec << endl;
	      tlen-=linelen;
	      data+=linelen;
	      if(tlen > 0)
		os << tabs << "     ";
            }
        }
      else
	os << tabs << "data=";
    }
  
  tabs[ (level+1)*2 ]=' ';  // was tabs[ (level+1)*4 ]=' ';
}


INT4 
rec_len(DATA_DIR *dir)
{
  DATA_DIR *child;

  if(!dir)
    return 0;
  dir->length=0;
  if(dir->form==ASN1_CONSTRUCTED)
    {
      for(child=dir->ptr.child; child; child=child->next)
	dir->length+=rec_len(child);

#ifdef DEBUG
      cerr << "in rec_len: total length for field " << dir->fldid
	   << " is " << dir->length << endl;
#endif

    }
  else
    {
      dir->length=dir->count;

#ifdef DEBUG
      printf("in rec_len: length of field %d is %ld\n",
	     dir->fldid, dir->length);
#endif

    }

  return dir->length + LEN_LEN(dir) + TAG_LEN(dir);
}


INT 
IsCompleteBER(UCHR *ptr, INT4 len, INT4 *remainder)
{
  INT  lenlen, tag, taglen;
  INT4 fieldlen, headerlen;

  *remainder=0;
  if(!len)
    return FALSE;

  if( (taglen=get_tag(&tag, ptr, len))==0)  // no tag yet
    return FALSE;

  if( (lenlen=get_len(&fieldlen, ptr+taglen, len-taglen))==0)
    return FALSE;  // no len yet

  headerlen=taglen+lenlen;
  if(lenlen==1 && fieldlen == -1)  // indefinite length
    {
      INT4 fieldlen, totlen=0;

      // loop through the subfields and see if they are complete
      for(ptr+=headerlen, len-=headerlen; len>1 
	    && (ptr[0]!=0 || ptr[1]!=0);
	  ptr+=fieldlen, len-=fieldlen)
        {
	  if(!IsCompleteBER(ptr, len, &fieldlen))
	    return FALSE;
	  totlen+=fieldlen;
        }
      if(len>1 && ptr[0]==0 && ptr[1]==0)
        {
	  *remainder=headerlen+totlen+2;  // + 2 nulls at end
	  return TRUE;
        }
      *remainder = -1;  // special flag to indicate indefinite length
                        // items
      return FALSE;
    }

  if(fieldlen+headerlen<=len)
    {
      *remainder=fieldlen+headerlen;
      return TRUE;
    }
  *remainder=fieldlen+headerlen-len;
  return FALSE;
}


int 
IsIndefiniteLengthBer(UCHR *ptr, LONG len)
{
  INT  lenlen, tag, taglen;
  INT4 fieldlen;
 
  if(!len)
    return TRUE;
 
  if( (taglen=get_tag(&tag, ptr, len))==0)  // no tag yet
    return TRUE;

  if( (lenlen=get_len(&fieldlen, ptr+taglen, len-taglen))==0)
    return TRUE;  // no len yet
 
  if(lenlen==1 && fieldlen == -1)  // indefinite length
    return TRUE;

  return FALSE;
}
 

INT 
get_tag(INT *tag, UCHR *s, INT4 len)
{
  UCHR     c;
  INT      taglen;

  if(!len)                    /* nothing to look at */
    {
      *tag = -1;
      return 0;
    }

  c=(*s++)&0x1f;              /* the first byte of the tag */
  taglen=1;

  if(c<0x1f)                  /* if the tag is less than 31, then it */
    *tag=c;                   /* is fully contained in the one byte */

  else                        /* otherwise we keep looking at bytes */
    {                         /* until we find 1 with it's sign bit off */
                              /* catenating the last 7 bits of each byte */
      if(len==1)              /* no extra bytes to look at */
        {
	  *tag = -1;
	  return 0;
        }
      *tag=0;
      c= *s++;
      taglen+=1;
      while(c>0x80 && taglen<len)
        {
	  *tag+=c&0x7F;
	  *tag<<=7;
	  c= *s++;
	  taglen+=1;
        }
      if(c>0x80 && taglen==len)  /* missing part of tag */
        {
	  *tag = -1;
	  return 0;
        }
      *tag+=c;
    }

  return taglen;
}


INT 
get_len(INT4 *fieldlen, UCHR *s, INT4 len)
{
  INT  i;
  UCHR c;
  INT4 tlen;

  if(!len)
    {
      *fieldlen = -1;
      return 0;
    }
  c= *s++;

  /* if the sign bit is turned on in the first byte of the length field, then
   * the first byte contains the number of subsequent bytes necessary to 
   * contain the length, or the length is indefinite if the remaining bits 
   * are zero.  Otherwise, the first byte contains the length */

  if(c<128)  // sign bit off
    {
      *fieldlen=c;
      return 1;
    }

  tlen=0;
  if( (INT)c-128>4)  // paranoia check: no lengths greater than 2 billion
    {
      *fieldlen = -1;
      return 0;
    }

  // Make sure we got enough length bytes.
  if ( (int)(c-127) > len)
    {
      *fieldlen = -1;
      return 0;
    }
  
  if(c==128)  // indefinite length
    {
      *fieldlen = -1;  // who knows
      return 1;        // the length of the length field IS 1
    }
  /*
  if(c-127>len)  // not enough length
    {
      *fieldlen = -1;
      return 0;
    }
    */
  for(i=0; i < (INT)c-128; i++)
    {
      tlen<<=8;
      tlen+= *s++;
    }

  *fieldlen=tlen;

#ifdef DEBUG
  printf("in get_len: ASN.1 length field is %d bytes long, *len=%ld\n",
	 (INT)c-127, *fieldlen);
#endif

  return (INT)c-127;
}


DATA_DIR *
daddbits(DATA_DIR *dir, unsigned fldid, UCHR Class, UCHR *bits)
{
  UCHR *data, *t;
  INT   i, j, len, num_bytes, unused;

  static UCHR mask[8]={0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

  len=strlen((CHR*)bits);
  num_bytes=1+len/8+((len%8)?1:0);
  t=data=dmalloc(dir, num_bytes);
  if(!data)
    return NULL_DIR;

  unused=8-len%8;
  if(unused==8)
    unused=0;

  *t=unused;
  for(i=0, t++; i<len; i+=8, t++)
    {
      *t='\0';
      for(j=0; j<8 && i+j<len; j++, bits++)
	if(*bits=='y' || *bits=='Y' || *bits=='t' || *bits=='T' ||
	   *bits=='1')
	  t[0]|=mask[j];
    }
  return daddchar(dir, fldid, Class, data, num_bytes);
}


UCHR *
dgetbits(DATA_DIR *dir)
{
  UCHR *bits;
  INT   j, last_used;
  INT4  i;

  static UCHR mask[8]={0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

  last_used=8 - dir->ptr.data[0];
  bits=(UCHR*)dmalloc(dir, 8*(INT)dir->count);
  for(i=1; i<dir->count-1; i++)
    for(j=0; j<8; j++)
      bits[(i-1)*8+j]=dir->ptr.data[i]&mask[j]?'y':'n';
  for(j=0; j<last_used; j++)
    bits[(i-1)*8+j]=dir->ptr.data[i]&mask[j]?'y':'n';
  bits[(i-1)*8+j]='\0';
  return bits;
}


UCHR *
dmalloc(DATA_DIR *dir, INT len)
{
  BLK_DIR  *blk_dir;
  DATA_DIR *t;

  if(!dir)
    return FALSE;
  blk_dir=dir->block;
  blk_dir->num_blocks++;

#ifdef DEBUG
  INT i;
  INT4 total=0;

  printf("in dmalloc1: dir = %lx blk_dir = %lx ", dir, blk_dir);
  printf("block_array = %lx num_blocks = %d\n",
	 blk_dir->block_array, (INT)blk_dir->num_blocks);
  for(i=0;i<blk_dir->num_blocks-1;i++)
    total += sizeof(*blk_dir->block_array[i]);
  printf("len=%i, total memory allocated = %d\n", len,total);
#endif

  if( (blk_dir->block_array=
       (DATA_DIR **)realloc((UCHR *)blk_dir->block_array,
			    blk_dir->num_blocks*sizeof(UCHR*)))==NULL )
    {

#ifdef DEBUG
      printf("in dmalloc: realloc of block_array failed\n");
#endif

      return NULL;
    }

  if( (t = blk_dir->block_array[ blk_dir->num_blocks-1 ] =
       (DATA_DIR *)malloc(len))==NULL )
    {

#ifdef DEBUG
      printf("in dmalloc: malloc of %d bytes\n", len);
#endif

      return NULL;
    }

#ifdef DEBUG
  printf("in dmalloc2: dir = %lx blk_dir = %lx ", dir, blk_dir);
  printf("block_array = %lx num_blocks = %d\n",
	 blk_dir->block_array, (INT)blk_dir->num_blocks);
#endif

  return (UCHR *)t;
}


DATA_DIR *
daddoid(DATA_DIR *dir, unsigned fldid, UCHR Class, CHR *cstring)
{
  UCHR place[100], *ptr;  /* seems like enough */
  INT  offset=0;
  INT4 value;

  while(cstring && isdigit(*cstring))
    {
      if(offset>90)
	return NULL_DIR;
      value = atol(cstring);
      if(offset==0)  /* first 2 numbers get special treatment */
        {
	  cstring=strchr(cstring, '.');
	  if(cstring)
	    cstring++;  /* skip '.' */
	  else
	    return NULL_DIR;  /* can't be this short */
	  value = value * 40 + atol(cstring);
        }

      if (value >= 0x80)
        {
	  INT  count = 0;
	  UCHR bits[12];    /* Save a 84 (12*7) bit number */

	  while (value)
            {
	      bits[count++] = (UCHR)value & 0x7F;
	      value >>= 7;
            }

	  /* Now place in the correct order */
	  while (--count > 0)
	    place[offset++] = bits[count] | 0x80;

	  place[offset++] = bits[count];
        }
      else
	place[offset++] = (UCHR)value;

      cstring=strchr(cstring, '.');
      if(cstring)
	cstring++;  /* skip '.' */
    }

  ptr=dmalloc(dir, offset);
  if(!ptr)
    return NULL_DIR;
  memcpy(ptr, place, offset);
  return daddchar(dir, fldid, Class, ptr, offset);
}


CHR *
dgetoid(DATA_DIR *dir)
{
  CHR  *s;
  UCHR *oid;
  INT   i, nvals=0, slen=0;
  INT4  len, offset=0, value[30];  /* 30 numbers should be enough */

  if(!dir || dir->form!=ASN1_PRIMITIVE)
    return NULL;
  oid=dir->ptr.data;
  len=dir->count;
  /* Convert encoded value to a c-string */
  while(offset<len)
    {
      value[nvals] = 0;
      do
        {
	  value[nvals] <<= 7;
	  value[nvals] |= oid[offset] & 0x7f;
        } while (oid[offset++] & 0x80);

      if(nvals == 0)
	slen=22;
      else
	slen+=11;
      nvals++;
    }
  s=(CHR *)dmalloc(dir, slen+1);
  if(!s)
    return NULL;

  ostrstream coid(s, slen + 1);
  
  coid << (value[0] / 40) << '.' << (value[0] % 40);
  for (i = 1; i < nvals; i++)
    coid << '.' << value[i];
  coid << '\0';

  return s;
}


DATA_DIR *
dinstag(DATA_DIR *dir, unsigned fldid, UCHR Class)
{
    DATA_DIR *new_node;
    //int           count;

    if(dir->block->next_dir > dir->block->last_dir)  /* out of new_node dirs */
        if(!dmake_bigger(dir))
            return FALSE;
    new_node=dir->block->next_dir++;
    if(dir->block->next_dir <= dir->block->last_dir)
        new_node->block->next_dir->block=dir->block;

    if(dir->parent)
    {
        new_node->ptr.child=dir->parent->ptr.child;
        dir->parent->ptr.child=new_node;
        dir->parent->count=1;
    }
    else
        new_node->ptr.child=dir;
    new_node->parent=dir->parent;

    for(new_node->count=0, dir=new_node->ptr.child; dir; 
        new_node->count++, dir=dir->next)
        dir->parent=new_node;

    new_node->next = new_node->prev = NULL_DIR;
    new_node->user=0;

    new_node->fldid=fldid;
    new_node->Class=Class;
    new_node->form =ASN1_CONSTRUCTED;

    return new_node;
}


DATA_DIR *
dreplace_num(DATA_DIR *msg, INT4 number)
{
  UCHR     *t;
 
  if(msg->form!=ASN1_PRIMITIVE)
    return(0);
 
  t=(UCHR*)(&msg->number);
  if(berutil_ByteOrder == 1)
    msg->number=number;
  else {
    UCHR *ptr;
    ptr=(UCHR*)(&number);
    t[0]= ptr[3];
    t[1]= ptr[2];
    t[2]= ptr[1];
    t[3]= ptr[0];
  }
  /*
    #if (NATIVE_ORDER&PDP_ORDER)==PDP_ORDER
    UCHR *ptr;
    ptr=(UCHR*)(&number);
    t[0]= ptr[1];
    t[1]= ptr[0];
    t[2]= ptr[3];
    t[3]= ptr[2];
    #endif
  */
  msg->ptr.data=t;
  msg->count = 4;
  if(msg->ptr.data[0]=='\0')
    while(msg->count>1 && msg->ptr.data[0]=='\0' &&
	  msg->ptr.data[1]<128)
      {
	msg->ptr.data++;
	msg->count--;
      }
  else
    if(msg->ptr.data[0]==0xff)
      while(msg->count>1 && msg->ptr.data[0]==0xff &&
	    msg->ptr.data[1]>127)
	{
	  msg->ptr.data++;
	  msg->count--;
	}
 
  return(msg);
}
