/* $Id: berutil.hxx,v 1.4 1997/05/31 21:08:02 cnidr Exp $ */
/*
(c)1992 OCLC Online Computer Library Center, Inc., 6565 Frantz Road,
Dublin, Ohio 43017-0702.  OCLC is a registered trademark of OCLC Online
Computer Library Center, Inc.

NOTICE TO USERS: The BER Utilities ("Software") has been developed by
OCLC Online Computer Library Center, Inc.  Subject to the terms and
conditions set forth below, OCLC grants to user a perpetual,
non-exclusive, royalty-free license to use, reproduce, alter, modify,
and create derivative works from Software, and to sublicense Software
subject to the following terms and conditions:

SOFTWARE IS PROVIDED AS IS.  OCLC MAKES NO WARRANTIES, REPRESENTATIONS,
OR GUARANTEES WHETHER EXPRESS OR IMPLIED REGARDING SOFTWARE, ITS FITNESS
FOR ANY PARTICULAR PURPOSE, OR THE ACCURACY OF THE INFORMATION CONTAINED
THEREIN.

User agrees that OCLC shall have no liability to user arising therefrom,
regardless of the basis of the action, including liability for special,
consequential, exemplary, or incidental damages, including lost profits,
even if it has been advised of the possibility thereof.

User shall cause the copyright notice of OCLC to appear on all copies of
Software, including derivative works made therefrom.
*/

/****

NOTE: You will want to define your machine's integer byte order below,
if you are not defining it externally.  OCLC and Stanford use a file
named portinfo.h to define port specific values and include it before
including berutil.h.  If you do not define NATIVE_ORDER in such a way,
you will want to examine the definition of NATIVE_ORDER below.

You may also want to change the definition of COPY from memcpy to bcopy.

****/

#ifndef BERUTIL_HXX
#define BERUTIL_HXX

#include "gdt.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifndef NATIVE_ORDER
/* byteorder values */
#define HIGHTOLOW 1
#define LOWTOHIGH 2
#define PDP_ORDER 4

/* define native byte ordering here.  Options are HIGHTOLOW (most rational
   machines), LOWTOHIGH (ibm pc's) or PDP_ORDER */

#define NATIVE_ORDER HIGHTOLOW  /* most UNIX machines */
#endif

/* ASN.1 tag classes */
#define ASN1_UNIVERSAL   (UCHR)0
#define ASN1_APPLICATION (UCHR)1
#define ASN1_CONTEXT     (UCHR)2
#define ASN1_PRIVATE     (UCHR)3

/* ASN.1 tag forms */
#define ASN1_PRIMITIVE   (UCHR)0
#define ASN1_CONSTRUCTED (UCHR)1

/* ASN.1 UNIVERSAL data types */
#define ASN1_BOOLEAN           1
#define ASN1_INTEGER           2
#define ASN1_BITSTRING         3
#define ASN1_OCTETSTRING       4
#define ASN1_NULL              5
#define ASN1_OBJECTIDENTIFIER  6
#define ASN1_OBJECTDESCRIPTOR  7
#define ASN1_EXTERNAL          8
#define ASN1_SEQUENCE         16
#define ASN1_SET              17
#define ASN1_VISIBLESTRING    26
#define ASN1_GENERALSTRING    27

/* ASN.1 EXTERNAL encoding choices */
#define ASN1_single_ASN1_type  0
#define ASN1_octet_aligned     1
#define ASN1_arbitrary         2

#define NULL_DIR  ((DATA_DIR  *)NULL)

#define TAG_LEN(dir) ((dir->fldid<31) ? 1 : ((dir->fldid<128) ? 2 : 3))
#define LEN_LEN(dir) (dir->length<128L ? 1L:(dir->length<256L ? 2L :\
(dir->length<65536L ? 3L : 4L)))

typedef struct blk_dir
{
  INT2              block_type;
  struct dir_node  *dir;
  struct dir_node  *next_dir;
  struct dir_node  *last_dir;
  INT               num_blocks;
  INT               num_nodes;
  struct dir_node **block_array;
} BLK_DIR;

typedef struct dir_node
{
  struct blk_dir   *block;  /* dir block containing directory */
  struct dir_node  *parent; /* the parent of this node */
  struct dir_node  *prev;   /* the previous sibling of this node */
  struct dir_node  *next;   /* the next sibling of this node */
  union
  {
    struct dir_node  *child;  /* first child, if constructed */
    UCHR *           data;    /* data, if primitive */
  } ptr;
  INT4 length;             /* if rebuilding: length of encoded data */
  INT4 count;              /* number of subfields, if constructed field
			      field length, if primitive field */
  INT4 user;               /* buffer for user assigned data */
  INT4 number;             /* buffer for holding numeric data */
  UINT2 fldid;             /* type of field */
  UCHR           Class;    /* ASN.1 class:
			      00 = universal
			      01 = application
			      10 = context specific
		 	      11 = private */
  UCHR           form;     /* ASN.1 primitive vs. constructed identifier
			      0 = primitive
			      1 = constructed */
} DATA_DIR;
typedef DATA_DIR  *PDATA_DIR;

#ifdef SHORT_NAMES
#define asm_rec       asmrec
#define bld_dir       blddir
#define bld_rec       bldrec
#define dmake_bigger  dmbigger
#define dreplace_num  drepnum
#define dtag_found    dtagfnd
#define get_len       getlen
#define get_num       getnum
#define get_tag       gettag
#define IsCompleteBER iscomber
#define hex_dir       hexdir
#define hex_dirs      hexdirs
#define rec_len       reclen
#endif

#ifndef NO_PROTOTYPES
extern INT4          asm_rec(DATA_DIR  *, UCHR  *);
extern INT4          asn1len(UCHR  *);
extern INT           bld_dir(UCHR  *, DATA_DIR *);
extern UCHR         *bld_rec(DATA_DIR *, INT4*);
extern DATA_DIR     *daddchar(DATA_DIR *, unsigned, UCHR, UCHR  *, INT4);
extern DATA_DIR     *dadddir(DATA_DIR *, DATA_DIR *);
extern DATA_DIR     *daddbits(DATA_DIR *, unsigned, UCHR, UCHR  *);
extern DATA_DIR     *daddnum(DATA_DIR *, unsigned, UCHR, UCHR  *, INT);
extern DATA_DIR     *daddoid(DATA_DIR *, unsigned, UCHR, CHR  *);
extern DATA_DIR     *daddtag(DATA_DIR *, unsigned, UCHR);
extern DATA_DIR     *dalloc(INT);
extern INT           ddeldir(DATA_DIR *);
extern INT           dfree(DATA_DIR *);
extern UCHR         *dgetbits(DATA_DIR *);
extern INT4          dgetnum(DATA_DIR *);
extern CHR          *dgetoid(DATA_DIR  *dir);
extern DATA_DIR     *dinit(DATA_DIR *, unsigned, UCHR);
extern DATA_DIR     *dinstag(DATA_DIR *, unsigned, UCHR);
extern DATA_DIR     *dmake(unsigned, UCHR, INT);
extern INT           dmake_bigger(DATA_DIR *);
extern UCHR         *dmalloc(DATA_DIR  *dir, INT len);
extern DATA_DIR     *dreplace_num(DATA_DIR *, INT4);
extern INT           dtag_found(DATA_DIR *, unsigned, UCHR);
extern INT           get_len(INT4* reclen, UCHR  *ptr, INT4 len);
extern INT4          get_num(UCHR *, INT4);
extern INT           get_tag(INT *tag, UCHR  *s, INT4 len);
extern INT           IsCompleteBER(UCHR  *ptr, INT4 len, INT4 *remainder);
extern INT           IsIndefiniteLengthBer(UCHR *ptr, INT4 len);
extern void          hex_dir(DATA_DIR *, INT);
extern void          hex_dirs(DATA_DIR*, INT,  ostream &os);
extern INT4          rec_len(DATA_DIR *);

#else
extern INT4          asm_rec();
extern INT4          asn1len();
extern INT           bld_dir();
extern UCHR         *bld_rec();
extern DATA_DIR     *daddchar();
extern DATA_DIR     *dadddir();
extern DATA_DIR     *daddnum();
extern DATA_DIR     *daddtag();
extern DATA_DIR     *dalloc();
extern INT           ddeldir();
extern INT           dfree();
extern INT4          dgetnum();
extern DATA_DIR     *dinit();
extern DATA_DIR     *dinstag();
extern DATA_DIR     *dmake();
extern INT           dmake_bigger();
extern DATA_DIR     *dreplace_num();
extern INT           dtag_found();
extern INT           get_len();
extern INT4          get_num();
extern INT           get_tag();
extern void          hex_dir();
extern void          hex_dirs();
extern INT4          rec_len();
#endif

#ifdef __cplusplus
}
#endif

#endif  /* #ifndef BERUTIL_HXX */

