/* $Id: gate.cxx,v 1.14 1998/01/30 19:10:57 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994.

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact.

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software.

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR.

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
************************************************************************/

#include <iostream.h>
#include <fstream.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>

#ifdef UNIX
#include <unistd.h>
#endif

#ifdef GW_PASSWD
#include "gdbm.h"
//#include "ndbm.h"
#endif

#include <string.h>

#include "ztags.hxx"
#include "defs.hxx"
#include "common.hxx"
#include "string.hxx"
#include "vlist.hxx"
#include "strlist.hxx"
#include "tokengen.hxx"
#include "strstack.hxx"
#include "infix2rpn.hxx"
#include "gdt.h"
#include "cgi.hxx"
#include "defines.hxx"
#include "registry.hxx"
#include "gate.hxx"
#include "zutils.hxx"

extern CHR *Host;
extern CHR *QuerySource;
extern CHR *DBInfo;
extern INT NumDatabases;
extern CHR *database_names[MAX_SERVERS], *database_values[MAX_SERVERS];
extern CHR *DBName, *RecSyntax, *ESName, *RSName, *MaxRecords;
extern CHR *AttrSet,*RecsPerServer;
extern volatile GDT_BOOLEAN WaitingForSignal;
extern enum { FROM_COMMANDLINE, FROM_CGI } Source;
extern CHR MapString[1024];
extern CHR min_lat[80], max_lat[80], min_lon[80], max_lon[80];
extern ofstream LogFile;
extern char* IsiteVersion;

static CHR *SPATH;
static CHR LocalTempFile[512];
static INT LocalBreakLoop,LocalError;
static CHR LocalAction[256];
static INT LocalPID;

#define TIMEOUT_LIMIT 10
#define MAXYEAR "9998"
#define MINYEAR "1000"

#ifndef __SUNPRO_CC
void 
ZgateAlarmHandler(INT t)
#else
void 
ZgateAlarmHandler(INT t,...)
#endif
{
#ifndef DEBUG
  LogFile << "Deleting " << LocalTempFile << endl;
  unlink(LocalTempFile);
#endif

  fclose(stdout);
  fclose(stdin);
  fclose(stderr);
  LogFile << "Exit via ZgateAlarmHandler" << endl;
  LogFile.flush();
  LogFile.close();
  exit(0);
}


#ifndef __SUNPRO_CC
void 
ZgateStopHandler(INT t)
#else
void 
ZgateStopHandler(INT t,...)
#endif
{
  cout << "<p><b>Stop Button Pressed</b>" << endl;
  cout.flush();
  if (LocalPID>0)
    kill(LocalPID, SIGUSR2);
  ZgateAlarmHandler(t);
}


#ifndef __SUNPRO_CC
void 
SegvHandler(INT t)
#else
void 
SegvHandler(INT t,...)
#endif
{

  cout << "<H1>Help - received SEGV signal</H1>" << endl;
  cout.flush();

  // Kill off zcon - it's done
  if (LocalPID>0)
    kill(LocalPID, SIGUSR2);
  ZgateAlarmHandler(t);
}

/*
INT 
DoWait()
{
  INT ch;
  LocalBreakLoop=LocalError=0;
  for(;;){
    pause();
    
    if(LocalError==1){
      return(0);
    }
    if(LocalBreakLoop==1) 
      break;
    
  }
  return(1);
}
*/


//---------------------------------------------------------------//
// DoWait	                                                 //
//                                                               //
// This routine forces synchronization between zgate and zcon    //
// by blocking zgate until either an error condition is detected //
// or zcon has completed the previous processing request.        //
//                                                               //
// This routine should be called immediately after calling       //
// PingManager.                                                  //
//                                                               //
// Note:                                                         //
// LocalBreakLoop and LocalError are semaphores that are         //
// locked (set to 0) by PingManager, and released (set to 1)     //
// by USR1Handler. If these semaphores have already been         // 
// released, then this process is already synchronized with zcon //
// (or an error has already occurred), and the pause is ignored. //
//                                                               //
//---------------------------------------------------------------//

INT 
DoWait()
{
  for (;;)
  {

    // if semaphores have already been released by USR1, don't pause

    if ((LocalBreakLoop==0) && (LocalError==0))   
        pause();
    
    if(LocalError==1) 
        return(0);

    if(LocalBreakLoop==1)  
        break;
  }

  return(1);
}

// Called when zcon signals that data is ready
// reads from the zcon.pid file and writes contents out to the web server
// if the requested function was INIT or LOGOUT, display
// special HTML forms
// INIT displays the search selection form
// Logout displays the login gateway page

#ifndef __SUNPRO_CC
void   
USR1Handler(int t)
#else
void   
USR1Handler(int t,...)
#endif
{
  
  /* Read results from TempFile */
  FILE           *fp,*fx;
  INT i=0;

  CHR buf[2024];
  CHR MapString[1024];
  CHR min_lat[80], max_lat[80], min_lon[80], max_lon[80];

  pid_t PID = getpid();
  LogFile << "[" << PID << "] ";
  LogFile << "Got USR1 signal from zcon" << endl;
  LogFile.flush();

  strcpy(MapString,"world-300x300.gif");
  strcpy(min_lat,"-90");
  strcpy(max_lat,"90");
  strcpy(min_lon,"-180");
  strcpy(max_lon,"180");

  if (StrCaseCmp(LocalAction,"INIT") 
        && StrCaseCmp(LocalAction,"LOGOUT")) {
    //  if (StrCaseCmp(LocalAction,"INIT")) {
    while((fp = fopen(LocalTempFile,"r"))==NULL) {
      ++i;
      sleep(1);
      if(i>TIMEOUT_LIMIT){
	LocalError=1;
	return;
      }
    }
    rewind(fp);
    while (fgets(buf, sizeof(buf), fp)) {
      cout << buf;
      cout.flush();
      if (!StrCaseCmp(buf, "SUCCESS\n")) {
	// Display search form
	fgets(buf, sizeof(buf), fp);
	PID = atoi(buf);

	if (!StrCaseCmp(QuerySource,"HTML")) 
	  DisplaySearchForm(DBInfo, MapString, min_lat, max_lat, min_lon,
			    max_lon, NumDatabases, database_names, 
			    database_values, PID, (PCHR)NULL);
	else
	  cout << "PID=" << PID << endl;
      }
    }
    fclose(fp);

#ifndef DEBUG
    LogFile << "Deleting " << LocalTempFile << endl;
    unlink(LocalTempFile);
#endif

    LogFile << "Exit via USR1Handler" << endl;
    exit(0);    
  } 
  LocalError=0;
  LocalBreakLoop=1;
}


GATE::GATE(char *mp,CGI *xcgi)
{
  STRING Filename, Group;
  STRLIST ResultList, Position, Value;
  INT count;
  struct stat info;

  cgi=xcgi;
  strcpy(mypath,mp);
  BreakLoop=0;

  pid_t PID = getpid();

#ifdef DEBUG
  LogFile << "[" << PID << "] ";
  LogFile << "Entered GATE constructor" << endl;
  LogFile.flush();
#endif

  signal(SIGUSR1,USR1Handler);
  signal(SIGINT,ZgateStopHandler); // when stop button pressed

#ifdef DEBUG
  LogFile << "[" << PID << "] ";
  LogFile << "Signals set." << endl;
  LogFile.flush();
#endif

  Defaults = new REGISTRY("zgate");

  Filename = ZGATE_INI_NAME;
  if (IsFile(Filename)) {
    LogFile << "[" << PID << "] ";
    LogFile << "Reading from " << ZGATE_INI_NAME << endl;
    LogFile.flush();
    Defaults->ProfileLoadFromFile(Filename, Position);
  } else {
    // Didn't find the compiled-in name - let's see if we can guess the name
    Filename = mypath;
    Filename.Cat("/gateway.ini");
    LogFile << "[" << PID << "] ";
    LogFile << ZGATE_INI_NAME << " not found.  Reading from " << endl;
    LogFile << Filename << " instead." << endl;
    LogFile.flush();
    if (IsFile(Filename))
      Defaults->ProfileLoadFromFile(Filename, Position);
  }

#ifdef DEBUG
  LogFile << "[" << PID << "] ";
  LogFile << "Profile loaded from file " << Filename << endl;
  LogFile.flush();
#endif

  Position.AddEntry("sites");
  Position.AddEntry("location");
  Defaults->GetData(Position,&DatabaseList);

#ifdef DEBUG
  LogFile << "[" << PID << "] ";
  LogFile << "Got location from [sites] block" << endl;
  LogFile.flush();
#endif

  STRING Section,Entry,Hold;

  Section="config";
  /*
    Entry="WEBFORMS";
    Defaults->ProfileGetString(Section, Entry, "/usr/local/etc/httpd/htdocs", 
    &Hold);
    WebForms = Hold.NewCString();
  */

  Entry="SCRATCH_DIR";
  //  Defaults->ProfileGetString(Section, Entry, "/tmp", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_TMP_DIR, &Hold);
  ScratchDir = Hold.NewCString();

  Entry="SPATH";
  //  Defaults->ProfileGetString(Section, Entry, "/tmp/queries", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_SPATH, &Hold);
  SPATH = Hold.NewCString();

  if(stat(SPATH,&info) != 0) {
    // Directory doesn't exist - make it
    if (mkdir(SPATH, S_IROTH | S_IRUSR | S_IRGRP | S_IWUSR | S_IWGRP | S_IXUSR | S_IXGRP) != 0) {
      // The mkdir failed, so just use /tmp instead
      strcpy(SPATH,GATEWAY_TMP_DIR);
    }
  }
  if (S_ISDIR(info.st_mode) == 0) {
    // It exists, but it isn't a directory so just use /tmp instead
    strcpy(SPATH, GATEWAY_TMP_DIR);
  }

  Entry="TIMEOUT";
  Defaults->ProfileGetString(Section, Entry, DEFAULT_TIMEOUT, &Hold);
  Timeout = Hold.GetInt();
  signal(SIGALRM,ZgateAlarmHandler);
  alarm(Timeout);

  Entry="TRACKER_PATH";
  Defaults->ProfileGetString(Section, Entry, ScratchDir, &Hold);
  ZgateTrackerPath = Hold.NewCString();

  Entry="TRACKER_NAME";
  //Defaults->ProfileGetString(Section, Entry, "recent", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_TRACKER, &Hold);
  ZgateTrackerName = Hold.NewCString();

  Entry="HISTORY_PATH";
  Defaults->ProfileGetString(Section, Entry, ScratchDir, &Hold);
  ZconHistoryPath = Hold.NewCString();

  Entry="HISTORY_NAME";
  //  Defaults->ProfileGetString(Section, Entry, "hx", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_HISTORY, &Hold);
  ZconHistoryName = Hold.NewCString();

  Entry="LOCK_PATH";
  Defaults->ProfileGetString(Section, Entry, ScratchDir, &Hold);
  ZconLockPath = Hold.NewCString();

  Entry="LOCK_NAME";
  //Defaults->ProfileGetString(Section, Entry, "lock", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_LOCK, &Hold);
  ZconLockName = Hold.NewCString();

  Entry="PASSWD_DIR";
  Defaults->ProfileGetString(Section, Entry, PASSWD_DIRECTORY, &Hold);
  PasswdDir = Hold.NewCString();

  // zgate configuration info
  Section="gateway";

  Entry="GATEFORM_PATH";
  //Defaults->ProfileGetString(Section, Entry, "/usr/local/etc/httpd/htdocs", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_HTDOCS, &Hold);
  ZgateGatewayPath = Hold.NewCString();

  Entry="GATEFORM_HTML";
  //Defaults->ProfileGetString(Section, Entry, "gateway.html", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_FORM, &Hold);
  ZgateGatewayHTML = Hold.NewCString();

  Entry="GATEFORM_URL";
  //Defaults->ProfileGetString(Section, Entry, "http://localhost/gateway.html", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_LINK, &Hold);
  ZgateGatewayURL = Hold.NewCString();

  Entry="PROG_PATH";
  //Defaults->ProfileGetString(Section, Entry, "/usr/local/etc/httpd/cgi-bin", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_CGIDIR, &Hold);
  ZgateFilePath = Hold.NewCString();

  Entry="PROG_NAME";
  //Defaults->ProfileGetString(Section, Entry, "zgate", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_ZGATE, &Hold);
  ZgateFileName = Hold.NewCString();

  Entry="PROG_URL";
  //Defaults->ProfileGetString(Section, Entry, "http://localhost/cgi-bin/zgate", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_GATELINK, &Hold);
  ZgateFileURL = Hold.NewCString();


  // zcon configuration info
  Section="connection";
  Entry="FILE_PATH";
  Defaults->ProfileGetString(Section, Entry, ScratchDir, &Hold);
  ZconFilePath = Hold.NewCString();

  Entry="FILE_NAME";
  //Defaults->ProfileGetString(Section, Entry, "zcon", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_ZCON, &Hold);
  ZconFileName = Hold.NewCString();

  Entry="PROG_PATH";
  //Defaults->ProfileGetString(Section, Entry, "/usr/local/etc/httpd/cgi-bin", &Hold);
  Defaults->ProfileGetString(Section, Entry, CGIBIN, &Hold);
  ZconProgName = Hold.NewCString();

  Entry="PROG_NAME";
  //Defaults->ProfileGetString(Section, Entry, "zcon", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_ZCON, &Hold);
  ZconProgName = Hold.NewCString();

#ifdef DEBUG
  LogFile << "[" << PID << "] ";
  LogFile << "Leaving GATE constructor" << endl;
  LogFile.flush();
#endif

}


GATE::~GATE()
{
  //  delete WebForms;
  delete ScratchDir;
  delete ZgateGatewayPath;
  delete ZgateGatewayHTML;
  delete ZgateGatewayURL;
  delete ZgateFilePath;
  delete ZgateFileName;
  delete ZgateFileURL;
  delete ZgateTrackerPath;
  delete ZgateTrackerName;
  delete ZconFilePath;
  delete ZconFileName;
  delete ZconHistoryPath;
  delete ZconHistoryName;
  delete ZconLockPath;
  delete ZconLockName;
  delete ZconProgPath;
  delete ZconProgName;
}


CHR* 
GATE::SNumber(STRING ETerm,char *GlobalUserId)
{
  CHR *p;
  CHR STerm[256],FN[256],Sbuf[256];
  FILE *fu; 
  INT pos;
  INFIX2RPN qp;
  
  STRING In,Out;
  // check for #S
  // if found, look up the term and replace it.
  //
  STRING j;
  INT val;
  if((pos=ETerm.Search("#S")) || (pos=ETerm.Search("#s"))){
    CHR c;
    for(;;){
      c=ETerm.GetChr(pos);
      if(c==']'){
	j.Cat(c);
	break;
      }
      j.Cat(c);
      pos++;
    }
    
    j.GetCString(STerm,256);
    
    p=strchr(STerm,'[');
    if(p)
      *p='\0';
    
    val=atoi(&STerm[2]);	
    
    sprintf(FN,"%s/%s/%05d",SPATH,GlobalUserId,val);
    
    fu=fopen(FN,"r");
    if(fu==NULL)
      return(NULL);
    fgets(Sbuf,256,fu);		// discard dbname
    fgets(Sbuf,256,fu);
    fclose(fu);
    In=Sbuf;
    //    qp.Parse(&In, &Out);
    qp.Parse(In, &Out);
    Out.GetCString(Sbuf,256);
    j.GetCString(STerm,256);
    ETerm.Replace(STerm,Sbuf);
    
    p=SNumber(ETerm,GlobalUserId);
    if(p==NULL)
      p=ETerm.NewCString();
    return(p); 
    
  }else{
    return(ETerm.NewCString());
  }
}


void 
GATE::SetErrorMessage(CHR *s) {
  pid_t ThisPID=getpid();

  strcpy(LocalErrorMessage,s);
  LogFile << "[" << ThisPID << "] ";
  LogFile << LocalErrorMessage << endl;
  LogFile.flush();
}


void 
GATE::GetErrorMessage(CHR *s)
{
  strcpy(s,LocalErrorMessage);
}


void 
//GATE::DisplayGWForm(CHR* Filename, INT SessionID)
GATE::DisplayGWForm()
{
  CHR    Name[256];
  FILE  *fp;

  pid_t  PID = getpid();

#ifdef DEBUG
  cout << "WWW dir: " << ZgateGatewayPath << "<BR>" << endl;
  cout << "Form: " << ZgateGatewayHTML << "<BR>" << endl;
  exit(1);
#endif

  sprintf(Name, "%s/%s",ZgateGatewayPath,ZgateGatewayHTML);

  LogFile << "[" << PID << "] ";
  LogFile << "Displaying Gateway form " << Name << endl;
  LogFile.flush();

  if ((fp = fopen(Name, "r")) == NULL) {
    cout << "<HTML><HEAD>\n<TITLE>Successful Logout</TITLE>\n</HEAD>" << endl;
    cout << "<BODY>\n<h1>Logout Completed</h1>" << endl;
    cout << "Successful Logout Completed<BR><HR>" << endl;
    cout << "<A HREF=\"http://www.cnidr.org\">CNIDR's</A> Stateful Z39.50 Gateway" << endl;
    cout << "</BODY></HTML>" << endl;
    return;
  } else {
    fclose(fp);
    STRING HTMLgate;
    HTMLgate.ReadFile(Name);
    cout << HTMLgate << endl;
  }
}


INT 
GATE::UpdateSNumberCount()
{
  CHR             UserDir[256];
  INT             Value=0;
  CHR             buf[256];
  FILE           *fu;
  
  sprintf(UserDir, "%s/%s/Count", SPATH,InitUserId);
  fu = fopen(UserDir, "r");
  if (fu) {
    fgets(buf, 256, fu);
    fclose(fu);
    Value = atoi(buf);
    fu = fopen(UserDir, "w");
    if(fu) {
      fprintf(fu, "%d\n", Value + 1);
      fclose(fu);
    }
  }
  return (Value);
}


INT 
GATE::CheckSession()
{
  CHR *Tmp;
  if ((Tmp = cgi_GetValueByName(cgi, "SESSION_ID")) == NULL) {
    Tmp = (CHR *)malloc(3);
    strcpy(Tmp, "-1");
  }
/*
  if (*Tmp == '\0') {
    //   cout << "Your form didn't include a SESSION_ID!";
    //   cout << endl;
    //  MyExit(1);
  }
*/
  LocalPID=atoi(Tmp);
  return(LocalPID);
}


INT 
GATE::GenerateSearchRequest(INT nDBs)
{
  CHR *DateBool, *GeoBool, *BooleanOp;
  CHR *Coords, *North, *South, *East, *West;
  CHR *TempCoords, *Space;
  CHR *Synonym;
  CHR *Tmp;
  
  CHR *Term1, *Use1, *Struct1, *Relation1, *BooleanOp1;
  CHR *Term2, *Use2, *Struct2, *Relation2, *BooleanOp2;
  CHR *Term3, *Use3, *Struct3, *Relation3;

  CHR *NumTerm1, *NumUse1, *NumStruct1, *NumRelation1, *NumBooleanOp;
  CHR *NumTerm2, *NumUse2, *NumStruct2, *NumRelation2;
  
  CHR *AddDate, *DatePresent;
  CHR *DateYear, *DateMonth, *DateDay;
  CHR *DateUse1, *DateRel1, *DatePresent1, *DateStruct1, *PDays1;
  
  CHR *DateUse2, *DateRel2, *DatePresent2, *DateStruct2;
  CHR *MinMonth, *MinDay, *MinYear, *MaxMonth, *MaxDay, *MaxYear;
  CHR             tf[256];
  FILE           *fq;
  
  pid_t           PID = CheckSession();
  pid_t           MyPID = getpid();
  CHR            *dat;
  INT             Start;
  GDT_BOOLEAN     SingleDB;

  TempFile = new CHR[1024];
  sprintf(TempFile, "%s/%s.%d", ScratchDir,ZconFileName,PID);
  strcpy(LocalTempFile,TempFile);

#ifdef DEBUG
  cgi_DisplayContents(cgi);
#endif

  LogFile << "[" << MyPID << "] ";
  LogFile << "TempFile name = " << TempFile << endl;
  LogFile.flush();

  if ((DBName = cgi_GetValueByName(cgi, "DBNAME")) == NULL) {
    DBName = new CHR[11];
    strcpy(DBName, "xxdefault");
    SingleDB = GDT_FALSE;
    LogFile << "[" << MyPID << "] ";
    LogFile << "DBNAME not found in form. Loading list from INI file." << endl;
    LogFile.flush();
  } else
    SingleDB=GDT_TRUE;

  LogFile << "[" << MyPID << "] ";
  LogFile << "DBName = " << DBName << endl;
  LogFile.flush();

#ifdef TRACK_QUERIES
  sprintf(tf, "%s/%s.%d", ZgateTrackerPath,ZgateTrackerName,PID);
  fq = fopen(tf, "w");
  if (fq==NULL) {
    SetErrorMessage("Couldn't open Query Page Tracker File");
    LogFile << "[" << MyPID << "] ";
    LogFile << "Couldn't open Query Page Tracker File " << tf << endl;
    LogFile.flush();
    //    return(0);
  }
  LogFile << "[" << MyPID << "] ";
  LogFile << "Query Page Tracker File " << tf << " opened" << endl;
  LogFile.flush();
#endif

  // General Search stuff
  if ((dat = cgi_GetValueByName(cgi, "START")) == NULL) {
    Start = 1;
  } else {
    Start = atoi(dat);
  }

  if((ESName = cgi_GetValueByName(cgi, "ESNAME")) == NULL) {
    ESName = new CHR[2];
    strcpy(ESName, "B");
  }

  if((RecSyntax = cgi_GetValueByName(cgi, "RECSYNTAX")) == NULL) {
    RecSyntax = new CHR[strlen(SUTRS_OID)+1];
    strcpy(RecSyntax, SUTRS_OID);
  }
  
  if((AttrSet = cgi_GetValueByName(cgi, "ATTRSET")) == NULL) {
    AttrSet = new CHR[strlen(BIB1_ATTRSET_OID)+1];
    strcpy(AttrSet,BIB1_ATTRSET_OID);
  }
  
  // Get the text search terms
  if((Term1 = cgi_GetValueByName(cgi, "TERM_1")) == NULL) {
    Term1 = (CHR *)malloc(1);
    strcpy(Term1, "");
  }
  if((Term2 = cgi_GetValueByName(cgi, "TERM_2")) == NULL) {
    Term2 = (CHR *)malloc(1);
    strcpy(Term2, "");
  }
  if((Term3 = cgi_GetValueByName(cgi, "TERM_3")) == NULL) {
    Term3 = (CHR *)malloc(1);
    strcpy(Term3, "");
  }
  
  // Text search attributes
  if((Use1 = cgi_GetValueByName(cgi, "USE_1")) == NULL) {
    Use1 = (CHR *)malloc(5);
    //    strcpy(Use1, "1035");
    strcpy(Use1, "1016");
  }
  if((Use2 = cgi_GetValueByName(cgi, "USE_2")) == NULL) {
    Use2 = (CHR *)malloc(5);
    //    strcpy(Use2, "1035");
    strcpy(Use2, "1016");
  }
  if((Use3 = cgi_GetValueByName(cgi, "USE_3")) == NULL) {
    Use3 = (CHR *)malloc(5);
    //    strcpy(Use3, "1035");
    strcpy(Use3, "1016");
  }

  GDT_BOOLEAN GotOldPage=GDT_TRUE;
  if((BooleanOp = cgi_GetValueByName(cgi, "BOOLEAN_OP"))==NULL) {
    BooleanOp = (CHR *)malloc(4);
    strcpy(BooleanOp, "AND");
    GotOldPage=GDT_FALSE;
  } 
  if((Struct1 = cgi_GetValueByName(cgi, "STRUCT_1"))==NULL) {
    Struct1 = (CHR *)malloc(10);
    strcpy(Struct1, "WORD_LIST");
  }
  if((Struct2 = cgi_GetValueByName(cgi, "STRUCT_2"))==NULL) {
    Struct2 = (CHR *)malloc(10);
    strcpy(Struct2, "WORD_LIST");
  }
  if((Struct3 = cgi_GetValueByName(cgi, "STRUCT_3"))==NULL) {
    Struct3 = (CHR *)malloc(10);
    strcpy(Struct3, "WORD_LIST");
  }
  if((Relation1 = cgi_GetValueByName(cgi, "REL_1"))==NULL) {
    Relation1 = (CHR *)malloc(6);
    strcpy(Relation1, "EQUAL");
  }
  if((Relation2 = cgi_GetValueByName(cgi, "REL_2"))==NULL) {
    Relation2 = (CHR *)malloc(6);
    strcpy(Relation2, "EQUAL");
  }
  if((Relation3 = cgi_GetValueByName(cgi, "REL_3"))==NULL) {
    Relation3 = (CHR *)malloc(6);
    strcpy(Relation3, "EQUAL");
  }
  
  // Text search Boolean operators - just stack them left to right
  if((BooleanOp1 = cgi_GetValueByName(cgi, "BOOLEAN_OP_1"))==NULL) {
    BooleanOp1 = (CHR *)malloc(4);
    strcpy(BooleanOp1, "AND");
  } 
  if((BooleanOp2 = cgi_GetValueByName(cgi, "BOOLEAN_OP_2"))==NULL) {
    BooleanOp2 = (CHR *)malloc(4);
    strcpy(BooleanOp2, "AND");
  } 
  
  // Eventually, we may be able to do synomym expansion
  if((Synonym = cgi_GetValueByName(cgi, "syn_list"))==NULL) {
    Synonym = (CHR *)malloc(5);
    strcpy(Synonym, "WORD");
  } 
  
  // retrieve values to build a numeric query
  if((NumTerm1 = cgi_GetValueByName(cgi, "NUM_TERM_1")) == NULL) {
    NumTerm1 = (CHR *)malloc(1);
    strcpy(NumTerm1, "");
  }
  if((NumTerm2 = cgi_GetValueByName(cgi, "NUM_TERM_2")) == NULL) {
    NumTerm2 = (CHR *)malloc(1);
    strcpy(NumTerm2, "");
  }
  
  // Numeric search attributes
  if((NumUse1 = cgi_GetValueByName(cgi, "NUM_USE_1")) == NULL) {
    NumUse1 = (CHR *)malloc(5);
    //    strcpy(NumUse1, "1035");
    strcpy(NumUse1, "1016");
  }
  if((NumUse2 = cgi_GetValueByName(cgi, "NUM_USE_2")) == NULL) {
    NumUse2 = (CHR *)malloc(5);
    //    strcpy(NumUse2, "1035");
    strcpy(NumUse2, "1016");
  }
  if((NumStruct1 = cgi_GetValueByName(cgi, "NUM_STRUCT_1"))==NULL) {
    NumStruct1 = (CHR *)malloc(10);
    strcpy(NumStruct1, "WORD_LIST");
  }
  if((NumStruct2 = cgi_GetValueByName(cgi, "NUM_STRUCT_2"))==NULL) {
    NumStruct2 = (CHR *)malloc(10);
    strcpy(NumStruct2, "WORD_LIST");
  }
  if((NumRelation1 = cgi_GetValueByName(cgi, "NUM_REL_1"))==NULL) {
    NumRelation1 = (CHR *)malloc(6);
    strcpy(NumRelation1, "EQUAL");
  }
  if((NumRelation2 = cgi_GetValueByName(cgi, "NUM_REL_2"))==NULL) {
    NumRelation2 = (CHR *)malloc(6);
    strcpy(NumRelation2, "EQUAL");
  }
  if((NumBooleanOp = cgi_GetValueByName(cgi, "NUM_BOOLEAN_OP"))==NULL) {
    NumBooleanOp = (CHR *)malloc(4);
    strcpy(NumBooleanOp, "AND");
  } 
  
  // Are we adding a date query?  What kind?
  if((AddDate = cgi_GetValueByName(cgi, "add-date"))==NULL) {
    AddDate = (CHR *)malloc(3);
    strcpy(AddDate, "NO");
  }
  
  // For single date, do we input a specific one, or search on "PRESENT"
  if((DatePresent1 = cgi_GetValueByName(cgi, "date_present_1"))==NULL) {
    DatePresent1 = (CHR *)malloc(5);
    strcpy(DatePresent1, "DATE");
  }
  if((PDays1 = cgi_GetValueByName(cgi, "relative-days"))==NULL) {
    PDays1 = (CHR *)malloc(3);
    strcpy(PDays1, "30");
  }
    
  // query terms/attributes for single date search
  if((DateUse1 = cgi_GetValueByName(cgi, "date_use_1"))==NULL) {
    DateUse1 = (CHR *)malloc(5);
    strcpy(DateUse1, "3103");
  }
  if((DateRel1 = cgi_GetValueByName(cgi, "date_rel_1"))==NULL) {
    DateRel1 = (CHR *)malloc(2);
    strcpy(DateRel1, "3");
  }
  if((DateStruct1 = cgi_GetValueByName(cgi, "date_struct_1"))==NULL) {
    DateStruct1 = (CHR *)malloc(5);
    strcpy(DateStruct1, "DATE");
  }
  if((DateYear = cgi_GetValueByName(cgi, "date_year"))==NULL) {
    DateYear = (CHR *)malloc(5);
    strcpy(DateYear, "1996");
  }
  if((DateMonth = cgi_GetValueByName(cgi, "date_month"))==NULL) {
    DateMonth = (CHR *)malloc(1);
    strcpy(DateMonth, "");
  }
  if((DateDay = cgi_GetValueByName(cgi, "date_day"))==NULL) {
    DateDay = (CHR *)malloc(1);
    strcpy(DateDay, "");
  }
    
  // For date range, do we input specific dates, or search on "PRESENT"
  if((DatePresent2 = cgi_GetValueByName(cgi, "date_present_2"))==NULL) {
    DatePresent2 = (CHR *)malloc(5);
    strcpy(DatePresent2, "DATE");
  }
    
  // construct Date Range Query USE Attr 3103
  if((DateUse2 = cgi_GetValueByName(cgi, "date_use_2"))==NULL) {
    DateUse2 = (CHR *)malloc(5);
    strcpy(DateUse2, "3103");
  }
  if((DateRel2 = cgi_GetValueByName(cgi, "date_rel_2"))==NULL) {
    DateRel2 = (CHR *)malloc(2);
    strcpy(DateRel2, "3");
  }
  if((DateStruct2 = cgi_GetValueByName(cgi, "date_struct_2"))==NULL) {
    DateStruct2 = (CHR *)malloc(5);
    strcpy(DateStruct2, "DATE");
  }
  if((MinMonth = cgi_GetValueByName(cgi, "begdate-month"))==NULL) {
    MinMonth = (CHR *)malloc(1);
    strcpy(MinMonth, "");
  }
  if((MinDay = cgi_GetValueByName(cgi, "begdate-day"))==NULL) {
    MinDay = (CHR *)malloc(1);
    strcpy(MinDay, "");
  }
  if((MinYear = cgi_GetValueByName(cgi, "begdate-year"))==NULL) {
    MinYear = (CHR *)malloc(10);
    strcpy(MinYear, MINYEAR);
  }
  
  if((MaxMonth = cgi_GetValueByName(cgi, "enddate-month"))==NULL) {
    MaxMonth = (CHR *)malloc(1);
    strcpy(MaxMonth, "");
  }
  if((MaxDay = cgi_GetValueByName(cgi, "enddate-day"))==NULL) {
    MaxDay = (CHR *)malloc(1);
    strcpy(MaxDay, "");
  }
  if((MaxYear = cgi_GetValueByName(cgi, "enddate-year"))==NULL) {
    MaxYear = (CHR *)malloc(10);
    strcpy(MaxYear, MAXYEAR);
  }
  
  /* construct Rectangle query N S E W 3111 - formatted */
  CHR Rectangle[2048],Dummy[2048];
  
  if ((Coords = cgi_GetValueByName(cgi, "coords"))!=NULL) {
    North = (CHR *)malloc(10);
    TempCoords=Coords;
    Space=strchr(TempCoords,' ');
    *Space='\0';

    strcpy(North, TempCoords);
    TempCoords=Space+1;
    South = (CHR *)malloc(10);
    Space=strchr(TempCoords,' ');
    *Space='\0';

    strcpy(South, TempCoords);
    TempCoords=Space+1;
    East = (CHR *)malloc(10);
    Space=strchr(TempCoords,' ');
    *Space='\0';

    strcpy(East, TempCoords);
    TempCoords=Space+1;
    West = (CHR *)malloc(10);
    strcpy(West, TempCoords);

  } else {
    if((North = cgi_GetValueByName(cgi, "NORTH"))==NULL) {
      North = (CHR *)malloc(10);
      strcpy(North, "90");
    }
    if((South= cgi_GetValueByName(cgi, "SOUTH"))==NULL) {
      South = (CHR *)malloc(10);
      strcpy(South, "-90");
    }
    if((East= cgi_GetValueByName(cgi, "EAST"))==NULL) {
      East = (CHR *)malloc(10);
      strcpy(East, "180");
    }
    if((West= cgi_GetValueByName(cgi, "WEST"))==NULL) {
      West = (CHR *)malloc(10);
      strcpy(West, "-180");
    }
  }
  
#ifdef DEBUG
  cout << "<BR>" << endl;
  cout << "NORTH = " << North << "<BR>" << endl;
  cout << "SOUTH = " << South << "<BR>" << endl;
  cout << "EAST = " << East << "<BR>" << endl;
  cout << "WEST = " << West << "<BR>" << endl;
#endif
  
  if( (Term1 != NULL) && (Term1[0] != '\0') ){
    strcpy(Dummy,Term1);
    free(Term1);
    Term1=BuildKWAQS(Dummy,Use1,Struct1,Relation1);
    free(Use1);
    free(Struct1);
    free(Relation1);
  }
  if( (Term2 != NULL) && (Term2[0] != '\0') ){
    strcpy(Dummy,Term2);
    free(Term2);
    Term2=BuildKWAQS(Dummy,Use2,Struct2,Relation2);
    free(Use2);
    free(Struct2);
    free(Relation2);
  }
  if( (Term3 != NULL) && (Term3[0] != '\0') ){
    strcpy(Dummy,Term3);
    free(Term3);
    Term3=BuildKWAQS(Dummy,Use3,Struct3,Relation3);
    free(Use3);
    free(Struct3);
    free(Relation3);
  }
  
#ifdef DEBUG
  cout << "<BR>" << endl;
  cout << "Term1KWAQS = " << Term1 << "<BR>" << endl;
  cout << "Term2KWAQS = " << Term2 << "<BR>" << endl;
  cout << "Term3KWAQS = " << Term2 << "<BR>" << endl;
#endif
  
  // At this point, the text terms have been converted to KWAQS
  // so they look like term[1,n,2,,3,n,4,n,5,n,6,n]
  // and are ready to be combined with the Booleans
  
  CHR *TheTextTerm=(CHR*)NULL;
  CHR *TheNumTerm=(CHR*)NULL;
  CHR *TheTerm=(CHR*)NULL;
  
  if( (Term1 != NULL) && (Term1[0] != '\0') ) {
    if ((TheTextTerm = (CHR *)malloc(strlen(Term1)+64))) {
      sprintf(TheTextTerm,"%s",Term1);
      free(Term1);
    } else {
      cout << "malloc for Term1 failed" << endl;
      MyExit(1);
    }
      
    if( (Term2 != NULL) && (Term2[0] != '\0') ) {
      // stash the current text query into Term1
      if ((Term1 = (CHR *)malloc(strlen(TheTextTerm)+64))) {
	strcpy(Term1, TheTextTerm);
	free(TheTextTerm);
      } else {
	cout << "temp malloc for Term2 failed" << endl;
	MyExit(1);
      }
      
      // and combine the existing text query with Term2
      if ((TheTextTerm = (CHR *)malloc(strlen(Term1)+strlen(Term2)+64))) {
	// Check to see if we got the Boolean operator from the old
	// search page - they use different form variables
	if (GotOldPage)       
	  sprintf(TheTextTerm, "%s(%s,%s)", BooleanOp, Term1, Term2);
	else
	  sprintf(TheTextTerm, "%s(%s,%s)", BooleanOp1, Term1, Term2);
	free(Term1);
	free(Term2);
      } else {
	cout << "malloc for Term2 failed" << endl;
	MyExit(1);
      }
    } 
    
    if( (Term3 != NULL) && (Term3[0] != '\0') ) {
      // stash the current text query into Term1
      if ((Term1 = (CHR *)malloc(strlen(TheTextTerm)+64))) {
	strcpy(Term1, TheTextTerm);
	free(TheTextTerm);
      } else {
	cout << "temp malloc for Term3 failed" << endl;
	MyExit(1);
      }
      
      // and combine the existing text query with Term3
      if ((TheTextTerm = (CHR *)malloc(strlen(Term1)+strlen(Term3)+64))) {
	sprintf(TheTextTerm, "%s(%s,%s)", BooleanOp2, Term1, Term3);
	free(Term1);
	free(Term3);
      } else {
	cout << "malloc for Term3 failed" << endl;
	MyExit(1);
      }
    } 
  }
  
#ifdef DEBUG
  cout << "TheTextTerm = " << TheTextTerm << "<BR>" << endl;
#endif
    
  // At this point, the text query should be fully formed
  // Now, preprocess the numeric stuff
  
  if( (NumTerm1 != NULL) && (NumTerm1[0] != '\0') ){
    strcpy(Dummy,NumTerm1);
    free(NumTerm1);
    NumTerm1=BuildKWAQS(Dummy,NumUse1,NumStruct1,NumRelation1);
    free(NumUse1);
    free(NumStruct1);
    free(NumRelation1);
  }
  if( (NumTerm2 != NULL) && (NumTerm2[0] != '\0') ){
    strcpy(Dummy,NumTerm2);
    free(NumTerm2);
    NumTerm2=BuildKWAQS(Dummy,NumUse2,NumStruct2,NumRelation2);
    free(NumUse2);
    free(NumStruct2);
    free(NumRelation2);
  }
  
#ifdef DEBUG
  cout << "NumTerm1KWAQS = " << NumTerm1 << "<BR>" << endl;
  cout << "NumTerm2KWAQS = " << NumTerm2 << "<BR>" << endl;
#endif
  
  // The numeric terms have been converted to KWAQS
  // so they look like term[1,n,2,n,3,n,4,n,5,n,6,n]
  // and are ready to be combined with the Booleans
  INT NTlen;
  NTlen = strlen(NumTerm1)+strlen(NumTerm2)+64;

  if ((TheNumTerm = (CHR *)malloc(NTlen))) {
    if ( ((NumTerm1 != NULL) && (NumTerm1[0] != '\0'))
	&& ((NumTerm2 != NULL) && (NumTerm2[0] != '\0')) ) {
	sprintf(TheNumTerm, "%s(%s,%s)", NumBooleanOp, NumTerm1, NumTerm2);
	free(NumTerm1);
	free(NumTerm2);

    } else if( (NumTerm1 != NULL) && (NumTerm1[0] != '\0') ) {
      sprintf(TheNumTerm,"%s",NumTerm1);
      free(NumTerm1);

    } else if( (NumTerm2 != NULL) && (NumTerm2[0] != '\0') ) {
      sprintf(TheNumTerm,"%s",NumTerm2);
      free(NumTerm1);

    } else {
      TheNumTerm = "";
    }

  } else {
    cout << "malloc for TheNumTerm failed" << endl;
    MyExit(1);
  }    

#ifdef DEBUG
  cout << "TheNumTerm = " << TheNumTerm << "<BR>" << endl;
#endif
  
  // So, TheTextTerm and TheNumTerm have been constructed
  // and are ready to be combined with the Booleans
  
  sprintf(Rectangle,"%s %s %s %s[1,2060]",North,South,East,West);
  
#ifdef DEBUG
  cout << "Rectangle = " << Rectangle << "<BR>" << endl;
#endif
  
  // Time to build the date query - lots of cases to handle
  CHR DateTerm[1024];
  CHR *TheDate,*MaxDate,*MinDate;
  INT Day;
  time_t today;
  struct tm    *t;
  
  if( (!StrCaseCmp(AddDate,"single")) ) {
#ifdef DEBUG
    cout << "AddDate=" << AddDate << "<BR>" << endl;
#endif
    // Search on single date
    if (strlen(DateDay) > 0)
      Day = atoi(DateDay);
    else
      Day = 0;
    TheDate=BuildFGDCDate(DateYear,DateMonth,Day);
    strcpy(Dummy,TheDate);
    free(TheDate);
    TheDate=BuildKWAQS(Dummy,DateUse1,DateStruct1,DateRel1);
    strcpy(DateTerm,TheDate);
    free(TheDate);
    free(DateUse1);
    free(DateStruct1);
    free(DateRel1);
    
  } else if ( (!StrCaseCmp(AddDate,"range")) ) {
#ifdef DEBUG
    cout << "AddDate=" << AddDate << "<BR>" << endl;
#endif
    
    // Search on range of dates
    if (strlen(MaxDay) > 0)
      Day = atoi(MaxDay);
    else
      Day = 0;
    MaxDate=BuildFGDCDate(MaxYear,MaxMonth,Day);
    
    if (strlen(MinDay) > 0)
      Day = atoi(MinDay);
    else
      Day = 0;
    MinDate=BuildFGDCDate(MinYear,MinMonth,Day);
    
    sprintf(Dummy,"%s %s",MinDate,MaxDate);
    TheDate=BuildKWAQS(Dummy,DateUse2,DateStruct2,DateRel2);
    strcpy(DateTerm,TheDate);
    free(TheDate);
    free(MaxDate);
    free(MinDate);
    
  } else if ( (!StrCaseCmp(AddDate,"relative")) ) {
#ifdef DEBUG
    cout << "AddDate=" << AddDate << "<BR>" << endl;
#endif
    time_t Ndays;
    
    // Use today's date for ending
    today = time((time_t *)NULL);
    t = localtime(&today);
    if ((MaxDate = (CHR *)malloc(9))) {
      strftime(MaxDate,9,"%Y%m%d",t); // and put today's date in MaxDate
    } else {
      cout << "temp malloc for MaxDate failed" << endl;
    }
    
    // Convert PDays1 to seconds and subtract from today's date
    Ndays = atoi(PDays1) * 86400;
    today -= Ndays;
    t = localtime(&today);
    if ((MinDate = (CHR *)malloc(9))) {
      strftime(MinDate,9,"%Y%m%d",t); // and put today's date in MinDate
    } else {
      cout << "temp malloc for MinDate failed" << endl;
    }
    sprintf(Dummy,"%s %s",MinDate,MaxDate);
    TheDate=BuildKWAQS(Dummy,DateUse2,DateStruct2,DateRel2);
    strcpy(DateTerm,TheDate);
    free(TheDate);
    free(MaxDate);
    free(MinDate);
    
  } else {
#ifdef DEBUG
    cout << "AddDate=" << AddDate << "<BR>" << endl;
#endif
    // no date search at all
    strcpy(DateTerm,"");
  }
  
#ifdef DEBUG
  cout << "DateTerm = " << DateTerm << "<BR>" << endl;
#endif
    
  /* at this point, we must AND the query types together to form
     one combined query */
  
  if ((TheTerm=(CHR *)malloc(2048)) != NULL) {
    strcpy(TheTerm,"");
  } else {
    cout << "malloc failed for TheTerm" << endl;
    MyExit(1);
  }
  
  GDT_BOOLEAN DoGeo=GDT_FALSE,DoDate=GDT_FALSE;
  
  GDT_BOOLEAN GotText,GotNum,GotDate,GotGeo;
  
  // Put the text part of the query into the main query
  if( (TheTextTerm != NULL) && (strlen(TheTextTerm) > 0) ) {
    strcpy(TheTerm,TheTextTerm);
    GotText = GDT_TRUE;
  } else
    GotText = GDT_FALSE;
  
#ifdef DEBUG
  cout << "TheTerm = " << TheTerm << "<BR>" << endl;
#endif

  CHR *AddProgress, *Progress;

  // Are we searching on Progress?
  if((AddProgress = cgi_GetValueByName(cgi, "add-progress"))==NULL) {
    AddProgress = (CHR *)malloc(3);
    strcpy(AddProgress, "NO");
  }
  if((Progress = cgi_GetValueByName(cgi, "progress"))==NULL) {
    Progress = (CHR *)malloc(1);
    strcpy(Progress, "");
  }

  GDT_BOOLEAN DoProgress;
  CHR *ProgressTerm=(CHR*)NULL;
  if( (!StrCaseCmp(AddProgress,"yes")) ) {
    CHR PrUse[]="3108";
    CHR PrStruct[]="WORD_LIST";
    CHR PrRelation[]="EQUAL";
    DoProgress = GDT_TRUE;
    ProgressTerm=BuildKWAQS(Progress,PrUse,PrStruct,PrRelation);
    free(Progress);
  } else
    DoProgress = GDT_FALSE;

#ifdef DEBUG
  cout << "ProgressTerm = " << ProgressTerm << "<BR>" << endl;
#endif

  // Attach ProgressTerm to TheTerm, if necessary
  if(DoProgress) {
    if (strlen(TheTerm) > 0) {
      strcpy(Dummy,TheTerm);
      sprintf(TheTerm,"AND(%s,%s)",Dummy,ProgressTerm);
      GotText = GDT_TRUE;
    } else {
      strcpy(TheTerm,ProgressTerm);
      GotText = GDT_TRUE;
    }
  }
  
#ifdef DEBUG
  cout << "TheTerm = " << TheTerm << "<BR>" << endl;
#endif
  
  // Put the numeric part of the query into the main query
  if( (TheNumTerm != NULL) && (strlen(TheNumTerm) > 0) ) {
    if (strlen(TheTerm) > 0) {
      strcpy(Dummy,TheTerm);
      sprintf(TheTerm,"AND(%s,%s)",TheTextTerm,TheNumTerm);
    } else {
      strcpy(TheTerm,TheNumTerm);
    }
    GotNum = GDT_TRUE;
  } else {
    GotNum = GDT_FALSE;
  }
  
#ifdef DEBUG
  cout << "TheTerm = " << TheTerm << "<BR>" << endl;
#endif
    
  // Put the date part of the query into the main query
  if( (DateTerm != NULL) && (strlen(DateTerm) > 0) ) {
    if(strlen(TheTerm) > 0) {
      strcpy(Dummy,TheTerm);
      sprintf(TheTerm,"AND(%s,%s)",Dummy,DateTerm);
    } else {
      strcpy(TheTerm,DateTerm);
    }
    GotDate = GDT_TRUE;
  } else {
    GotDate = GDT_FALSE;
  }
    
#ifdef DEBUG
  cout << "TheTerm = " << TheTerm << "<BR>" << endl;
#endif

  if((GeoBool = cgi_GetValueByName(cgi, "add-geo"))==NULL) {
    GeoBool = (CHR *)malloc(4);
    strcpy(GeoBool, "NO");
  } 
  
  if (!StrCaseCmp(GeoBool,"yes")) {
    if( (Rectangle != NULL) && (strlen(Rectangle) > 0) ) {
      if(strlen(TheTerm) > 0) {
	strcpy(Dummy,TheTerm);
	sprintf(TheTerm,"AND(%s,%s)",Dummy,Rectangle);
      } else {
	strcpy(TheTerm,Rectangle);
      }
      GotGeo = GDT_TRUE;
    } else {
      GotGeo = GDT_FALSE;
    }
    GotGeo = GDT_FALSE;
  }

#if defined(DEBUG) || defined (DEBUGRPN)
  cout << "TheTerm = " << TheTerm << "<BR>" << endl;
#endif

  LogFile << "[" << MyPID << "] ";
  LogFile << "TheTerm = " << TheTerm << endl;
  LogFile.flush();

  if (strlen(TheTerm) == 0) {
    cout << "You didn't include a search term!";
    cout << "Go back and enter something to search on" << endl;
    MyExit(1);
  }
  
  if((MaxRecords = cgi_GetValueByName(cgi, "MAXRECORDS"))==NULL) {
    MaxRecords = (CHR *)malloc(3);
    strcpy(MaxRecords, "10");
  }

#ifdef DEBUG
  cout << "MaxRecords= " << MaxRecords << "<BR>" << endl;
#endif

  if ((RecsPerServer = cgi_GetValueByName(cgi, "SERVERRECORDS"))==NULL) {
    RecsPerServer = (CHR *)malloc(3);
    strcpy(RecsPerServer, "-1");
  }

#ifdef DEBUG
  cout << "ServerRecords= " << RecsPerServer << "<BR>" << endl;
#endif

  if (!SingleDB) {
    CHR *DBType,*DBHold;
    if((DBType = cgi_GetValueByName(cgi, "DATABASE-TYPE")) == NULL) {
      DBType = new CHR[4];
      strcpy(DBType, "all");
    }

#ifdef DEBUG
    cout << "DBType = " << DBType << "<BR>" << endl;
#endif

    STRLIST Databases;
    STRING Database;
    INT i;

    LogFile << "[" << MyPID << "] ";
    LogFile << "Getting multiple databases" << endl;
    LogFile.flush();

    if (nDBs > 0) {
      if (StrCaseCmp(DBType,"all")) {
	i=0;
	while((DBHold = cgi_GetMultipleValueByName(cgi, "DATABASES",i)) != NULL) {
	  Database = DBHold;
	  i++;
	  Databases.AddEntry(Database);
	  free(DBHold);
	} 
      } else {

	//	for(i=0;i<nDBs;i++) {
	for(i=1;i<nDBs;i++) {
	  Database = database_names[i];
	  Databases.AddEntry(Database);
	}
      }
  
      Databases.Join("+",&Database);
      LogFile << "[" << MyPID << "] ";
      LogFile << "Databases = " << Database << endl;
      LogFile.flush();
      
      if (Database.GetLength() > 0) {
	DBName = Database.NewCString();
      }
    }
  } else {

#ifdef DEBUG
  cout << "Skipping database-type<BR>" << endl;
#endif

  }

  FILE *fp;
  
  // Make sure that another zgate is not still running by
  // looking for other zcon.<PID> files in /tmp
  //
  if ((fp = fopen(TempFile, "r")) != NULL) {
    fclose(fp);
    cout << "<PRE>The HTTP server is still terminating your last" << endl;
    cout << "stop request.  This may take a few seconds. Please " << endl;
    cout << "attempt the search again or <A HREF=\"" ;
    cout << ZgateFileURL;
    cout << "\">reconnect</A>." << endl;
    cout << "Please be patient.</PRE>\n";
    MyExit(0);

  } else {

    if ((fp = fopen(TempFile, "w")) == NULL) {
      SetErrorMessage("Couldn't Open Gateway Transfer File");
      LogFile << "[" << MyPID << "] ";
      LogFile << "Couldn't Open Gateway Transfer File " << TempFile << endl;
      LogFile.flush();
      return(0);
    }
    fprintf(fp, "%d\nSEARCH\n%s\n%s\n%s\n%s\n%s\n%s\n%d\n%s\n", getpid(),
	    DBName, TheTerm, ESName, RecSyntax, MaxRecords, RecsPerServer,
	    Start,AttrSet);
    fclose(fp);
    LogFile << "[" << MyPID << "] ";
    LogFile << "Writing query " << TheTerm << " for " << DBName;
    LogFile << " to " << TempFile << endl;
    LogFile.flush();

    strcpy(LocalAction,"SEARCH");
    if(PingManager(PID)<1){

#ifndef DEBUG
      LogFile << "Deleting " << TempFile << endl;
      unlink(TempFile);
#endif

      LogFile << "Returning from GenerateSearchRequest" << endl;
      LogFile.flush();
      return(0);
    }
  }
  return(DoWait());
}


// handle a logout request 

INT 
GATE::GenerateLogOutRequest()
{
  
  CHR             TmpFile[1024],TmpFile1[1024];
  FILE           *fp;
  
  pid_t           PID = CheckSession();
  pid_t           ThisPID = getpid();
  
  // clean up old files
  
  sprintf(TmpFile, "%s/%s.%d", ZgateTrackerPath, ZgateTrackerName, PID);
  unlink(TmpFile);
  sprintf(TmpFile, "%s/%s.%d", ZconHistoryPath, ZconHistoryName, PID);
  unlink(TmpFile);
  
  sprintf(TmpFile, "%s/%s.%d", ScratchDir,ZconFileName, PID);

  sprintf(TmpFile1, "%s/%s", ZgateGatewayPath,ZgateGatewayHTML);

  strcpy(LocalAction,"LOGOUT");
  strcpy(LocalTempFile,TmpFile);

  CHR MsgBuf[512];

  if ((fp = fopen(TmpFile, "w")) == NULL) {
    sprintf(MsgBuf,"Couldn't Open Gateway Transfer File %s", TmpFile);
    SetErrorMessage(MsgBuf);
    LogFile << "[" << ThisPID << "] ";
    LogFile << MsgBuf << endl;
    LogFile.flush();
    return(0);
  } else {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Logout: Opened Gateway Transfer file " << TmpFile << endl;
    LogFile.flush();
  }

  fprintf(fp, "%d\nLOGOUT\n", ThisPID);
  fclose(fp);
  
  if(PingManager(PID)<1){
    unlink(TmpFile);
    SetErrorMessage("Session Already Closed (Timeout or Logout)");
    return(0);
  }
  
  LogFile << "[" << ThisPID << "] ";
  LogFile << "Logout succeeded." << endl;
  LogFile.flush();

#ifndef DEBUG
  unlink(TmpFile);
  LogFile << "[" << ThisPID << "] ";
  LogFile << "Logout: Deleting Gateway Transfer file " << TmpFile << endl;
  LogFile.flush();
#endif

  DoWait();
  if ((fp = fopen(TmpFile,"r")) == NULL) {
    DisplayGWForm();
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Gateway form displayed" << endl;
    LogFile.flush();
  } else {
    fclose(fp);
    STRING HTMLout;
    HTMLout.ReadFile(TmpFile);
    cout << HTMLout << endl;
    unlink(TmpFile);
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Logout: deleting last communication file " << TmpFile << endl;
    LogFile.flush();
  }

  return(1);
}


// Create a SCAN Request and pass to zcon

INT 
GATE::GenerateScanRequest()
{
  
  CHR TempFile[512],tf[256];
  FILE *fq,*fp;
  STRING In,Out;
  CHR *TheTerm,*TermM,*NumReq,*DBScan,*PrefPos;
  INFIX2RPN       qp;
  
  pid_t           PID = CheckSession();
  sprintf(TempFile, "%s/%s.%d", ScratchDir,ZconFileName, PID);
  strcpy(LocalTempFile,TempFile);
  sprintf(tf, "%s/%s.%d", ZgateTrackerPath,ZgateTrackerName, PID);
  fq = fopen(tf, "w");
  if(fq==NULL){
    SetErrorMessage("Couldn't open Query Page Tracker File");
    LogFile << "[" << PID << "] ";
    LogFile << "Couldn't open Query Page Tracker File " << tf << endl;
    LogFile.flush();
    //    return(0);
  }

  if (fq) {
    fprintf(fq, "%s/scan.html",ZgateGatewayPath);
    fclose(fq);
  }

  if ((DBScan = cgi_GetValueByName(cgi, "DBNAME")) == NULL) {
    DBScan = new CHR[11];
    strcpy(DBScan, "XXDEFAULT");
  }
  if ((NumReq = cgi_GetValueByName(cgi, "NUMREQ")) == NULL) {
    NumReq = (CHR *) malloc(3);
    strcpy(NumReq, "12");
  }
  if ((PrefPos = cgi_GetValueByName(cgi, "PREFPOS")) == NULL) {
    PrefPos = (CHR *) malloc(3);
    strcpy(PrefPos, "1");
  }
  
  
  
  TheTerm = (CHR *) malloc(2048);
  TermM = cgi_GetValueByName(cgi, "TERM_M");
  In = TermM;
  qp.Parse(In, &Out);
  Out.GetCString(TheTerm, 256);
  
  if ((fp = fopen(TempFile, "w")) == NULL) {
    SetErrorMessage("Couldn't Open Gateway Transfer File");
    return(0);
  }
  
  fprintf(fp, "%d\nSCAN\n%s\n%s\n%s\n%s\n", getpid(),
	  DBScan, TheTerm, PrefPos, NumReq);
  fclose(fp);
  strcpy(LocalAction,"SCAN");
  if(PingManager(PID)<1){
    unlink(TempFile);
    return(0);
  }
  
  return(DoWait());
}


INT 
GATE::GeneratePresentRequest()
{
  
  pid_t           PID;
  INT4            Start, Count, HitCount=0;
  CHR TempFile[1024];
  CHR *ESName,*RecSyntax,*RSName,*Tmp;
  FILE *fp;

  if((Tmp=cgi_GetValueByName(cgi,"SESSION_ID")) == NULL) {
    cout << "Form didn't include a SESSION_ID!";
    cout << endl;
    unlink(TempFile);
    MyExit(1);
  }
  PID= CheckSession();
  
  if ((ESName = cgi_GetValueByName(cgi, "ESNAME")) == NULL) {
    ESName = new CHR[11];
    strcpy(ESName, "B");
  }

  if ((RecSyntax = cgi_GetValueByName(cgi, "RECSYNTAX")) == NULL) {
    RecSyntax = new CHR[strlen(SUTRS_OID) + 1];
    strcpy(RecSyntax, SUTRS_OID);
  }

  if ((RSName = cgi_GetValueByName(cgi, "RSNAME")) == NULL) {
    RSName = new CHR[11];
    strcpy(RSName, "DEFAULT");
  }

  if (!(StrCaseCmp(ESName,"B"))) {
    if ((Tmp = cgi_GetValueByName(cgi, "HITCOUNT")) == NULL) {
      SetErrorMessage("Form Didn't include HITCOUNT");
      unlink(TempFile);
      return(0);
    }
    HitCount = atoi(Tmp);
  }

  if ((Tmp = cgi_GetValueByName(cgi, "START")) == NULL) {
    SetErrorMessage("Form Didn't include START");
    unlink(TempFile);
    return(0);
  }
  Start = atoi(Tmp);

  if ((Tmp = cgi_GetValueByName(cgi, "COUNT")) == NULL) {
    SetErrorMessage("Form Didn't include COUNT");
    unlink(TempFile);
    return(0);
  }
  Count = atoi(Tmp);

  sprintf(TempFile, "%s/%s.%d", ScratchDir,ZconFileName, PID);
  //	  ZconFilePath,ZconFileName, PID);
  strcpy(LocalTempFile,TempFile);
  
  if ((fp = fopen(TempFile, "w")) == NULL) {
    SetErrorMessage("Couldn't Open Gateway Transfer File");
    return(0);
  }
  fprintf(fp, "%d\nPRESENT\n%s\n%d\n%d\n%s\n%s\n%d\n", getpid(),
	  RSName, Start, Count, ESName, RecSyntax, HitCount);
  fclose(fp);
/*
    FILE *fy;
    fy=fopen("/tmp/present.txt","w");
    fprintf(fy, "%ld\nPRESENT\n%s\n%d\n%d\n%s\n%s\n%d\n", getpid(),
	  RSName, Start, Count, ESName, RecSyntax, HitCount);
    fclose(fy);
*/
  strcpy(LocalAction,"PRESENT");
  if(PingManager(PID)<1){
    unlink(TempFile);
    return(0);
  }
  return(DoWait());
}


// Generate an INIT request
// On error, return an error code and set a message
// return 1 on success

INT 
GATE::GenerateInitRequest()
{
  STRING          TempString;
  CHR             UserDir[256],buf[1024];
  INT             Port;
  CHR            *Tmp;
  CHR            *Host;
  STRLIST         List;
  CHR            *GroupId = (CHR*)NULL;
  CHR            *UserId = (CHR*)NULL;
  CHR            *Password = (CHR*)NULL;
  CHR            *UserPassUser = (CHR*)NULL;
  CHR            *UserPassPass = (CHR*)NULL;
  FILE           *fu,*fp;
  struct stat     statbuf;
  INT             err,PID;
  pid_t           ThisPID = getpid();
  GDT_BOOLEAN     NewPWtype=GDT_FALSE;
  
  if ((Tmp = cgi_GetValueByName(cgi, "FORM_HOST_PORT")) == NULL) {
    SetErrorMessage("Your form didn't include a FORM_HOST_PORT Field");
    return(-1);
  }

  TempString = Tmp;
  List.Split(",", TempString);
  List.GetEntry(1, &TempString);
  DBInfo = TempString.NewCString();

  // If this file doesn't exist, die now
  if (stat(DBInfo, &statbuf) == -1) {
    SetErrorMessage("Missing Search Form");
    return(-1);
  }

  List.GetEntry(2, &TempString);
  Host = TempString.NewCString();
  List.GetEntry(3, &TempString);
  Port = TempString.GetInt();

#ifdef DEBUG
  cout << "DBInfo=" << DBInfo << "<br>" << endl;
  cout << "Port=" << Port << "<br>" << endl;
  cout << "Host=" << Host << "<br>" << endl;
#endif


  //  Did the query come from an HTTP page or a JAVA applet?
  //  If from the Java applet, we don't have to display a search form
  //
  if((QuerySource = cgi_GetValueByName(cgi, "CLIENT_TYPE")) == NULL) {
    QuerySource = (CHR *)malloc(5);
    strcpy(QuerySource, "HTML");
  }

  /*
   * Do we have a groupid/userid/password?
   */
  
  TempString = "";
  if ((GroupId = cgi_GetValueByName(cgi, "GROUPID")) == NULL) {
    GroupId = (CHR *)malloc(8);
    strcpy(GroupId, "DEFAULT");
  }
  if ((UserId = cgi_GetValueByName(cgi, "USERID")) == NULL) {
    UserId = (CHR *)malloc(7);
    strcpy(UserId, "nobody");
  }
  if ((Password = cgi_GetValueByName(cgi, "PASSWORD")) == NULL) {
    Password = (CHR *)malloc(2);
    strcpy(Password, "-");
  }

  if (!(ValidateUserAndPassword(GroupId,UserId,Password)) ) {
    SetErrorMessage("Bad username or password");
    return(-1);
  }

#ifdef DEBUG
  cout << "GroupId=" << GroupId << "<br>" << endl;
  cout << "UserId=" << UserId << "<br>" << endl;
  cout << "Password=" << Password << "<br>" << endl;
#endif

  /*
   * These are to support the old-style user/pass style encoding
   * Only look for this if we didn't get the USERID variable in the form

   if (!NewPWtype) {
   if ((UserPassUser = cgi_GetValueByName(cgi, "USERPASS_USER")) == NULL) {
   UserPassUser = (CHR *)malloc(1);
   strcpy(UserPassUser, "");
   }

   #ifdef DEBUG
   cout << "UserPassUser=" << UserPassUser << "<br>" << endl;
   #endif

   if ((UserPassPass = cgi_GetValueByName(cgi, "USERPASS_PASS")) == NULL) {
   UserPassPass = (CHR *)malloc(1);
   strcpy(UserPassPass, "");
   }

   #ifdef DEBUG
   cout << "UserPassPass=" << UserPassPass << "<br>" << endl;
   #endif

   strcpy(InitUserId, UserPassUser);
   } else {
   strcpy(InitUserId, UserId);
   }

   */

  if (strlen(UserId) > 0)
   strcpy(InitUserId, UserId);
  else
   strcpy(InitUserId, "nobody");

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Request from user " << InitUserId << endl;
  LogFile.flush();

  sprintf(UserDir, "%s/%s", SPATH,InitUserId);

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Opening user dir " << UserDir << endl;
  LogFile.flush();

  fu=fopen(UserDir,"r");

  if (!fu) {
    // create initial stuff for new users...
    //sprintf(UserDir, "%s/%s", SPATH,InitUserId);
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Open failed.  Creating user dir " << UserDir << endl;
    LogFile.flush();

    mkdir(UserDir, 0x1ff);	/* make the user S directory */
  }

  // See if we can make the Count file in that directory now...
  sprintf(UserDir, "%s/%s/Count", SPATH,InitUserId);
  fu = fopen(UserDir, "w");
  if (!fu) {
    // It didn't get created, bail out for now.
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Count file " << UserDir << " not created." << endl;
    LogFile.flush();
    CHR MsgBuf[512];
    sprintf(MsgBuf,"Couldn't Open %s",UserDir);
    SetErrorMessage(MsgBuf);
    return(-1);
  } else {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Count file " << UserDir << " created." << endl;
    LogFile.flush();
    fprintf(fu, "1\n");
    fclose(fu);
  }

  //  if (NewUserPass) {
    if (GroupId) {
      TempString.Cat(" -gid ");
      TempString.Cat(GroupId);
    }
    if (UserId) {
      TempString.Cat(" -uid ");
      TempString.Cat(UserId);
    }
    if (Password) {
      TempString.Cat(" -pass ");
      TempString.Cat(Password);
    }
    /*
      } else {
      if ((UserPassUser) && (UserPassPass)) {
      TempString.Cat(" -pass ");
      TempString.Cat(UserPassUser);
      TempString.Cat("/");
      TempString.Cat(UserPassPass);
      }
      }
      */

#ifdef DEBUG
  cout << "PW string =" << TempString << "<BR>" << endl;
#endif

  LogFile << "[" << ThisPID << "] ";
  LogFile << "CLIENT_TYPE=" << QuerySource << endl;
  LogFile.flush();

  char            cmd[1024];

  TempFile = tempnam("/tmp", ZgateFileName);

#ifndef OLD_CODE
  if (TempString.GetLength() > 0) {
    CHR *GroupUserPass;
    GroupUserPass = TempString.NewCString();
    sprintf(cmd, "%s/%s %s %d %s %i %s >/dev/null 2>&1 &", 
	    mypath, ZconProgName, TempFile, getpid(), Host, Port, 
	    GroupUserPass);
    delete GroupUserPass;
  } else {
    sprintf(cmd, "%s/%s %s %d %s %i >/dev/null 2>&1 &", 
	    mypath, ZconProgName, TempFile, getpid(), Host, Port);
  }
#else
  if (TempString.GetLength() > 0) {
    CHR *GroupUserPass;
    GroupUserPass = TempString.NewCString();
    sprintf(cmd, "%s/%s %s %d %s %i %s &", mypath, ZconProgName,
	    TempFile, getpid(), Host, Port, GroupUserPass);
    delete GroupUserPass;
  } else {
    sprintf(cmd, "%s/%s %s %d %s %i &", mypath, ZconProgName,
	    TempFile, getpid(), Host, Port);
  }
#endif

#ifdef DEBUG
  FILE *fy;
  fy=fopen("/tmp/dump.txt","a");
  fprintf(fy,"%s\n",cmd);
  fclose(fy);
#endif

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Calling Connection Manager:" << cmd << endl;
  LogFile.flush();

  // Prepare for signal from zcon after launching
  strcpy(LocalTempFile,TempFile);
  LocalBreakLoop=0;
  LocalError=0;
  strcpy(LocalAction,"INIT");

  // Launch zcon - it'll signal when it has initialized
  err = system(cmd);

  if (err == -1) {
    // Oops - the system call failed
    CHR MsgBuf[512];
    sprintf(MsgBuf,"Failed to start Connection Manager (%s)",cmd);
    SetErrorMessage(MsgBuf);
    LogFile << "[" << ThisPID << "] ";
    LogFile << MsgBuf << endl;
    LogFile.flush();
    return(-1);
  }

  // Wait for signal from zcon
  for(;;) {
    if (LocalBreakLoop != 1) // Don't pause if signal has been handled
      pause();
    if(LocalError==1) {
      SetErrorMessage("Error in Signal Handler");
      LogFile << "[" << ThisPID << "] ";
      LogFile << "Error in Signal Handler" << endl;
      LogFile.flush();
      return(0);
    }
    if(LocalBreakLoop==1) 
      break;
  }

  if ((fp = fopen(TempFile, "r")) == NULL) {
    unlink(TempFile);
    SetErrorMessage("Couldn't Open Results of Init");
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Couldn't Open Results of INIT" << endl;
    LogFile.flush();
    return(0);
  } else {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Opened transfer file " << TempFile << endl;
    LogFile.flush();
  }
  fgets(buf, sizeof(buf), fp);

  if (!StrCaseCmp(buf, "SUCCESS\n")) {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Login succeeded." << endl;
    LogFile.flush();

    /* Display search form */
    CHR MapString[1024];
    CHR min_lat[80], max_lat[80], min_lon[80], max_lon[80];

    strcpy(MapString,"world-300x300.gif");
    strcpy(min_lat,"-90");
    strcpy(max_lat,"90");
    strcpy(min_lon,"-180");
    strcpy(max_lon,"180");

    fgets(buf, sizeof(buf), fp);
    PID = atoi(buf);
    /* Read the message of the day - it's returned by INIT */
    int len = fread(buf, 1, sizeof(buf), fp);
    buf[len] = '\0';
    DisplaySearchForm(DBInfo, MapString, min_lat, max_lat, min_lon,
		      max_lon, NumDatabases, database_names, 
		      database_values, PID, (PCHR)NULL);
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Search Form Displayed" << endl;
    LogFile.flush();
  } else {
    // we probably need to improve this login failure message. 
    SetErrorMessage("Login Failed");
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Login Failed" << endl;
    LogFile.flush();
    fclose(fp);
    return(0);
  }
  fclose(fp);

  LogFile << "[" << ThisPID << "] ";
  LogFile << "GenerateInitRequest end - deleting " << TempFile << endl;
  LogFile.flush();

#ifndef DEBUG
  unlink(TempFile);
#endif

  return(1);
}


INT 
GATE::WriteQueryFiles(INT PID, CHR *DBName, CHR *TempTerm)
{
  CHR HistoryFile[512];
  FILE *fp;
  INT             SNumber = UpdateSNumberCount();
  CHR             SFile[256];
  
  sprintf(HistoryFile, "/%s/%s.%d", ZconHistoryPath, ZconHistoryPath, PID);
  sprintf(SFile, "/%s/%s/%05d", SPATH,InitUserId, SNumber);
  fp = fopen(SFile, "w");
  if(fp==NULL)
    return(-1);
  fprintf(fp, "%s\n%s\n", DBName, TempTerm);
  fclose(fp);
  
  fp = fopen(HistoryFile, "a");
  if(fp==NULL)
    return(-1);
  fprintf(fp,
	  "<TR><TD>%s</TD><TD>%s</TD>",
	  DBName, TempTerm);
  fclose(fp);
  return(1);
}

//---------------------------------------------------------------//
// PingManager                                                   //
//                                                               //
// This routine locks the two semaphore variables                //
//       LocalBreakLoop, and                                     //
//       LocalError,                                             //
// and then signals the zcon process indicated by PID.           // 
//                                                               //
// This routine should be called immediately before DoWait.      //
//                                                               //
// Note:                                                         //
// LocalBreakLoop and LocalError are semaphores that are         //
// locked (set to 0) by PingManager, and released (set to 1)     //
// by USR1Handler.                                               //
//                                                               //
//---------------------------------------------------------------//

INT 
GATE::PingManager(INT PID)
{

  CHR Msg[255];

  LocalBreakLoop=LocalError=0;                  // lock semaphores

  if ((PID>0) && (kill(PID, SIGUSR1) == -1)) {  // signal zcon
    sprintf(Msg,
	    "Your session has expired. Please <A HREF=\"%s\">reinitialize.</A>",
	    ZgateGatewayURL);

    pid_t ThisPID=getpid();
    SetErrorMessage(Msg);
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Session " << PID << " expired." << endl;
    LogFile.flush();
    return(-1);
  }
  else return(1);
}

INT 
GATE::LoadDBNames(CHR *database_names[], CHR *database_values[])
{
  INT nDB=0;
  INT i;
  STRING DBData,Name,Value;
  STRINGINDEX split;

  nDB = DatabaseList.GetTotalEntries();
  for (i=0;i<nDB;i++) {
    DatabaseList.GetEntry(i+1,&DBData);
    Name=DBData;
    split = Name.SearchReverse("+");
    Name.EraseAfter(split-1);
    database_names[i] = Name.NewCString();

    Value = DBData;
    Value.EraseBefore(split+1);
    database_values[i] = Value.NewCString();
  }
  return nDB;
}


CHR* 
GATE::GetFormsPath() {
  return(ZgateGatewayPath);
}


CHR* 
GATE::GetScratchDir() {
  return(ScratchDir);
}


UINT 
GATE::GetGatewayTimeout() {
  return(Timeout);
}


CHR* 
GATE::GetGatewayHTML() {
  return(ZgateGatewayHTML);
}


CHR* 
GATE::GetGatewayPath() {
  return(ZgateFilePath);
}


CHR* 
GATE::GetGatewayName() {
  return(ZgateFileName);
}


CHR* 
GATE::GetGatewayURL() {
  return(ZgateFileURL);
}


CHR* 
GATE::GetTrackerPath() {
  return(ZgateTrackerPath);
}


CHR* 
GATE::GetTrackerName() {
  return(ZgateTrackerName);
}


CHR* 
GATE::GetZconPath() {
  return(ZconFilePath);
}


CHR* 
GATE::GetZconName() {
  return(ZconFileName);
}


CHR* 
GATE::GetHistoryPath() {
  return(ZconHistoryPath);
}


CHR* 
GATE::GetHistoryName() {
  return(ZconHistoryName);
}


CHR* 
GATE::GetLockPath() {
  return(ZconLockPath);
}


CHR* 
GATE::GetLockName() {
  return(ZconLockName);
}


CHR* 
GATE::GetZconProgName() {
  return(ZconProgName);
}


void 
GATE::SetIniFilename(const STRING* NewName) {
  IniFileName = NewName;
}


STRING 
GATE::GetIniFilename() {
  return IniFileName;
}

#ifdef GW_PASSWD
// Check to see if the user is authorized

GDT_BOOLEAN 
GATE::ValidateUserAndPassword(const CHR* GroupId, const CHR* UserId,
					const CHR* Password) {
  // The passwords are in a GDBM database, in the directory PasswdDir
  // The filename will be <GroupId>.db
  GDBM_FILE   passwd_dbf;
  //DBM   *passwd_dbf;
  pid_t       PID = getpid();
  STRING      PasswdFile;
  CHR        *cPasswdFile;
  CHR        *pCryptPassword, *pTestPassword;
  CHR        *key;
  size_t      CryptLength, TestLength;
  datum       key_data,return_data;
  GDT_BOOLEAN Status=GDT_TRUE; // for now, don't return false
  INT         block_size, read_write, mode;

  if ((strlen(GroupId) == 0) || (strlen(UserId) == 0)
      || (GroupId == NULL) || (UserId == NULL)) {
    LogFile << "[" << PID << "] ";
    LogFile << "GroupId or UserId missing." << endl;
    LogFile.flush();
    return Status;
  }

  block_size = 512;
  read_write = GDBM_READER;
  //  read_write = 0;
  mode       = 0664;

  key = new CHR[strlen(UserId)+1];
  strcpy(key,UserId);

  // Build the password file name
  PasswdFile = PasswdDir;
  PasswdFile.Cat("/");
  PasswdFile.Cat(GroupId);
  PasswdFile.Cat(".db");
  cPasswdFile = PasswdFile.NewCString();

  LogFile << "[" << PID << "] ";
  LogFile << "Check " << cPasswdFile << " for ";
  LogFile << "UserId=" << UserId << endl;
  LogFile.flush();

  // Open the password file
  passwd_dbf  =  gdbm_open(cPasswdFile, block_size, read_write, mode, 0);
  //passwd_dbf  =  dbm_open(cPasswdFile, read_write, mode);
  if (passwd_dbf) { // it opened successfully

    // Store the user's ID as the database key
    key_data.dptr = key;
    key_data.dsize = strlen(key)+1;

    // fetch the associated password
    return_data = gdbm_fetch (passwd_dbf, key_data);
    //return_data = dbm_fetch (passwd_dbf, key_data);

    if (return_data.dptr != NULL) { // Got something back
      // Save the retrieved value
      pTestPassword = return_data.dptr;

      // Now, encrypt the password the user entered
      pCryptPassword = pTestPassword;

      // And compare the results with the retrieved value
      if (strcmp(pTestPassword,pCryptPassword) == 0)
	Status = GDT_TRUE;
      free (return_data.dptr);
    } else { // Nothing returned from gdbm_fetch
      LogFile << "[" << PID << "] ";
      LogFile << "Failed to locate password for user " << UserId << endl;
      LogFile.flush();
    }
    // Clean up
    gdbm_close(passwd_dbf);
    //dbm_close(passwd_dbf);
  } else { // cPasswdFile didn't open
    LogFile << "[" << PID << "] ";
    LogFile << "Failed to open password file." << endl;
    LogFile.flush();
  }

  delete [] cPasswdFile;
  delete [] key;

  return Status;
}
#else
GDT_BOOLEAN 
GATE::ValidateUserAndPassword(const CHR* GroupId, const CHR* UserId,
					const CHR* Password) {
  // No password checking - just return GET_TRUE
  return GDT_TRUE;
}
#endif
