/* $Id: gate.hxx,v 1.7 1998/01/30 19:10:58 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/
/*@@@
File:		gate.hxx
Version:	1.00
$Revision: 1.7 $
Description:	Z39.50 Gateway Input Class
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#ifndef _GATE_HXX_
#define _GATE_HXX_

#ifndef __SUNPRO_CC
void ZgateAlarmHandler(int t);
void ZgateStopHandler(int t);
void USR1Handler(int t);
void SegvHandler(int t);
#else
void ZgateAlarmHandler(int t,...);
void ZgateStopHandler(int t,...);
void USR1Handler(int t,...);
void SegvHandler(int t,...);
#endif

/// Z39.50 Gateway Input Class
/*@Doc:
\begin{description}

\item[Function]  This class performs all conversions from HTML form input to
file input to the {\bf zcon} Z39.50 client process.  In general, this class parses HTML forms input, generates a suitable input file to zcon based on this input material, and sends signals to {\bf zcon} indicating the availability of query or retrieval information.  The class then waits for a signal from {\bf zcon}
indicating the availability of formatted results in the transfer file.
\item[Algorithms]  This class uses a variety of algorithms to accomplish it's tasks -- these algorithms are described for each method.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the comiling software.
\item[Security] This class extracts username and password information from
the initialization page transmitted from the web server. This information
is transferred to {\bf zcon} for transmission to the Z39.50 server. 
\end{description}
*/

class GATE {
 private:
  CHR          InitUserId[256],*TempFile,*DBInfo,*Action;
  CHR          mypath[1024],LocalErrorMessage[512];
  INT          Usr1Error;		// 0 if Usr1 was OK, else 1
  INT          BreakLoop;
  CGI         *cgi;
  REGISTRY    *Defaults;
  STRLIST      DatabaseList; 
  STRING       IniFileName;

  //  Some configuration information
  CHR         *WebForms;
  CHR         *ScratchDir;
  UINT        Timeout;
  CHR         *PasswdDir; 

  // zgate paths and names
  CHR         *ZgateGatewayPath;
  CHR         *ZgateGatewayHTML;
  CHR         *ZgateGatewayURL;
  CHR         *ZgateFilePath;
  CHR         *ZgateFileName;
  CHR         *ZgateFileURL;
  CHR         *ZgateTrackerPath;
  CHR         *ZgateTrackerName;

  // zcon paths and names
  CHR         *ZconFilePath;
  CHR         *ZconFileName;
  CHR         *ZconHistoryPath;
  CHR         *ZconHistoryName;
  CHR         *ZconLockPath;
  CHR         *ZconLockName;
  CHR         *ZconProgPath;
  CHR         *ZconProgName;

/// Shows main login form
  void DisplayGWForm(CHR * Filename, INT SessionId);

/*@Doc: Display the login form that prompts for username and password, and
causes {\bf zgate} and {\bf zcon} to initiate a connection.  The form is loaded from the file HTML\_FULL\_PATH/gw.html.  HTML\_FULL\_PATH is a system-level define string.*/

/// Shows main login form
  void DisplayGWForm(void);
/*@Doc: Display the login form that prompts for username and password, and
causes {\bf zgate} and {\bf zcon} to initiate a connection.  The form is loaded from the file HTML\_FULL\_PATH/gw.html.  HTML\_FULL\_PATH is a system-level define string.*/

/// Update the current S Number
  INT UpdateSNumberCount();
/*@Doc:  Open the current S-Number tracking file and increment the current 
count */

/// Update history and S number data
  INT WriteQueryFiles(INT PID, CHR *DBName, CHR *TempTerm);

/// Signal ZCON to read command
  INT PingManager(INT PID);
/*@Doc: This method sends a SIGUSR1 signal to the {\bf zcon} process 
designated by the internally tracked process id passed to {\bf zgate} (and 
gate) by the Web client.  SIGUSR1 signifies that operation data is available 
in the file {\tt zcon.pid} where {\tt pid} is the process ID passed from the 
web client.*/

/// Set text of error message
  void SetErrorMessage(CHR *s);
/*@Doc: Set a textual error message for later processing and display to the user */

/// Add S Number Support
  CHR *SNumber(STRING ETerm,char *GlobalUserId);
/*@Doc: Look up a S number value and replace it with the associated query.*/

/// Grab defaults from INI file
/*
  void LoadDefaultsFromFile(const STRING & Filename,
                                    const STRING & Group);
*/

public:

/// Basic Constructor
  GATE(char *mp,CGI *cgi);
/*@Doc:  Sets up appropriate signal interfaces to internal signal handlers*/

/// Basic Destructor
  ~GATE();  
/*@Doc: Basic destructor.  Called for C++ system cleanup and garbage collection.*/

/// Retrieve latest error message
  void GetErrorMessage(CHR *s);
/*@Doc: Retrieves the latest error message set by the internal private {\it SetErrorMessage()} method*/

/// Get the session ID
  INT CheckSession( );
/*@Doc:  This method extracts the current session (pid) id from the submitted HTML page.*/

/// Do a search 
  INT GenerateSearchRequest(INT nDBs);
/*@Doc:  This method extracts and processes query information from a Search page, processes the queries into the internal query format, and signals 
{\bf zcon} to read the input file and execute the query operation.*/

/// Do a SCAN
  INT GenerateScanRequest();
/*@Doc:  This method extracts and process information related to a Messenger 
EXPAND operation, writes it to the input file and signals {\bf zcon}.*/

/// Log Out
  INT GenerateLogOutRequest();
/*@Doc: This method writes a log-off request to the {\bf zcon} input file and 
signals {\bf zcon} to process the request.*/

/// Present records
  INT GeneratePresentRequest();
/*@Doc: Extracts information from any page that can cause records to be 
presented to the user.  Creates the {\bf zcon} input file with parameters 
that describe which group of records to retrieve, the format in which they 
should be presented, and the elemetn sets that should be presented.*/

/// Perform an INIT
  INT GenerateInitRequest();
/*@Doc:  Generate {\bf zcon} input file information to log a user onto the 
system. This information is {\bf not} passed in the standard {\bf zcon.pid} file, as a zcon process has not yet been launched.  Information is instead passed in {\bf /tmp/zconxxxxxxx}, where {\it xxxxxxx} is a system-generated random string.  This method actually launches the {\bf zcon} connection manager process, which is found in the {\bf bin} directory designated at compile time */

/// Get the list of DB names from the INI file
  INT LoadDBNames(CHR *database_names[], CHR *database_values[]);  

/// Get the local path to the HTML forms (WEBFORMS)
  CHR* GetFormsPath();

/// Get the name of a scratch directory
  CHR* GetScratchDir();

/// Get the process timeout limit
  UINT GetGatewayTimeout();

/// Get the URL path to the gateway HTML form (ZgateGatewayPath)
  CHR* GetGatewayHTMLPath();

/// Get the name of the gateway HTML form (ZgateGatewayHTML)
  CHR* GetGatewayHTML();

/// Get the path to the gateway binary (ZgateFilePath)
  CHR* GetGatewayPath();

/// Get the name of the gateway binary (ZgateFileName)
  CHR* GetGatewayName();

/// Get the URL of the gateway binary (ZgateFileURL)
  CHR* GetGatewayURL();

/// Get the path to the tracker file (ZgateTrackerPath)
  CHR* GetTrackerPath();

/// Get the name of the tracker file (ZgateTrackerName)
  CHR* GetTrackerName();

/// Get the path to the zcon binary (ZconFilePath)
  CHR* GetZconPath();

/// Get the name of the zcon binary (ZconFileName)
  CHR* GetZconName();

/// Get the path to the history file (ZconHistoryPath)
  CHR* GetHistoryPath();

/// Get the name of the history file (ZconHistoryName)
  CHR* GetHistoryName();

/// Get the path to the lock file (ZconLockPath)
  CHR* GetLockPath();

/// Get the name of the lock file (ZconLockName)
  CHR* GetLockName();

/// Get the zcon program name (ZconProgName)
  CHR* GetZconProgName();

/// Saves the name of the INI file
  void SetIniFilename(const STRING* NewName);
/*@Doc: Store the name of the default gateway initialization file */

/// Retrieves the name of the INI file */
  STRING GetIniFilename();
/*@Doc: Gets the name of the current gateway initialization file */

  /// Validates the user name and password against the gateway access file
  GDT_BOOLEAN  ValidateUserAndPassword(const CHR* GroupId, const CHR* UserId,
				       const CHR* Password);
  /*@Doc: Checks to see if the user's name and password match against the
    group database */
};

#endif  // _GATE_HXX_
