/* $Id: ipstream.hxx,v 1.4 1997/09/28 16:28:28 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		ipstream.hxx
Version:        1.00
Description:    Class ipstream
Author:         Kevin Gamiel, Kevin.Gamiel@cnidr.org
@@@*/

#ifndef _ipstream_
#define _ipstream_

#include <iostream.h>
#include "tcpbuf.hxx"
#include "ipbuf.hxx"

///Stream-based Network I/O Base Class
/*@Doc:
\begin{description}

\item[Function]  Provide basic I/O management for TCP/IP C++ streams
\item[Algorithms]  Basic open, close and status routine management

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/

class ipstreambase : public iostream {
  tcpbuf c_tcpbuf;
  void __ib_init();

public:
	///Basic Constructor
  ipstreambase();	
/*@Doc: Create basic ipstream object*/
	
/// Open an ipstream
  void Open(const CHR *host, UINT port);
/*@Doc: Opens a connection to a given host/port*/

///Close this connection
  void Close();
/*@Doc: Closes connection managed by this object*/

/// Determine connection status
  INT IsConnected() const { return rdbuf()->IsConnected(); }
/*@Doc:  Returns 1 if this object is maintaining a connection, 0 otherwise*/

protected:
  void setbuf(PCHR ptr, INT len) { rdbuf()->setbuf(ptr, len); }
  tcpbuf* rdbuf() const { return (tcpbuf*)&c_tcpbuf; }

};

/// Input IP stream
/*@Doc:
\begin{description}

\item[Function]  This class manages a basic IP Input stream
\item[Algorithms] Basic input open and close operations.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class iipstream : public ipstreambase {
public:
///Basic Input stream constructor
	iipstream() : ipstreambase() { }
/*@Doc: derive an input stream from base class*/

///Open an input stream
	void Open(const CHR *host, UINT port)
		{ ipstreambase::Open(host, port); }
/*@Doc: Call base constructor to open input stream*/
};


/// Output IP stream
/*@Doc:
\begin{description}

\item[Function] Manages a basic TCP/IP Output stream
\item[Algorithms] Basic output open and close operations.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class oipstream : public ipstreambase {
public:
///Basic Output stream constructor
	oipstream() : ipstreambase() { }

///Open an output stream
	void Open(const CHR *host, UINT port)
		{ ipstreambase::Open(host, port); }
};

/// Provides Basic  IP Stream support

/*@Doc:
\begin{description}

\item[Function] Provides read/write access to TCP/IP sockets via a C++ stream interface.
\item[Algorithms] Basic input open and close operations.

\item[Flow] This is a fundamental c++ class.    Flow is described as needed within the individual method descriptions.
\item[Packaging] This fundamental class is linked and archived as determined by the compiling software.
\item[Security] This class has no security role or function.
\end{description}
*/
class ipstream : public ipstreambase {
public:
///Basic IP stream constructor
	ipstream() : ipstreambase() { }

///Open an IP stream
	void open(const CHR *host, UINT port)
		{ ipstreambase::Open(host, port); }
/*@Doc: Open a generic IP stream.*/

///Close an IP stream
	void close() { rdbuf()->Close(); }
/*@Doc: Close a generic IP stream*/

///Get status of this stream
	INT is_open() const { return rdbuf()->IsConnected(); }
/*@Doc: Return the connection status of this object*/

///Show Last Error
	INT LastError() { return rdbuf()->LastError();}
/*@Doc: Get code of most recent error on this stream*/

///Get Last Error Message
	void ErrorMessage(INT err, PCHR msg, INT maxlen)
		{ rdbuf()->ErrorMessage(err, msg, maxlen);}
/*@Doc: Reeturn error message for given code*/

///Get Last Error
	void ErrorMessage(PCHR msg, INT maxlen) 
		{ rdbuf()->ErrorMessage(msg, maxlen);}
/*@Doc: Get error message for most recent error*/
};

#endif

