/* $Id: izclient.cxx,v 1.9 1997/12/08 20:31:01 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery
and Retrieval, 1994.

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby
granted without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact.

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of
noteworthy uses of this software.

3. The names of MCNC and Clearinghouse for Networked Information
Discovery and Retrieval may not be used in any advertising or publicity
relating to the software without the specific, prior written permission
of MCNC/CNIDR.

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF
THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
************************************************************************/

/*@@@
File:           izclient.cxx
Version:        2.00
$Revision: 1.9 $
Description:    full-screen Z39.50 client
Author:         Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#include <stdlib.h>
#include <string.h>
#include <locale.h>

#include "defs.hxx"
#include "string.hxx"
#include "vlist.hxx"
#include "strlist.hxx"
#include "common.hxx"

#include "zsession.hxx"
//#include "config.hxx"
#include "zclibase.hxx"
#include "marc.hxx"

#if defined(_WIN32)
#define CLEAR_SCREEN() system("cls")
#elif defined(_MSDOS)
#define CLEAR_SCREEN() cout << "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
#else
#define CLEAR_SCREEN() system("clear")
#endif

#define MAX_HITS 20

void 
DisplayHits(INT4 TotalHits, ZRECORDLIST & HitList, INT4 Start, INT4 Count);

class IZCLIENT : public ZCLIENT_BASE {
  ZRECORDLIST	c_hitlist;
  INT4		c_display_start,
                c_display_count,
                c_hitcount,
                c_maxhits;
  void DisplayName();	
  CHR c;
  STRING        c_host,
                c_database,
		c_term,
		c_esname,
		c_recsyntax,
                c_attrset,
                c_username,
                c_password;
  UINT		c_port;
  enum		{ CLOSED, INIT } c_state;
public:
  IZCLIENT();
  ~IZCLIENT();

  GDT_BOOLEAN DisplayMainMenu();	
  void ReleaseNotes();
  GDT_BOOLEAN Connect();
  void Configure();
  void UserInfo();
  virtual GDT_BOOLEAN Search();
  GDT_BOOLEAN Add_Record();
  GDT_BOOLEAN Delete_Record();
  INT GetState() { return c_state; }
  void ConnectionDetails();
};


void 
IZCLIENT::ConnectionDetails()
{
  CLEAR_SCREEN();
  DisplayName();

  cout << "Connection Details" << endl << endl;
  PrintConnectionDetails();
  cout << "Press Enter to Continue...";
  cin.get(c);
}


void 
IZCLIENT::ReleaseNotes()
{
  CLEAR_SCREEN();
  DisplayName();

  cout << "Release Notes" << endl << endl;
  
  cout << "This simple client is for programmer reference purposes only!";
  cout << endl;
  cout << "The intention is to provide guidelines for client developers.";
  cout << endl << endl;
  cout << "We'll gladly accept comments and suggestions." << endl;
  cout << endl;
  cout << "isite@cnidr.org" << endl << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << "Press Enter to Continue...";
  cin.get(c);
}


void 
IZCLIENT::Configure()
{
  CHR buf[64];
  STRING recsyntax;
  STRING AttrSet;

  CLEAR_SCREEN();
  DisplayName();

  cout << "Configure" << endl << endl;

  if (c_attrset.Equals(BIB1_ATTRSET_OID))
    AttrSet = "bib1";
  else if (c_attrset.Equals(STAS_ATTRSET_OID))
    AttrSet = "stas";
  else if (c_attrset.Equals(GILS_ATTRSET_OID))
    AttrSet = "gils";
  else if (c_attrset.Equals(GEO_ATTRSET_OID))
    AttrSet = "geo";

  if (c_recsyntax.Equals(SUTRS_OID))
    recsyntax = "text";
  else if (c_recsyntax.Equals(HTML_OID))
    recsyntax = "HTML";
  else if (c_recsyntax.Equals(USMARC_OID))
    recsyntax = "USMARC";


  // Preferred Record Syntax
  cout << "Enter Attribute Set ID [" << AttrSet << "]: ";
  cin.getline(buf, sizeof(buf));
  if (*buf != '\0')
    AttrSet = buf;

  if ((AttrSet.CaseEquals("stas")) || (AttrSet.Equals(STAS_ATTRSET_OID)))
    c_attrset = STAS_ATTRSET_OID;
  else if ((AttrSet.CaseEquals("gils")) || (AttrSet.Equals(GILS_ATTRSET_OID)))
    c_attrset = GILS_ATTRSET_OID;
  else if ((AttrSet.CaseEquals("geo")) || (AttrSet.Equals(GEO_ATTRSET_OID)))
    c_attrset = GEO_ATTRSET_OID;
  else
    c_attrset = BIB1_ATTRSET_OID;


  // Preferred Record Syntax
  cout << "Enter Default Record Syntax [" << c_recsyntax << "]: ";
  cin.getline(buf, sizeof(buf));
  if (*buf != '\0')
    recsyntax = buf;

  if (recsyntax.CaseEquals("SUTRS"))
    c_recsyntax = SUTRS_OID;
  else if (recsyntax.CaseEquals("TEXT"))
    c_recsyntax = SUTRS_OID;
  else if (recsyntax.CaseEquals(SUTRS_OID))
    c_recsyntax = SUTRS_OID;
  else if ((recsyntax.CaseEquals("HTML")) || (recsyntax.Equals(HTML_OID)))
    c_recsyntax = HTML_OID;
  else if ((recsyntax.CaseEquals("USMARC")) || (recsyntax.Equals(USMARC_OID)))
    c_recsyntax = USMARC_OID;

  // Element Set Name
  cout << "Enter Default Element Set for display [" << c_esname << "]: ";
  cin.getline(buf, sizeof(buf));
  if (*buf != '\0')
    c_esname = buf;

  // MaxHits
  cout << "Enter Maximum # of hits [" << c_maxhits << "]: ";
  cin.getline(buf, sizeof(buf));
  if (*buf != '\0')
    c_maxhits = atoi(buf);
}


void 
IZCLIENT::UserInfo()
{
  CHR buf[64];

  CLEAR_SCREEN();
  DisplayName();

  cout << "User Info" << endl << endl;

  // User name
  cout << "Enter Login name [" << c_username << "]: ";
  cin.getline(buf, sizeof(buf));
  if (*buf != '\0')
    c_username = buf;

  // Password
  cout << "Enter password [" << c_password << "]: ";
  cin.getline(buf, sizeof(buf));
  if (*buf != '\0')
    c_password = buf;

}


GDT_BOOLEAN 
IZCLIENT::Search()
{
  CLEAR_SCREEN();
  DisplayName();
  cout << "Search" << endl << endl;
  cout << "Enter Database [" << c_database << "]: ";
  CHR buf[64];
  cin.getline(buf, sizeof(buf));

  if (*buf == '\0') {
    if(c_database == "")
      c_database = "xxdefault";
  } else 
    c_database = buf;

  cout << "Enter Term [" << c_term << "]: ";
  cin.getline(buf, sizeof(buf));

  if (*buf == '\0') {
    if(c_term == "") 
      return GDT_FALSE;
  } else 
    c_term = buf;

  STRING a="xx";

  if (!ZCLIENT_BASE::Search(c_database, c_term, "Default", c_attrset,
			   &c_hitcount, a, 1, c_maxhits, &c_hitlist)) 
    {
      c_hitcount = -1;
      return GDT_FALSE;
    }

  INT4 FetchCount;

  if (c_hitcount <= c_maxhits)
    FetchCount = c_hitcount;
  else
    FetchCount = c_maxhits;

  c_hitlist.Clear();

  if (c_hitcount) {
    if(!ZCLIENT_BASE::Present(1, FetchCount, c_recsyntax, "Default","B", 
			      &c_hitlist)) {
      cout << "Failed to present" << endl;
      return GDT_FALSE;
    }
 
    if(c_hitlist.GetRecordCount() <= 0) {
      cout << "No records returned!" << endl;
      return GDT_FALSE;
    }
  }

  DisplayHits(c_hitcount, c_hitlist, 1, c_maxhits);

  int num;
  ZRECORDLIST RecordList;
  INT4 Count;
  ZRECORDSCHOICE *Record;
  STRING Data, OID;
  CHR cmd[1024];
  CHR *TempFile = tmpnam((CHR *)NULL);
#if defined(_MSDOS) || defined(_WIN32)
  sprintf(cmd, "type %s", TempFile);
#else
  sprintf(cmd, "more %s", TempFile);
#endif
  FILE *fp;
  CHR c;
  for(;;) {
    RecordList.Clear();

    if (c_hitcount > 0) {
      cout << endl << "Enter # of record to view (<CR> returns to menu): ";
      cin.getline(buf, sizeof(buf));
      if(!buf[0])
	break;
    } else {
      cout << endl << "<CR> returns to menu: ";
      cin.getline(buf, sizeof(buf));
      break;
    }
    num = atoi(buf);
    Count = c_hitlist.GetRecordCount();	
    if((num <= 0) || (num > Count))
      break;
    //    if(!ZCLIENT_BASE::Present(num, 1, c_recsyntax, "Default","F", 
    //			      &RecordList)){
    if (!ZCLIENT_BASE::Present(num, 1, c_recsyntax, "Default",
			       c_esname, &RecordList)) {
      cout << "Failed to present" << endl;
      exit(1);
    }
    Count = RecordList.GetRecordCount();	
    if(Count < 1) {
      cout << "No records returned!" << endl;
      exit(1);
    }
    Record = RecordList.GetRecord(0);
    if(!Record) {
      cout << "Record not available!" << endl;
      continue;
    }
    Record->GetOID(&OID);
    Record->GetRecordData(&Data);
    if((fp = fopen(TempFile, "w")) == NULL) {
      perror(TempFile);
      exit(1);
    }
    fprintf(fp, "CNIDR izclient, Version %s, Copyright (c) 1995,1996,1997 MCNC/CNIDR\n", VERS);
    fprintf(fp, "______________________________________________________________________________\n");
    fprintf(fp, "Record %d of %d\n\n", num, c_hitcount);
    if(OID == USMARC_OID) {
      MARC *m;
      m = new MARC(Data);
      m->Print(fp);
      delete m;
    } else {
      CHR *buf;
      buf = Data.NewCString();
      fwrite(buf, 1, Data.GetLength(), fp);
    }
    fclose(fp);
    CLEAR_SCREEN();
		
#if defined(_MSDOS) && !defined(_WIN32)
    cout << endl;
    cout << "Sorry...I can't view records in this enviroment!";
#else
    system(cmd);
#endif
    cout << endl << "Press return continue...";
    cin.get(c);
    CLEAR_SCREEN();
    DisplayHits(c_hitcount, c_hitlist, 1, 10);
  }
  return GDT_TRUE;
}


GDT_BOOLEAN 
IZCLIENT::Add_Record()
{
  CLEAR_SCREEN();
  DisplayName();
  cout << "Add Record" << endl << endl;
  cout << "Enter Database [" << c_database << "]: ";
  
  CHR buf[132];
  
  cin.getline(buf, sizeof(buf));
  if(*buf == '\0') {
    if(c_database == "")
      c_database = "xxdefault";
  } else 
    c_database = buf;

  ZESUPDATERECORD *ur;
  ZEXTSINGLE *s;
  ZRECORDLIST RecordList(ES_TS_NOTTOKEEP_TAG);

  cout << "Enter data to index : ";

  cin.getline(buf, sizeof(buf));
  if(*buf == '\0')
    return GDT_FALSE;

  strcat(buf, "\n");

  //  s = new ZEXTSINGLE(SUTRS_OID, buf, GDT_FALSE);
  s = new ZEXTSINGLE(c_recsyntax, buf, GDT_FALSE);

  cout << "Enter the record id to associate with the data : ";

  cin.getline(buf, sizeof(buf));
  if(*buf == '\0')
    return GDT_FALSE;

  ur = new ZESUPDATERECORD(buf, s);
  RecordList.AddRecord(ur);

  if( !ZCLIENT_BASE::Add_Records(c_database, &RecordList) )
    return GDT_FALSE;

  return GDT_TRUE;
}


GDT_BOOLEAN 
IZCLIENT::Delete_Record()
{
  CLEAR_SCREEN();
  DisplayName();
  cout << "Delete Record" << endl << endl;
  cout << "Enter Database [" << c_database << "]: ";

  CHR buf[132];

  cin.getline(buf, sizeof(buf));
  if(*buf == '\0') {
    if(c_database == "")
      c_database = "xxdefault";
  } else c_database = buf;

  ZESUPDATERECORD *ur;
  ZEXTSINGLE *s;
  ZRECORDLIST RecordList(ES_TS_NOTTOKEEP_TAG);

  cout << "Enter the record id to delete : ";

  cin.getline(buf, sizeof(buf));
  if(*buf == '\0')
    return GDT_FALSE;
  
  //  s = new ZEXTSINGLE(SUTRS_OID, "", GDT_FALSE);
  s = new ZEXTSINGLE(c_recsyntax, "", GDT_FALSE);

  ur = new ZESUPDATERECORD(buf, s);
  RecordList.AddRecord(ur);

  if( !ZCLIENT_BASE::Delete_Records(c_database, &RecordList) )
    return GDT_FALSE;

  return GDT_TRUE;
}


GDT_BOOLEAN 
IZCLIENT::Connect()
{
  CLEAR_SCREEN();
  DisplayName();

  cout << "Connect to Service" << endl << endl;
  cout << "Enter Host Name [localhost]: ";
  CHR buf[64];
  cin.getline(buf, sizeof(buf));
  if(*buf == '\0')
    strcpy(buf, "localhost");
  c_host = buf;
  cout << "Enter Port [210]: ";
  cin.getline(buf, sizeof(buf));
  if(*buf == '\0')
    strcpy(buf, "210");
  c_port = atoi(buf);

  //  if(!Initialize(c_host, c_port)) {
  if(!Initialize(c_host, c_port,"",c_username,c_password)) {
    cout << "Failed to initialize to " << c_host << "[";
    cout << c_port << "]" << endl;
    cout << "Press Enter to Continue...";
    cin.get(c);
  }
  c_state = INIT;
  return GDT_TRUE;
}


void 
IZCLIENT::DisplayName()
{
  cout << "CNIDR izclient " << VERS;
  cout << ", Copyright (c) 1995,1996,1997 CNIDR";
  if(c_state != CLOSED) {
    cout << " - " << c_host;
    if(c_hitcount > -1)
      cout << " - " << c_hitcount << " hits";
  } else
    cout << " - No Connection";
  cout << endl;
  cout << "_____________________________________________________________________________";
  cout << endl << endl;
}


GDT_BOOLEAN 
IZCLIENT::DisplayMainMenu()
{
  CLEAR_SCREEN();
  DisplayName();
	
  cout << "\t" << "R\t\t- Release Notes" << endl << endl; 
  cout << "\t" << "U\t\t- Login Info (if required)" << endl << endl;
  cout << "\t" << "C\t\t- Connect to Service" << endl << endl;
  cout << "\t" << "S\t\t- Search" << endl << endl;
  cout << "\t" << "A\t\t- Add Record" << endl << endl;
  cout << "\t" << "E\t\t- Erase Record" << endl << endl;
  cout << "\t" << "O\t\t- Options" << endl << endl;
  if(GetState() != CLOSED) {
    cout << "\t" << "D\t\t- Print Connection Details";
    cout << endl << endl;
  } else {
    cout << endl;
    cout << endl;
  }
  cout << "\t" << "Q\t\t- Quit" << endl << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << endl;
	
  cout << "Enter Selection: ";

  CHR buf[64];
  cin.getline(buf, sizeof(buf));

  switch(toupper(buf[0])) {
  case 'D':
    if(GetState() == CLOSED)
      break;
    ConnectionDetails();
    break;
  case 'R':
    ReleaseNotes();
    break;
  case 'O':
    Configure();
    break;
  case 'U':
    UserInfo();
    break;
  case 'C':
    if(!Connect())
      return GDT_FALSE;
    break;
  case 'S':
    if(GetState() == CLOSED) {
      if(!Connect()) {
	break;
      }
    }
    Search();
    break;
  case 'A':
    if(GetState() == CLOSED) {
      if(!Connect()) {
	break;
      }
    }
    Add_Record();
    break;
  case 'E':
    if(GetState() == CLOSED) {
      if(!Connect()) {
	break;
      }
    }
    Delete_Record();
    break;
  case 'Q':
    return GDT_FALSE;
  default:
    return GDT_FALSE;
  }
  return GDT_TRUE;
}


IZCLIENT::IZCLIENT()
{
  c_display_start = 1;
  c_display_count = 8;
  c_state = CLOSED;	
  c_database = "xxdefault";
  c_term = "";
  c_hitcount = -1;
  c_recsyntax = "";
  c_attrset = BIB1_ATTRSET_OID;
  c_esname = "F";
  c_maxhits=MAX_HITS;
  c_username = "";
  c_password = "";
}


IZCLIENT::~IZCLIENT()
{
}


int 
main(int argc, char **argv)
{
  IZCLIENT client;

  if (!setlocale(LC_CTYPE,"")) {
    cout << "Warning: Failed to set the locale!" << endl;
  }

  for(;;) {
    if(!client.DisplayMainMenu())
      break;
  }
  
  return 0;
}


void 
DisplayHits(INT4 TotalHits, ZRECORDLIST & HitList, INT4 Start, INT4 Count)
{
  ZRECORDSCHOICE *Record;
  STRING Data, OID;
  INT4 i;
  if((Start+Count-1) > HitList.GetRecordCount())
    Count = HitList.GetRecordCount() - Start + 1;
  
  CLEAR_SCREEN();
  cout << "CNIDR izclient, Version " << VERS << endl;
  cout << "Copyright (c) 1995,1996,1997 MCNC/CNIDR" << endl;
  cout << "_____________________________________________________________________________";
  cout << endl;
  cout << TotalHits << " total records matched your query" << endl;
  cout << endl;
  for(i=Start-1;i < Count;i++) {
    Record = HitList.GetRecord(i);
    if(!Record)
      continue;
    Record->GetOID(&OID);
    Record->GetRecordData(&Data);
    cout << i+1 << ") ";
    if(OID == USMARC_OID) {
      MARC *m;
      m = new MARC(Data);
      m->SetDisplayFormat(MARC_FORMAT_TITLE);
      m->SetDisplayWidth(70);
      m->Print();
      delete m;
    } else
      cout << Data << endl;
  }
}
