/* $Id: zbatch.cxx,v 1.4 1998/02/26 01:53:02 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery
and Retrieval, 1994.

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby
granted without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact.

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of
noteworthy uses of this software.

3. The names of MCNC and Clearinghouse for Networked Information
Discovery and Retrieval may not be used in any advertising or publicity
relating to the software without the specific, prior written permission
of MCNC/CNIDR.

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF
THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
************************************************************************/

/*@@@
File:           zbatch.cxx
Version:        $Revision: 1.4 $
Description:    Batch-mode command line Z39.50 client
Author:         Archie Warnock (warnock@clark.net), A/WWW Enterprises
Notes:          Adapted from zclient, originally written by Kevin Gamiel
@@@*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <iostream.h>
#include <fstream.h>

#include "defs.hxx"
#include "string.hxx"
#include "vlist.hxx"
#include "strlist.hxx"
#include "common.hxx"

#include "zsession.hxx"
#include "zclibase.hxx"
#include "marc.hxx"

#if defined(_WIN32)
#define CLEAR_SCREEN() system("cls")
#elif defined(_MSDOS)
#define CLEAR_SCREEN() cout << "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
#else
#  define CLEAR_SCREEN() system("clear")
#endif

#define MAXFETCH 20
#define DEBUGLEVEL 4

void 
DisplayHits(INT4 TotalHits, ZRECORDLIST & HitList, INT4 Start, 
	    INT4 Count, GDT_BOOLEAN Quit);

void 
main(int argc, char **argv)
{
  ZCLIENT_BASE    client;
  GDT_BOOLEAN     Quit=GDT_TRUE;
  STRING          Flag;
  INT             ArgNum;
  INT             Fetch=MAXFETCH, iPort, Start=1;
  STRING          Host, Port, DBname, ESN, Query, recsyntax;
  STRING          Buffer, Remarks;
  STRING          CommandFilename;
  STRING          Msg, Nstr, pw, UserID="";
  STRLIST         MsgList;
  ZRECORDLIST     HitList;
  INT4            FetchCount, HitCount;
  ifstream        CommandFile;
  INT4            i, Count;
  STRING          OID, Data, ReturnMsg, Reason;
  CHR             lBuf[20];
  ZRECORDSCHOICE *Record;
  INT             Condition;
  STRING          AttrSet,AttrSetOID;


  cout << "CNIDR zbatch, Version " << VERS << endl;
  cout << "Copyright (c) 1997 MCNC/CNIDR" << endl;
  cout << endl;
  
  if(argc < 3) {
    cout << "Usage: " << argv[0]
	 << " [-n MaxHits, def=20] host port query-filename" << endl;
    cout << "       Each entry in the query file requires 4 lines:"<< endl;
    cout << "             Database name" << endl;
    cout << "             Attribute Set ID (BIB1|GILS|GEO|STAS)" << endl;
    cout << "             Preferred Record Syntax (SUTRS|HTML|...)" << endl;
    cout << "             ElementSetName (B|F)" << endl;
    cout << "             KWAQS query" << endl;
    cout << endl;
    cout << "       Comment lines starting with \"#\" will be echoed." 
	 << endl;
    exit(1);
  }

  if (!setlocale(LC_CTYPE,"")) {
    cout << "Warning: Failed to set the locale!" << endl;
  }

  client.SetDebugLevel(DEBUGLEVEL);

  ArgNum = 1;

  if (argv[ArgNum][0] == '-') {
    Flag = argv[ArgNum];
    if (Flag.Equals("-n")) {
      if (++ArgNum >= argc) {
	cerr << "ERROR: No value specified after -n." << endl << endl;
	exit(1);
      }
      Fetch = atoi(argv[ArgNum]);
      ArgNum++;
    }
  }

  Nstr="";
  //pw="pto1/pto1";
  pw="";

  Host=argv[ArgNum++];
  Port=argv[ArgNum++];
  iPort=Port.GetInt();
  CommandFilename=argv[ArgNum];

  if(!client.Initialize(Host, iPort,Nstr,Nstr,pw)) {
    INT4 i;
    client.GetDiagnosticList(&MsgList);	
    for(i=1;i <= MsgList.GetTotalEntries();i++) {
      MsgList.GetEntry(i, &Msg);
      cout << "Error: " << Msg << endl;
    }
    exit(1);
  }

  client.GetMOTD(&Msg);	
  if (Msg.GetLength() > 0)
    cout << "Message of the day: " << Msg << endl;

  // Start reading the command file

  //  CommandFile.open(CommandFilename);
  CommandFile.open(argv[ArgNum]);
  if(!CommandFile) {
    cout << "Error: Can't open " << CommandFilename << endl;
    exit(1);
  }

  while (!CommandFile.eof()) {

    Remarks = "";
    //    CommandFile >> DBname;
    CommandFile >> Buffer;
    while (Buffer.Search('#') == 1) {
      Remarks.Cat(Buffer);
      Remarks.Cat("\n");
      CommandFile >> Buffer;
    }
    DBname = Buffer;

    // There might be a stray \n at the end of the file, so let's see 
    // if we got a real string
    if (DBname.GetLength()<=0)
      break;

    //    CommandFile >> Attribute Set
    CommandFile >> Buffer;
    while (Buffer.Search('#') == 1) {
      Remarks.Cat(Buffer);
      Remarks.Cat("\n");
      CommandFile >> Buffer;
    }
    AttrSet = Buffer;

    //    CommandFile >> recsyntax;
    CommandFile >> Buffer;
    while (Buffer.Search('#') == 1) {
      Remarks.Cat(Buffer);
      Remarks.Cat("\n");
      CommandFile >> Buffer;
    }
    recsyntax = Buffer;

    //    CommandFile >> ESN;
    CommandFile >> Buffer;
    while (Buffer.Search('#') == 1) {
      Remarks.Cat(Buffer);
      Remarks.Cat("\n");
      CommandFile >> Buffer;
    }
    ESN = Buffer;

    //    CommandFile >> Query;
    CommandFile >> Buffer;
    while (Buffer.Search('#') == 1) {
      Remarks.Cat(Buffer);
      Remarks.Cat("\n");
      CommandFile >> Buffer;
    }
    Query = Buffer;

    if (recsyntax.CaseEquals("SUTRS"))
      recsyntax = SUTRS_OID;
    else if (recsyntax.CaseEquals("TEXT"))
      recsyntax = SUTRS_OID;
    else if (recsyntax.CaseEquals("HTML"))
      recsyntax = HTML_OID;
    else if (recsyntax.CaseEquals("USMARC"))
      recsyntax = USMARC_OID;

    if (AttrSet.CaseEquals("stas"))
      AttrSetOID = STAS_ATTRSET_OID;
    else if (AttrSet.CaseEquals("gils"))
      AttrSetOID = GILS_ATTRSET_OID;
    else if (AttrSet.CaseEquals("geo"))
      AttrSetOID = GEO_ATTRSET_OID;
    else
      AttrSetOID = BIB1_ATTRSET_OID;

    if (!client.Search(DBname, Query,"Default", AttrSetOID, &HitCount,
		       UserID,Start,Fetch,&HitList)) {
      cout << "Failed to search" << endl;
      break;
    }

    if (HitCount == 0) {
      Count = HitList.GetRecordCount();

      if (Count > 0) {
	
	for (i = 0; i < Count; i++) {
	  Record = HitList.GetRecord(i);
	  if (!Record)
	    continue;
	  Record->GetOID(&OID);
	  Record->GetRecordData(&Data);
	  Record->GetRecordCondition(&Condition);
	  client.MapError(Condition,&Reason);
	  sprintf(lBuf,"%d ",Condition);
	  Msg="Server Replied with Error Code: ";
	  Msg.Cat(lBuf);
	  Msg.Cat("-- ");
	  Msg.Cat(Reason);
	  Msg.Cat(" - ");
	  Msg.Cat(Data);
	  ReturnMsg = "";
	  if (recsyntax.Equals(HTML_OID)) {
	    ReturnMsg.Cat("<HTML>\n<HEAD>\n");
	    ReturnMsg.Cat("<TITLE>Error</TITLE>\n");
	    ReturnMsg.Cat("</HEAD>\n<BODY>");
	    ReturnMsg.Cat("<STRONG>");
	  }

	  ReturnMsg.Cat("ERROR: ");
	  ReturnMsg.Cat(Msg);

	  if (recsyntax.Equals(HTML_OID)) {
	    ReturnMsg.Cat("</STRONG>");
	    ReturnMsg.Cat("</BODY>\n</HTML>\n");
	  }
	  cout << endl;
	  ReturnMsg.Print();
	  cout << endl;
	}
      }
    }
  
    if(HitCount <= Fetch)
      FetchCount = HitCount;
    else
      FetchCount = Fetch;

    if (Remarks.GetLength() > 0)
      cout << Remarks;

    cout << "For database " << DBname << " from host " << Host << endl;
    cout << HitCount << " total records matched your query " 
	 << Query << endl;

    if(HitList.GetRecordCount() <= 0) {
      cout << "No records returned!" << endl;
    } else
      DisplayHits(HitCount, HitList, 1, FetchCount,Quit);

    cout << "*******************************************************" << endl;
    HitList.Clear();
  }

  exit(0);

}


void 
DisplayHits(INT4 TotalHits, ZRECORDLIST & HitList, INT4 Start, 
	    INT4 FetchCount,GDT_BOOLEAN Quit)
{
  ZRECORDSCHOICE *Record;
  STRING Data, OID;
  INT4 i;

  if((Start+FetchCount-1) > TotalHits)
    FetchCount = TotalHits - Start + 1;

  for(i=Start-1;i < FetchCount;i++) {
    Record = HitList.GetRecord(i);
    if(!Record)
      continue;
    Record->GetOID(&OID);
    Record->GetRecordData(&Data);
    cout << i+1 << ") ";
    if(OID == USMARC_OID) {
      MARC *m;
      m = new MARC(Data);
      m->SetDisplayFormat(MARC_FORMAT_TITLE);
      m->SetDisplayWidth(70);
      m->Print();
      delete m;
    } else
      cout << Data << endl;
  }
}
