/* $Id: zclibase.cxx,v 1.7 1997/11/13 01:54:09 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery and
Retrieval, 1994. 

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact. 

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or Extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software. 

3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR. 

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
************************************************************************/

/*@@@
File:		zclibase.cxx
Version:	1.00
Description:	Z39.50 client API
Author:		Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include "defs.hxx"
#include "string.hxx"
#include "vlist.hxx"
#include "strlist.hxx"
#include "common.hxx"

#include "zsession.hxx"
//#include "config.hxx"
#include "zclibase.hxx"

ZCLIENT_BASE::ZCLIENT_BASE() : ZSESSION()
{
	c_hitcount = 0;
	c_error = 0;
}


ZCLIENT_BASE::~ZCLIENT_BASE()
{
}


GDT_BOOLEAN 
ZCLIENT_BASE::Initialize(const STRING &Host, const UINT Port, STRING Pass)
{
  const char* IsiteVersion = VERS;
  return Initialize(Host, Port, "",ZPROTVER,c_options,1048576,1048576,
		    "","",Pass, "34","zclient_base",IsiteVersion);
}


GDT_BOOLEAN 
ZCLIENT_BASE::Initialize(const STRING &Host, const UINT Port)
{
  const char* IsiteVersion = VERS;
  return Initialize(Host, Port, "",ZPROTVER,c_options,1048576,1048576,
		    "","","", "34","zclient_base",IsiteVersion);
}


GDT_BOOLEAN 
ZCLIENT_BASE::Initialize(const STRING &Host, 
			 const UINT Port,
			 const STRING & GroupId, 
			 const STRING & UserId, 
			 const STRING & Password)
{
  const char* IsiteVersion = VERS;
  return Initialize(Host, Port, "",ZPROTVER, c_options,1048576,1048576,
		    GroupId, UserId, Password, "34","zclient_base",
		    IsiteVersion);
}


GDT_BOOLEAN 
ZCLIENT_BASE::Initialize(const STRING &Host, 
			 const UINT Port,
			 const STRING &RefId,
			 const STRING &ProtVer,
			 const STRING &Options,
			 const INT4 PrefMsgSize,
			 const INT4 ExceptionalRecSize,
			 const STRING &GroupId,
			 const STRING &UserId,
			 const STRING &Password,
			 const STRING &ImpId,
			 const STRING &ImpName,
			 const STRING &ImpVersion)
{
  CHR msg[512], *tmpbuf;
  UCHR *buf;
  ZINITREQUEST *Request;
  STRING Value;
  INT4 NumValue;

  c_error = 0;

  if(c_tcp != NULL)
    delete c_tcp;

  c_tcp = new TCPSOCK();

  tmpbuf = Host.NewCString();
  c_tcp->Open(tmpbuf, Port);
  delete [] tmpbuf;
  if((c_error = c_tcp->LastError())) {
    c_tcp->ErrorMessage(c_error, msg, (sizeof(msg) - 1));
    c_error_msg = msg;
    if(c_debuglevel > 0) {
      cerr << "Initialize:" << msg << endl;
    }
    return GDT_FALSE;
  }
  if(c_debuglevel >= 5) {
    cerr << "Connection open to " << Host << "[" << Port << "]";
    cerr << endl;
  }
	
  Request = new ZINITREQUEST(RefId,ProtVer,Options,PrefMsgSize,
			     ExceptionalRecSize, GroupId, UserId, Password,
			     ImpId, ImpName, ImpVersion);

  if(!SendPDU(Request)) {
    delete Request;
    return GDT_FALSE;
  }

  delete Request;

  //
  // Read the response PDU
  //
  INT4 len;
  unsigned short type;
  if((buf = ReadPDU(&len, &type)) == NULL)  {
    c_tcp->Close();
    return GDT_FALSE;
  }
  
  //
  // Is it an Init Response?
  //
  if(type != INITRESPONSE_TAG) {
    c_tcp->Close();
    delete [] buf;
    if(c_debuglevel > 0) {
      cerr << "Expected init response, got " << type << endl;
    }
    c_error = 9999;
    c_error_msg = "Expected INITIALIZE_RESPONSE";
    return GDT_FALSE;
  }
  
  //
  // Did they accept our Init Request?
  //
  ZINITRESPONSE *Response;
  Response = new ZINITRESPONSE(buf, len);
  delete [] buf;	
  GDT_BOOLEAN return_val = GDT_FALSE;
  if(Response->HasResult()) {
    INT4 num;
    Response->GetResult(&num);
    if(num) {
      c_state = open;
      return_val = GDT_TRUE;
    } else {
      c_error = ERR_INIT_DENIED;
      c_error_msg = "Initialization Request denied";
      c_tcp->Close();
      return_val = GDT_FALSE;
    }
  } else {
    c_error = 9997;
    c_error_msg = "Protocol Error: No Result Tag in Init Response";
    c_tcp->Close();
    return_val = GDT_FALSE;
  }
  
  //
  // What options do they support?
  //
  if(Response->HasOptions()) {
    Response->GetOptions(&Value);
    SetOptions(Value);
  } else SetOptions(ZOPTIONS);
  
  //
  // What protocol version will we be using?
  //
  if(Response->HasProtocolVersion()) {
    Response->GetProtocolVersion(&Value);
    SetProtocolVersion(Value);
  } else SetProtocolVersion(ZPROTVER);
  
  //
  // Message sizes
  //
  if(Response->HasExceptionalRecordSize()) {
    Response->GetExceptionalRecordSize(&NumValue);
    SetExceptionalRecordSize(NumValue);
  } else SetExceptionalRecordSize(ZMAXRECORDSIZE);
  
  if(Response->HasPreferredMessageSize()) {
    Response->GetPreferredMessageSize(&NumValue);
    SetPreferredMessageSize(NumValue);
  } else SetPreferredMessageSize(ZPREFMSGSIZE);
  
  if(c_debuglevel >= 5)
    PrintConnectionDetails();
  
  delete Response;
  
  return return_val;
}


GDT_BOOLEAN 
ZCLIENT_BASE::Search(const STRING &Database, 
		     const STRING &Term,
		     const STRING &ResultSetName, 
		     INT4 *HitCount,
		     STRING userId, 
		     INT Start, 
		     INT Fetch,
		     ZRECORDLIST *Records) 
{
  GDT_BOOLEAN result;
  result = Search(Database, Term, ResultSetName, BIB1_ATTRSET_OID,
		  HitCount, userId, Start, Fetch, Records);
  return result;
}


GDT_BOOLEAN 
ZCLIENT_BASE::Search(const STRING &Database, 
		     const STRING &Term,
		     const STRING &ResultSetName, 
		     const STRING& AttrSetId,
		     INT4 *HitCount,
		     STRING userId, 
		     INT Start, 
		     INT Fetch,
		     ZRECORDLIST *Records) 
{
  ZSEARCHREQUEST *s;
  INT4 len;
  PUCHR buf;
  INT SmallUpper, LargeLower=1000000, MediumPresent;
 

  *HitCount = 0;

#ifdef NOPIG
  Start=1;
  Fetch=0;
  LargeLower=1;
#endif

  if(Start>1){
    SmallUpper=0;
    LargeLower=1;
  }else
    SmallUpper=0;

  MediumPresent=Fetch;
  if (c_debuglevel >=5)
    cout << "Search: " << Database << " " << Term << " " <<ResultSetName<<endl;
  /*
  s = new ZSEARCHREQUEST("",SmallUpper,LargeLower,MediumPresent,1,
			 ResultSetName,Database,"","",SUTRS_OID, Term, 
			 BIB1_ATTRSET_OID,userId);
			 */
  s = new ZSEARCHREQUEST("",SmallUpper,LargeLower,MediumPresent,1,
			 ResultSetName,Database,"","","", Term, 
			 AttrSetId,userId);
  
  if(!SendPDU(s)) {
    delete s;
    return GDT_FALSE;
  }
  delete s;
  
  unsigned short type;
  if((buf = ReadPDU(&len, &type)) == NULL)  {
    c_tcp->Close();
    return GDT_FALSE;
  }
  
  if(type != SEARCHRESPONSE_TAG) {
    c_tcp->Close();
    delete [] buf;
    if(c_debuglevel > 0) {
      cerr << "Expected search response, got " << type << endl;
    }
    return GDT_FALSE;
  }
  
  ZSEARCHRESPONSE *r;
  r = new ZSEARCHRESPONSE(buf, len);
  delete [] buf;
  r->GetResultCount(HitCount);
  c_hitcount=*HitCount;
  c_dbname=Database;
  BERBROWSER *b;
  b = new BERBROWSER(r);
  GetRecordsOrDiagnostics(b,Records);
  
  delete r;
  delete b;
  return GDT_TRUE;	
}


GDT_BOOLEAN 
ZCLIENT_BASE::Sort(const STRING &InputResultSetNames,
		   const STRING &SortedResultSetName,
		   const STRING &SortSequence)
{
  ZSORTREQUEST *s;
  INT4 len;
  INT4 *Status;
  PUCHR buf;
  
  /*
    s = new ZSORTREQUEST("",InputResultSetNames,SortedResultSetName,
    SortSequence);
     
     
    if(!SendPDU(s)) {
    delete s;
    return GDT_FALSE;
    }
    delete s;
     
    unsigned short type;
    if((buf = ReadPDU(&len, &type)) == NULL)  {
    c_tcp->Close();
    return GDT_FALSE;
    }
     
    if(type != SORTRESPONSE_TAG) {
    c_tcp->Close();
    delete [] buf;
    if(c_debuglevel > 0) {
    cerr << "Expected sort response, got " << type << endl;
    }
    return GDT_FALSE;
    }
     
    ZSORTRESPONSE *r;
    r = new ZSORTRESPONSE(buf, len);
    delete [] buf;
    r->GetSortStatus(Status);
     
    delete r;
    if (Status == 0)
    return GDT_TRUE;	
    else
  */
  return GDT_FALSE;	
  
}

/*
   GDT_BOOLEAN ZCLIENT_BASE::Present(const INT4 Start, const INT4 Count,
   const STRING &PreferredRecordSyntax, 
   const STRING &ElementSetName,
   ZRECORDLIST *Records)
   {
   STRING RSN;
   RSN="Default";
   
   return(Present(Start, Count, PreferredRecordSyntax, RSN,
   ElementSetName, Records));
   }
   */


GDT_BOOLEAN 
ZCLIENT_BASE::Present(const INT4 Start, 
		      const INT4 Count,
		      const STRING &PreferredRecordSyntax, 
		      const STRING &ResultSetName,
		      const STRING &ElementSetName,
		      ZRECORDLIST *Records)
{
  ZPRESENTREQUEST *s;
  INT4 len;
  PUCHR buf;
  FILE *fp;
  CHR t1[256];
  INT St,Ct;
  St=Start;
  Ct=Count;
  
  ElementSetName.GetCString(t1,256);
  
  while(Ct>5){
    Present(St,5,PreferredRecordSyntax,
	    ResultSetName,ElementSetName,Records);
    St+=5;
    Ct-=5;
  }
  s = new ZPRESENTREQUEST("", ResultSetName, St, Ct,
			  ElementSetName, PreferredRecordSyntax);
  //			  ElementSetName, SUTRS_OID);
  
  if(!SendPresentPDU(s)) {
    delete s;
    return GDT_FALSE;
  }
  
  delete s;
  unsigned short type;
  if((buf = ReadPresentPDU(&len, &type)) == NULL)  {
    c_tcp->Close();
    return GDT_FALSE;
  }
  if(type != PRESENTRESPONSE_TAG) {
    c_tcp->Close();
    delete [] buf;
    if(c_debuglevel > 0) {
      cerr << "Expected present response, got " << type << endl;
    }
    return GDT_FALSE;
  }
  
  //
  // Build a ZRECORDLIST from the BER
  //
  ZPRESENTRESPONSE *rsp;
  rsp = new ZPRESENTRESPONSE(buf, len);
  delete [] buf;
  
  // How many records returned?
  INT4 i;
  rsp->GetNumberOfRecordsReturned(&i);
  
  if (i == 0)
    // Everything is OK, but no records returned
    return GDT_TRUE;
  BERBROWSER *b;
  b = new BERBROWSER(rsp);
  GetRecordsOrDiagnostics(b,Records);
  delete b;
  delete rsp; // this was leaking
  return GDT_TRUE;	
}


void 
ZCLIENT_BASE::GetErrorMessage(STRING *Message) {
  if(c_error == 0)
    *Message = "No Error";
  *Message = c_error_msg;
}


//*********************************************************************
// In zclibase.cxx:
//   Decoding of Scan Entry of is added to Scan method.
//
//*********************************************************************
GDT_BOOLEAN 
ZCLIENT_BASE::Scan(const CHR *DatabaseName,
		   const CHR *Term,
		   const INT PrefPos,
		   const INT StepSize,
		   const INT NumRequest,
		   ZRECORDLIST *Records)
{
  GDT_BOOLEAN result;
  result = Scan(DatabaseName, Term, STAS_ATTRSET_OID, PrefPos, StepSize,
		NumRequest, Records);
  return result;
}


GDT_BOOLEAN 
ZCLIENT_BASE::Scan(const CHR *DatabaseName,
		   const CHR *Term,
		   const CHR *AttributeSet,
		   const INT PrefPos,
		   const INT StepSize,
		   const INT NumRequest,
		   ZRECORDLIST *Records)
{
  ZSCANREQUEST *s;
  INT4 len;
  PUCHR buf;
  FILE *log;
  s = new ZSCANREQUEST("", DatabaseName, AttributeSet, Term, StepSize,
		       NumRequest, PrefPos);
  
  if(!SendPDU(s))
    return GDT_FALSE;
  unsigned short type;
  if((buf = ReadPDU(&len, &type)) == NULL)  {
    c_tcp->Close();
    return GDT_FALSE;
  }
  if(type != SCANRESPONSE_TAG) {
    c_tcp->Close();
    delete [] buf;
    if(c_debuglevel > 0) {
      cerr << "Expected scan response, got " << type << endl;
    }
    return GDT_FALSE;
  }
  
  ZSCANRESPONSE *r;
  r = new ZSCANRESPONSE(buf, len);
  delete [] buf;
  delete s;
  
  // How many records returned?
  
  INT4 RetStepSize;
  r->GetRetStepSize(&RetStepSize);
  INT4 ScanStatus;
  r->GetScanStatus(&ScanStatus);
  INT4 PosOfTerm;
  r->GetPositionOfTerm(&PosOfTerm);
  
  INT4 i;
  r->GetNumberOfEntriesReturned(&i);
  if(i == 0)			// Not OK
    return GDT_FALSE;
  
  BERBROWSER *b;
  b = new BERBROWSER(r);
  if(!b->HasTag(ENTRIES_TAG)) {
    return GDT_FALSE;
  }
  STRING Data;
  b->GetSubdirectory(ENTRIES_TAG, b);
  // Do we have response entries or nonsurrogatediag
  if (b->HasTag(ENTRY_TAG)) {
    // We have real entries
    Records->SetRecordType(ENTRY_TAG);
    BERBROWSER *sub;
    sub = new BERBROWSER(*b);
    INT j=0;
    b->GetSubdirectory(ENTRY_TAG, b);
    INT count = b->GetSubdirectoryCount();
    for(j=0;j < count;j++) {
      b->GetSubdirectoryX(j, sub);
      INT4    Occ;
      ASN1TAG Tag;
      sub->GetTag(&Tag);
      if (Tag==ASN1TAG(TERMINFO_TAG)) {
	sub->GetSubdirectory(TERMINFO_TAG, sub);
	sub->GetChar(TERMGENERAL_TAG, &Data);
	//Data.UpperWord();
	sub->GetInt(GLOBALOCCURRENCES_TAG, &Occ);
	ZENTRY *s;
	s = new ZENTRY(Data, Occ);
	ZNAMEPLUSRECORD *npr;
	npr = new ZNAMEPLUSRECORD(DatabaseName, s);
	Records->AddRecord(npr);
      } else if (Tag==ASN1TAG(SURROGATEDIAGNOSTIC_TAG)) {
	if(c_debuglevel > 0) {
	  cerr << "Surrogatediagnostic rec in TermInfo" << endl;
	}
      } else  {
	//cerr << "Unknown tag in TermInfo" << endl;
      }
    }
  } else if(b->HasTag(NONSURROGATEDIAGNOSTICS_TAG)) {
    // Diagnostic
    if(c_debuglevel > 0) {
      cerr << "NonSurrogatediagnostic rec in Entry" << endl;
    }
  } else {
    // cerr << "Unknown tag in Entry" << endl;
  }
  delete b;
  if (ScanStatus==6) return GDT_FALSE;
  else return GDT_TRUE;
}


void 
ZCLIENT_BASE::GetRecordsOrDiagnostics(BERBROWSER *b, ZRECORDLIST *Records)
{
  if(b->HasTag(NONSURROGATEDIAGNOSTIC_TAG)) {
    cout << "Diagnostic returned" << endl;
    Records->SetRecordType(NONSURROGATEDIAGNOSTIC_TAG);
    ZDEFAULTDIAGFORMAT *rec;
    // We have nsd records
    BERBROWSER *sub;
    sub = new BERBROWSER(*b);
    ASN1TAGU Tag(ASN1_EXTERNAL);
    b->GetSubdirectory(NONSURROGATEDIAGNOSTIC_TAG, b);
    
    STRING  diagnosticSetID, addinfo;
    INT4 Condition;
    Tag.Number = ASN1_OBJECTIDENTIFIER;
    b->GetOID(Tag, &diagnosticSetID);
    Tag.Number = ASN1_INTEGER;
    b->GetInt(Tag, &Condition);
    ASN1TAGU IntTag(26);
    b->GetChar(IntTag, &addinfo);
    rec = new ZDEFAULTDIAGFORMAT(Condition, addinfo, diagnosticSetID, 2);
    Records->AddRecord(rec);
    //  } else if(b->HasTag(SURROGATEDIAGNOSTIC_TAG)) {
    // put errors in the record list
  }else if(b->HasTag(MULTIPLENONSURDIAGNOSTICS_TAG)) {
    // Multiple
    cout << "Multiple Diagnostics returned" << endl;
    
  } else if(b->HasTag(RESPONSERECORDS_TAG)) {
    // We have real records
    BERBROWSER *sub;
    sub = new BERBROWSER(*b);
    INT4 j, count;
    b->GetSubdirectory(RESPONSERECORDS_TAG, b);
    count = b->GetSubdirectoryCount();
    STRING DBName, OID, Data;
    for(j=0;j < count;j++) {
      ZNAMEPLUSRECORD *npr;
      b->GetSubdirectoryX(j, sub);
      sub->GetChar(0, &DBName);
      //
      // What kind of record?
      //
      sub->GetSubdirectory(1, sub);
      if(sub->HasTag(RETRIEVALRECORD_TAG)) {
	ASN1TAGU Tag(ASN1_EXTERNAL);
	sub->GetSubdirectory(1, sub);
	sub->GetSubdirectory(Tag, sub);
	Tag.Number = ASN1_OBJECTIDENTIFIER;
	sub->GetOID(Tag, &OID);
	
	// USMARC or SUTRS?
	if(OID == USMARC_OID) {
	  sub->GetChar(1, &Data);
	  ZUSMARCRECORD *s;
	  s = new ZUSMARCRECORD(Data);
	  npr = new ZNAMEPLUSRECORD(DBName, s);
	  Records->AddRecord(npr);
	} if(OID == SUTRS_OID) {
	  //
	  // SUTRS isn't encoded very
	  // consistently across servers,
	  // unfortunately.  I have to test
	  // for every case here.
	  //
	  if(sub->HasTag(0)) {
	    ASN1TAGU IntTag(27);
	    sub->GetSubdirectory(0, sub);
	    sub->GetChar(IntTag, &Data);
	  } else if(sub->HasTag(1)) {
	    sub->GetChar(1, &Data);
	  }
	  ZSUTRSRECORD *s;
	  s = new ZSUTRSRECORD(Data);
	  npr = new ZNAMEPLUSRECORD(DBName, s);
	  if (c_debuglevel > 5)
	    cout << "Adding a record: db = "<<c_dbname << endl;
	  CHR *cc=c_dbname.NewCString();
	  Records->AddRecord(npr,cc);
	  delete cc;
	}
      } 
    }
    delete sub;
  }
}

static CHR *List[]={
  
  "Permanent system error",
  "Temporary system error",
  "Unsupported search",
  "Terms only exclusion (stop) words",
  "Too many argument words",
  "Too many boolean operators",
  "Too many truncated words",
  "Too many incomplete subfields",
  "Truncated words too short",
  "Invalid format for record number (search term)",
  "Too many characters in search statement",
  "Too many records retrieved",
  "Present request out of range",
  "System error in presenting records",
  "Record no authorized to be sent intersystem",
  "Record exceeds Preferred-message-size",
  "Record exceeds Maximum-record-size",
  "Result set not supported as a search term",
  "Only single result set as search term supported",
  "Only ANDing of a single result set as search term supported",
  "Result set exists and replace indicator off",
  "Result set naming not supported",
  "Combination of specified databases not supported",
  "Element set names not supported",
  "Specified element set name not valid for specified database",
  "Only a single element set name supported",
  "Result set no longer exists - unilaterally deleted by target",
  "Result set is in use",
  "One of the specified databases is locked",
  "Specified result set does not exist",
  "Resources exhausted - no results available",
  "Resources exhausted - unpredictable partial results available",
  "Resources exhausted - valid subset of results available",
  " ", " ", " ", " ", " ", " ", " ", " ", " ", " ",
  " ", " ", " ", " ", " ", " ", " ", " ", " ", " ",
  " ", " ", " ", " ", " ", " ", " ", " ", " ", " ",
  " ", " ", " ", " ", " ", " ", " ", " ", " ", " ",
  " ", " ", " ", " ", " ", " ", " ", " ", " ", " ",
  " ", " ", " ", " ", " ", " ", " ", " ", " ", " ", 
  " ", " ", " ", " ", " ", " ",
  
  "Unspecified error",
  "Access-control failure",
  "Security challenge required but could not be issued - request terminated",
  "Security challenge required but could not be issued - record not included",
  "Security challenge failed - record not included",
  "Terminated by negative continue response",
  "No abstract syntaxes agreed to for this record",
  "Query type not supported",
  "Malformed query",
  "Database unavailable",
  "Operator unsupported",
  "Too many databases specified",
  "Too many result sets created",
  "Unsupported attribute type",
  "Unsupported Use attribute",
  "Unsupported value for Use attribute",
  "Use attribute required but not supplied",
  "Unsupported Relation attribute",
  "Unsupported Structure attribute",
  "Unsupported Position attribute",
  "Unsupported Truncation attribute",
  "Unsupported Attribute Set",
  "Unsupported Completeness attribute",
  "Unsupported attribute combination",
  "Unsupported coded value for term",
  "Malformed search term",
  "Illegal term value for attribute",
  "Unparsable format for un-normalized value",
  "Illegal result set name",
  "Proximity search of sets not supported",
  "Illegal result set in proximity search",
  "Unsupported proximity relation",
  "Unsupported proximity unit code"};


void 
ZCLIENT_BASE::MapError(INT Code, STRING *Message)
{
  STRING b;
  CHR M[256];
  
  
  if(Code>132)
    *Message="Unknown Error";
  else{
    strcpy(M,List[Code-1]);
    b=M;
    b.Print();
    *Message=b;
  }
  
  
}


GDT_BOOLEAN 
ZCLIENT_BASE::Add_Records(const STRING &Database,
			  ZRECORDLIST *Records)
{
  INT4 len;
  PUCHR buf;
  ZESUPDATEREQUEST *s;
  unsigned short type;

  s = new ZESUPDATEREQUEST("", ES_UPDATE_RECORD_INSERT_ACTION,
			   Database, Records);

  if(!SendPDU(s)) {
    delete s;
    return GDT_FALSE;
  }

  delete s;

  if((buf = ReadPDU(&len, &type)) == NULL)  {
    c_tcp->Close();
    return GDT_FALSE;
  }

  if(type != EXTENDEDSERVICESRESPONSE_TAG) {
    c_tcp->Close();
    delete [] buf;
    if(c_debuglevel > 0) {
      cerr << "Expected extended service response, got " << type << endl;
    }
    return GDT_FALSE;
  }

  ZESUPDATERESPONSE *Response;

  Response = new ZESUPDATERESPONSE(buf, len);
  delete [] buf;

  INT4 OperationStatus;
  INT4 TaskStatus;
  INT4 UpdateStatus;

  if (	!Response->GetOperationStatus(&OperationStatus) ||
      OperationStatus != ES_DONE_OPERATIONSTATUS ||
      !Response->GetTaskStatus(&TaskStatus) ||
      TaskStatus != ES_TP_COMPLETE_TASKSTATUS ||
      !Response->GetUpdateStatus(&UpdateStatus) ||
      UpdateStatus != ES_UPDATE_SUCCESS_UPDATESTATUS) {
    delete Response;
    return GDT_FALSE;
  }

  delete Response;

  return GDT_TRUE;	
}


GDT_BOOLEAN 
ZCLIENT_BASE::Delete_Records(const STRING &Database,
			     ZRECORDLIST *Records)
{
  INT4 len;
  PUCHR buf;
  ZESUPDATEREQUEST *s;
  unsigned short type;

  s = new ZESUPDATEREQUEST("", ES_UPDATE_RECORD_DELETE_ACTION,
			   Database, Records);

  if(!SendPDU(s)) {
    delete s;
    return GDT_FALSE;
  }

  delete s;

  if((buf = ReadPDU(&len, &type)) == NULL)  {
    c_tcp->Close();
    return GDT_FALSE;
  }

  if(type != EXTENDEDSERVICESRESPONSE_TAG) {
    c_tcp->Close();
    delete [] buf;
    if(c_debuglevel > 0) {
      cerr << "Expected extended service response, got " << type << endl;
    }
    return GDT_FALSE;
  }

  ZESUPDATERESPONSE *Response;
  
  Response = new ZESUPDATERESPONSE(buf, len);
  delete [] buf;

  INT4 OperationStatus;
  INT4 TaskStatus;
  INT4 UpdateStatus;

  if (	!Response->GetOperationStatus(&OperationStatus) ||
	OperationStatus != ES_DONE_OPERATIONSTATUS ||
	!Response->GetTaskStatus(&TaskStatus) ||
	TaskStatus != ES_TP_COMPLETE_TASKSTATUS ||
	!Response->GetUpdateStatus(&UpdateStatus) ||
	UpdateStatus != ES_UPDATE_SUCCESS_UPDATESTATUS) {
    delete Response;
    return GDT_FALSE;
  }

  delete Response;
  
  return GDT_TRUE;	
}
