/* $Id: zclient.cxx,v 1.13 1998/01/30 19:13:51 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery
and Retrieval, 1994.

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby
granted without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact.

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of
noteworthy uses of this software.

3. The names of MCNC and Clearinghouse for Networked Information
Discovery and Retrieval may not be used in any advertising or publicity
relating to the software without the specific, prior written permission
of MCNC/CNIDR.

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF
THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
************************************************************************/

/*@@@
File:           zclient.cxx
Version:        2.00
$Revision: 1.13 $
Description:    command line Z39.50 client
Author:         Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <iostream.h>

//#include "config.hxx"
#include "defs.hxx"
#include "string.hxx"
#include "vlist.hxx"
#include "strlist.hxx"
#include "common.hxx"

#include "zsession.hxx"
#include "zclibase.hxx"
#include "marc.hxx"

#if defined(_WIN32)
#define CLEAR_SCREEN() system("cls")
#elif defined(_MSDOS)
#define CLEAR_SCREEN() cout << "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
#else
#  define CLEAR_SCREEN() system("clear")
#endif

#define MAXFETCH 20
#define DEBUGLEVEL 4

void 
DisplayHits(INT4 TotalHits, ZRECORDLIST & HitList, INT4 Start, 
	    INT4 Count, GDT_BOOLEAN Quit);

int 
main(int argc, char **argv)
{
  ZCLIENT_BASE client;
  INT          ArgNum;
  GDT_BOOLEAN  Quit=GDT_FALSE;
  INT          Nhits=MAXFETCH;
  STRING       Flag, Msg, Nstr, pw;
  STRLIST      MsgList;
  INT4         HitCount;
  STRING       a;
  ZRECORDLIST  HitList;
  INT4         FetchCount;
  STRING       recsyntax;
  STRING       AttrSet,AttrSetOID;

  
  cout << "CNIDR zclient, Version " << VERS << endl;
  cout << "Copyright (c) 1995,1996,1997 MCNC/CNIDR" << endl;
  
  if(argc < 5) {
    cout << "Usage: " << argv[0] 
	 << " [-q] [-n MaxHits, def=20] [-d DebugLevel] [-set AttrSet] host port "
	 << "database query [record_syntax]" << endl << endl;
    exit(1);
  }

  if (!setlocale(LC_CTYPE,"")) {
    cout << "Warning: Failed to set the locale!" << endl;
  }

  client.SetDebugLevel(DEBUGLEVEL);

  ArgNum = 1;

  //  while (ArgNum < 4) {
  while (ArgNum < argc) {
    if (argv[ArgNum][0] == '-') {
      Flag = argv[ArgNum];
      if (Flag.Equals("-q")) {
	Quit = GDT_TRUE;
	ArgNum++;
      }
      else if (Flag.Equals("-n")) {
	if (++ArgNum >= argc) {
	  cerr << "ERROR: No value specified after -n." << endl << endl;
	  exit(1);
	}
	Nhits = atoi(argv[ArgNum]);
	ArgNum++;
      }
      else if (Flag.Equals("-d")) {
	if (++ArgNum >= argc) {
	  cerr << "ERROR: No value specified after -d." << endl << endl;
	  exit(1);
	}
	client.SetDebugLevel(atoi(argv[ArgNum]));
	ArgNum++;
      }
      else if (Flag.Equals("-set")) {
	if (++ArgNum >= argc) {
	  cerr << "ERROR: No value specified after -set." << endl << endl;
	  exit(1);
	}
	AttrSet = argv[ArgNum];
	ArgNum++;
      }
    } else
      break;
  }

  Nstr="";
  //  pw="pto1/pto1";
  pw = "";

  //  if(!client.Initialize(argv[ArgNum], atoi(argv[ArgNum+1]),Nstr,Nstr,pw)) {
  if (!client.Initialize(argv[ArgNum], atoi(argv[ArgNum+1]))) {
    INT4 i;
    client.GetDiagnosticList(&MsgList);	
    for(i=1;i <= MsgList.GetTotalEntries();i++) {
      MsgList.GetEntry(i, &Msg);
      cout << "Error: " << Msg << endl;
    }
    exit(1);
  }

  client.GetMOTD(&Msg);	
  if (Msg.GetLength() > 0)
    cout << "Message of the day: " << Msg << endl;

  if(argc > (ArgNum+4))
    recsyntax = argv[ArgNum+4];
  else {
    recsyntax = SUTRS_OID;
  }

  if (recsyntax.CaseEquals("SUTRS"))
    recsyntax = SUTRS_OID;
  else if (recsyntax.CaseEquals("TEXT"))
    recsyntax = SUTRS_OID;
  else if (recsyntax.CaseEquals("HTML"))
    recsyntax = HTML_OID;
  else if (recsyntax.CaseEquals("USMARC"))
    recsyntax = USMARC_OID;

  if (AttrSet.CaseEquals("stas"))
    AttrSetOID = STAS_ATTRSET_OID;
  else if (AttrSet.CaseEquals("gils"))
    AttrSetOID = GILS_ATTRSET_OID;
  else if (AttrSet.CaseEquals("geo"))
    AttrSetOID = GEO_ATTRSET_OID;
  else
    AttrSetOID = BIB1_ATTRSET_OID;

  if (!client.Search(argv[ArgNum+2], argv[ArgNum+3],"Default", AttrSetOID,
		     &HitCount, a, 1, Nhits, &HitList)) {
    cout << "Failed to search" << endl;
    exit(1);
  }

  if (HitCount == 0) {
    INT4            i,Count;
    STRING          OID,Data,ReturnMsg,Reason;
    CHR lBuf[20];
    ZRECORDSCHOICE *Record;
    INT Condition;
    Count = HitList.GetRecordCount();

    if (Count == 0) {
      cout << "No records returned!" << endl;
    } else {
      for (i = 0; i < Count; i++) {
      
	Record = HitList.GetRecord(i);
	if (!Record)
	  continue;
	Record->GetOID(&OID);
	Record->GetRecordData(&Data);
	Record->GetRecordCondition(&Condition);
	client.MapError(Condition,&Reason);
	sprintf(lBuf,"%d ",Condition);
	Msg="Server Replied with Error Code: ";
	Msg.Cat(lBuf);
	Msg.Cat("-- ");
	Msg.Cat(Reason);
	Msg.Cat(" - ");
	Msg.Cat(Data);
	if (recsyntax.Equals(HTML_OID)) {
	  ReturnMsg.Cat("<HTML>\n<HEAD>\n");
	  ReturnMsg.Cat("<TITLE>Error</TITLE>\n");
	  ReturnMsg.Cat("</HEAD>\n<BODY>");
	  ReturnMsg.Cat("<STRONG>");
	}

	ReturnMsg.Cat("ERROR: ");
	ReturnMsg.Cat(Msg);

	if (recsyntax.Equals(HTML_OID)) {
	  ReturnMsg.Cat("</STRONG>");
	  ReturnMsg.Cat("</BODY>\n</HTML>\n");
	}
	cout << endl;
	ReturnMsg.Print();
	cout << endl;
      }
    }
    exit(0);
  }
  
  if(HitCount <= Nhits)
    FetchCount = HitCount;
  else
    FetchCount = Nhits;

  /*
     if(!client.Present(1, FetchCount, recsyntax,"Default", "B", &HitList)) {
     cout << "Failed to present" << endl;
     exit(1);
     }
     */
  if(HitList.GetRecordCount() <= 0) {
    cout << "No records returned!" << endl;
    exit(1);
  }

  DisplayHits(HitCount, HitList, 1, FetchCount,Quit);
  if(Quit)
    exit(0);

  int num;
  ZRECORDLIST RecordList;
  INT4 Count;
  ZRECORDSCHOICE *Record;
  STRING Data, OID;
  CHR cmd[1024];
  CHR *TempFile = tmpnam((CHR*)NULL);

  CHR *Pager;
  Pager = getenv("PAGER");
  if ((Pager) && (strlen(Pager)>0))
    sprintf(cmd,"%s %s",Pager,TempFile);
  else {
#if defined(_MSDOS) || defined(_WIN32)
    sprintf(cmd, "type %s", TempFile);
#else
    sprintf(cmd, "more %s", TempFile);
#endif
  }

  FILE *fp;
  CHR c;
  CHR buf[80];
  MARC *m;
  CHR *buffer;

  for(;;) {
    RecordList.Clear();
    cout << endl << "Enter # of record to view (return to quit): ";
    cin.getline(buf, sizeof(buf));
    if(!buf[0]) {
      cout << endl << endl;
      break;
    }
    num = atoi(buf);
    Count = HitList.GetRecordCount();	
    if((num <= 0) || (num > Count)) {
      cout << endl << endl;
      break;
    }
    if(!client.Present(num, 1, recsyntax, "Default", "F", &RecordList)){
      cout << "Failed to present" << endl;
      exit(1);
    }
    Count = RecordList.GetRecordCount();	

    if(Count < 1) {
      cout << "No records returned!" << endl;
      exit(1);
    }

    Record = RecordList.GetRecord(0);
    if(!Record) {
      cout << "Record not available!" << endl;
      continue;
    }

    Record->GetOID(&OID);
    Record->GetRecordData(&Data);

    if((fp = fopen(TempFile, "w")) == NULL) {
      perror(TempFile);
      exit(1);
    }
    fprintf(fp, "CNIDR zclient, Version %s\n", VERS);
    fprintf(fp, "Copyright (c) 1995,1996,1997 MCNC/CNIDR\n\n");
    fprintf(fp, "Record %d of %d\n\n", num, HitCount);

    if(OID == USMARC_OID) {
      m = new MARC(Data);
      m->Print(fp);
      delete m;
    } else {
      buffer = Data.NewCString();
      fwrite(buffer, 1, Data.GetLength(), fp);
      delete buffer;
    }
    fclose(fp);
    CLEAR_SCREEN();
		
#if defined(_MSDOS) && !defined(_WIN32)
    cout << endl;
    cout << "Sorry...I can't view records in this enviroment!";
#else
    system(cmd);
#endif
    cout << endl << "Press return continue...";
    cin.get(c);
    CLEAR_SCREEN();
    unlink(TempFile);

    DisplayHits(HitCount, HitList, 1, FetchCount,Quit);
  }
  return 0;
}


void 
DisplayHits(INT4 TotalHits, ZRECORDLIST & HitList, INT4 Start, 
	    INT4 FetchCount,GDT_BOOLEAN Quit)
{
  ZRECORDSCHOICE *Record;
  STRING Data, OID;
  INT4 i;

/*
  if((Start+FetchCount-1) > HitList.GetRecordCount())
    FetchCount = HitList.GetRecordCount() - Start + 1;
*/
  if((Start+FetchCount-1) > TotalHits)
    FetchCount = TotalHits - Start + 1;

  if (!Quit)
    CLEAR_SCREEN();
  cout << "CNIDR zclient, Version " << VERS;
  cout << ", Copyright (c) 1995,1996,1997 MCNC/CNIDR" << endl;
  cout << endl;
  cout << TotalHits << " total records matched your query" << endl;
  cout << endl;
  for(i=Start-1;i < FetchCount;i++) {
    Record = HitList.GetRecord(i);
    if(!Record)
      continue;
    Record->GetOID(&OID);
    Record->GetRecordData(&Data);
    cout << i+1 << ") ";
    if(OID == USMARC_OID) {
      MARC *m;
      m = new MARC(Data);
      m->SetDisplayFormat(MARC_FORMAT_TITLE);
      m->SetDisplayWidth(70);
      m->Print();
      delete m;
    } else
      cout << Data << endl;
  }
}
