/* $Id: zcon.cxx,v 1.15 1998/01/30 19:43:30 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery
and Retrieval, 1994-1997.

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby
granted without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact.

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of
noteworthy uses of this software.

3. The names of MCNC and Clearinghouse for Networked Information
Discovery and Retrieval may not be used in any advertising or publicity
relating to the software without the specific, prior written permission
of MCNC/CNIDR.

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF
THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
************************************************************************/

/*@@@
File:           zcon.cxx
Version:        2.00
$Revision: 1.15 $
Description:    Z39.50 connection manager
Author:         Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

/*  still needs to be converted to a class, with graceful error checking */

#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
//#include <values.h>
#include <limits.h>
#include <locale.h>

#if defined(_MSDOS) || defined(_WIN32)
#include <process.h>
#else
#include <unistd.h>
#endif

#include "defs.hxx"
#include "common.hxx"
#include "string.hxx"
#include "vlist.hxx"

#include "config.hxx"
#include "zclibase.hxx"
#include "marc.hxx"
#include "defines.hxx"
#include "registry.hxx"
#include "zcon.hxx"

ZCLIENT_BASE   *zclient;
char            HistoryFile[1024];
INT 	        DataReady;

//  Some configuration information
//CHR         *WEB_FORMS;
CHR         *SCRATCH;
CHR         *SPATH;

// zgate paths and names
CHR         *ZGATE_GATEWAY_PATH;
CHR         *ZGATE_GATEWAY_HTML;
CHR         *ZGATE_GATEWAY_URL;
CHR         *ZGATE_FILE_PATH;
CHR         *ZGATE_FILE_NAME;
CHR         *ZGATE_FILE_URL;
CHR         *ZGATE_TRACKER_PATH;
CHR         *ZGATE_TRACKER_NAME;

// ZCON PATHS AND NAMES
CHR         *ZCON_FILE_PATH;
CHR         *ZCON_FILE_NAME;
CHR         *ZCON_HISTORY_PATH;
CHR         *ZCON_HISTORY_NAME;
CHR         *ZCON_LOCK_PATH;
CHR         *ZCON_LOCK_NAME;
CHR         *ZCON_PROG_PATH;
CHR         *ZCON_PROG_NAME;

ofstream     LogFile;

#define MAX_CMD_LEN 10240

/*
 * Caller passes a temp filename as the first argument and its pid as the
 * second arg.  When this app finishes (trying) to initialize to the Z39.50
 * server, it creates the temporary file, writes a filename it will use for
 * subsequent IPC on line 1 and its pid on line 2.  The caller can use that
 * pid and filename combo to later send commands to this process (SIGUSR1).
 * 
 * Host should be argv[3], port argv[4]
 * 
 * args after that are optional:
 * 
 * -gid GroupId  -uid UserId  -pass Password
 */

int 
main(int argc, char **argv)
{
  pid_t ThisPID = getpid();

  char  mypath[1024];
  getcwd(mypath, sizeof(mypath));

  LogFile.open("/tmp/zcon.log",ios::app | ios::ate | ios::out,
	       S_IROTH | S_IRUSR | S_IRGRP | S_IWUSR | S_IWGRP);

  if (!setlocale(LC_CTYPE,"")) {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Warning: Failed to set the locale!" << endl;
    LogFile.flush();
  }

  ios::sync_with_stdio();

  fclose(stdin);
  fclose(stdout);
  fclose(stderr);
  //  freopen("/tmp/zcon.err","a",stderr);

  // Install an interrupt handler for a KILL signal
  signal(SIGKILL, KillHandler);

  // Install an interrupt handler for SIGUSR1 signal used by
  // zgate to tell zcon to read some input
  signal(SIGUSR1, USR1Handler);

  // Install an interrupt handler for SIGUSR2 signal used by
  // zgate to tell zcon that it has terminated
  signal(SIGUSR2, USR2Handler);

  Initialize(mypath);

  // Install an interrupt handler for an alarm signal
  signal(SIGALRM, AlarmHandler);
  alarm(LocalData.Timeout);

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Timeout set for " << LocalData.Timeout << " seconds." << endl;
  LogFile.flush();

  LocalData.CallersPID = atoi(argv[2]);

  FILE           *fp;
  if ((fp = fopen(argv[1], "w")) == NULL) {
    /*
     * Big trouble.  Nothing we can do.  Caller should time out
     * and/or check for existence of this file.
     */
    if(LocalData.CallersPID>0)
      kill(LocalData.CallersPID, SIGUSR1);
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Open " << argv[1] << " failed.  Exiting." << endl;
    LogFile.flush();
    LogFile.close();
  
    exit(1);
  }
  strcpy(LocalData.gCommandFile, argv[1]);
  LocalData.lnum=0;

  zclient = new ZCLIENT_BASE();

  GDT_BOOLEAN     FailedToInit = GDT_FALSE;

  /*
   * Initialize with the server Was a groupid, userid and password
   * included?
   */
  STRING          GroupId, UserId, Password;
  if (argc > 5) {
    // we have an id and password
    STRING          arg;
    INT             i;
    for (i = 5; i < argc; i += 2) {
      arg = argv[i];
      if (arg.CaseEquals("-gid"))
	GroupId = argv[i + 1];
      if (arg.CaseEquals("-uid"))
	UserId = argv[i + 1];
      if (arg.CaseEquals("-pass"))
	Password = argv[i + 1];
    }

    // Signal that no password was entered
    /*
    if (Password.Equals("-"))
    Password = "";
    */

    // save UserName - we need to add GlobalPassword
    Password.GetCString(LocalData.GlobalUserId,256);
    if (strchr(LocalData.GlobalUserId,'/') != NULL)
      *strchr(LocalData.GlobalUserId,'/')='\0'; // error
    if (!zclient->Initialize(argv[3], atoi(argv[4]),
			     GroupId, UserId, Password))
      FailedToInit = GDT_TRUE;

  } else {
    if (!zclient->Initialize(argv[3], atoi(argv[4])))
      FailedToInit = GDT_TRUE;
  }

  if (FailedToInit) {
    STRING          ReturnMsg;

    if (zclient->GetLastError() != ERR_INIT_DENIED) {
      /*
       * Not an server problem, but a *connection to* the server
       * problem
       */
      STRING          temp;
      ReturnMsg = "<STRONG>ERROR: </STRONG>";
      zclient->GetErrorMessage(&temp);
      ReturnMsg.Cat(temp);
      ReturnMsg.Cat("<BR>");
      LogFile << "[" << ThisPID << "] ";
      LogFile << "Connection error: " << temp << endl;
      LogFile.flush();

    } else {
      // Server Error
      INT4            i;
      STRLIST         MsgList;
      STRING          Msg;
      zclient->GetDiagnosticList(&MsgList);
      for (i = 1; i <= MsgList.GetTotalEntries(); i++) {
	MsgList.GetEntry(i, &Msg);
	ReturnMsg.Cat("<STRONG>Server ERROR: ");
	ReturnMsg.Cat("</STRONG>");
	ReturnMsg.Cat(Msg);
	ReturnMsg.Cat("<BR>");
	LogFile << "[" << ThisPID << "] ";
	LogFile << "Server error: " << Msg << endl;
	LogFile.flush();
      }
    }
    
    CHR *b;
    b = ReturnMsg.NewCString();
    fprintf(fp, "FAILURE [%d]\n%s\n", zclient->GetLastError(), b);
    delete[] b;
    fclose(fp);
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Failed to INIT -- Exiting." << endl;
    LogFile.flush();
    LogFile.close();
    if(LocalData.CallersPID>0)
      kill(LocalData.CallersPID, SIGUSR1);
    exit(1);
  }

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Session started." << endl;
  LogFile.flush();

  char           *motd;
  STRING          StrValue;
  zclient->GetMOTD(&StrValue);
  if (StrValue.GetLength() > 0) {
    motd = StrValue.NewCString();
    fprintf(fp, "SUCCESS\n%d\n%s", getpid(), motd);
    delete[] motd;
  } else {
    fprintf(fp, "SUCCESS\n%d\n--", getpid());
  }
  
  fclose(fp);

  if(LocalData.CallersPID>0)
    kill(LocalData.CallersPID, SIGUSR1);
  sprintf(LocalData.gCommandFile, "%s/%s.%d", 
	  SCRATCH,ZCON_FILE_NAME,getpid());
  
  // Accept all other commands via the signal handler
  for(;;){
    DataReady=0;
    for (;;){
      pause();
      if(DataReady)
	break;
    } 
    Process(&LocalData);
  }
  return 0;
}

// Called on session timeout
#ifndef __SUNPRO_CC
void 
AlarmHandler(int t)
#else
void 
AlarmHandler(int t,...)
#endif
{
  pid_t           ThisPID=getpid();

  fclose(stdout);
  fclose(stderr);
  fclose(stdin);
  
  unlink(LocalData.gCommandFile);
  
  zclient->Close();
  delete          zclient;

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Exit via AlarmHandler" << endl;
  LogFile.flush();
  LogFile.close();
  
  exit(0);
}


// Called on process kill
#ifndef __SUNPRO_CC
void 
KillHandler(int t)
#else
void 
KillHandler(int t,...)
#endif
{
  pid_t           ThisPID=getpid();

  fclose(stdout);
  fclose(stderr);
  fclose(stdin);
  
  unlink(LocalData.gCommandFile);
  
  zclient->Close();
  delete          zclient;

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Exit via KillHandler" << endl;
  LogFile.flush();
  LogFile.close();
  
  exit(0);
}


// Called on receipt of SIGUSR2.  This function should send a 
// triggerResourceControl (break) PDU to the server
// then longjmp back to the main command loop
#ifndef __SUNPRO_CC
void  
USR2Handler(int t)
#else
void  
USR2Handler(int t,...)
#endif
{
  pid_t           ThisPID=getpid();

  fclose(stdout);
  fclose(stderr);
  fclose(stdin);

  unlink(LocalData.gCommandFile);
  
  zclient->Close();
  delete zclient;

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Exit via USR2Handler" << endl;
  LogFile.flush();
  LogFile.close();
  
  exit(0);

}

#define CMD_SEARCH 0
#define CMD_PRESENT 1
#define CMD_LOGOUT 2
#define CMD_SCAN 3


// in our class, USR1handler should do nothing but load
// data from the file and set a flag
#ifndef __SUNPRO_CC
void 
USR1Handler(int t)
#else
void 
USR1Handler(int t,...)
#endif
{
  char            cmd[10240];
  CHR            *temp;
  INT             cmd_count;
  int             i, j;
  pid_t           ThisPID=getpid();

  // Reinstall an interrupt handler for a KILL signal
  signal(SIGKILL, KillHandler);

  // Reinstall an interrupt handler for SIGUSR1 signal used by
  // zgate to tell zcon to read some input
  signal(SIGUSR1, USR1Handler);

  // Reinstall an interrupt handler for SIGUSR2 signal used by
  // zgate to tell zcon that it has terminated
  signal(SIGUSR2, USR2Handler);

  // reset the alarm
  alarm(LocalData.Timeout);
  
  LocalData.command=-1;
  LocalData.start=LocalData.count=1;
  LocalData.NumReq=12;

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Taking commands from " << LocalData.gCommandFile << endl;
  LogFile.flush();

  FILE *fp, *flog;
  if ((fp = fopen(LocalData.gCommandFile, "r")) == NULL) {
    perror(LocalData.gCommandFile);
    LogFile << "[" << ThisPID << "] ";
    LogFile << "USR1Handler: Open " << LocalData.gCommandFile;
    LogFile << " failed." << endl;
    LogFile.flush();
    LogFile.close();
    exit(1);

  }

  // set up defaults
  strcpy(LocalData.esname, "B");
  strcpy(LocalData.recsyntax, SUTRS_OID);
  LocalData.resultset[0] = '\0';
  cmd_count=0;

  while (fgets(cmd, MAX_CMD_LEN, fp)) {
    int             len = strlen(cmd);
    if (len > 1)
      cmd[len - 1] = '\0';
    else
      break;

    LogFile << "[" << ThisPID << "] ";
    LogFile << "USR1Handler found command: " << cmd << endl;
    LogFile.flush();

    switch (cmd_count) {

    case 0:
      sscanf(cmd, "%d", &LocalData.CallersPID);
      break;

    case 1:
      if (!StrCaseCmp(cmd, "logout")) {
	LogFile << "[" << ThisPID << "] ";
	LogFile << "Got LOGOUT command" << endl;
	LogFile.flush();
	LocalData.command = CMD_LOGOUT;

      } else if (!StrCaseCmp(cmd, "search")) {
	LogFile << "[" << ThisPID << "] ";
	LogFile << "Got SEARCH command" << endl;
	LogFile.flush();
	LocalData.command = CMD_SEARCH;

      } else if (!StrCaseCmp(cmd, "present")) {
	LogFile << "[" << ThisPID << "] ";
	LogFile << "Got PRESENT command" << endl;
	LogFile.flush();
	LocalData.command = CMD_PRESENT;

      } else if (!StrCaseCmp(cmd, "scan")) {
	LogFile << "[" << ThisPID << "] ";
	LogFile << "Got SCAN command" << endl;
	LogFile.flush();
	LocalData.command = CMD_SCAN;

      } else {
	LogFile << "[" << ThisPID << "] ";
	LogFile << "Error:Got invalid command" << cmd << endl;
	LogFile.flush();
	//	cout << "Invalid command" << endl;
	if(LocalData.CallersPID>0)
	  kill(LocalData.CallersPID, SIGUSR1);
	//	LogFile << "USR1Handler: Invalid command: " 
	//		<< LocalData.command << endl;
	LogFile.flush();
	LogFile.close();
	exit(1);
      }
      break;

    case 2:
      if (LocalData.command == CMD_SEARCH || LocalData.command == CMD_SCAN)
	strcpy(LocalData.db, cmd);
      if (LocalData.command == CMD_PRESENT)
	strcpy(LocalData.resultset, cmd);
      break;

    case 3:
      if (LocalData.command == CMD_SEARCH || LocalData.command == CMD_SCAN)
	strcpy(LocalData.term, cmd);
      if (LocalData.command == CMD_PRESENT)
	LocalData.start = atoi(cmd);
      break;

    case 4:
      if (LocalData.command == CMD_SEARCH)
	strcpy(LocalData.esname, cmd);
      if (LocalData.command == CMD_PRESENT)
	LocalData.count = atoi(cmd);
      break;

    case 5:
      if (LocalData.command == CMD_SEARCH)
	strcpy(LocalData.recsyntax, cmd);
      if (LocalData.command == CMD_PRESENT)
	strcpy(LocalData.esname, cmd);
      if (LocalData.command == CMD_SCAN)
	LocalData.NumReq = atoi(cmd);
      break;

    case 6:
      if (LocalData.command == CMD_SEARCH) {
	LocalData.MaxRecords = atoi(cmd);
	LocalData.DisplayCnt = LocalData.MaxRecords;
      }
      if (LocalData.command == CMD_PRESENT)
	strcpy(LocalData.recsyntax, cmd);
      break;

    case 7:
      //      if (LocalData.command == CMD_SEARCH) {
      //	LocalData.start = atoi(cmd);
      //      }
      if (LocalData.command == CMD_SEARCH) {
	LocalData.RecsPerServer = atoi(cmd);
      }
      if (LocalData.command == CMD_PRESENT) {
	LocalData.MaxRecords = atoi(cmd);
      }
      break;
    case 8:
      if (LocalData.command == CMD_SEARCH) {
	LocalData.start = atoi(cmd);
      }
      break;
    case 9:
      if (LocalData.command == CMD_SEARCH) {
	strcpy(LocalData.attrset,cmd);
      }
      break;
    }
    cmd_count++;
  }
  fclose(fp);
 
#ifndef DEBUG
  LogFile << "[" << ThisPID << "] ";
  LogFile << "Deleting " << LocalData.gCommandFile << endl;
  LogFile.flush();
  unlink(LocalData.gCommandFile);
#endif

  DataReady=1;
  
}


void 
Initialize(char *mp) {
  STRING Filename, Group;
  STRLIST ResultList, Position, Value;
  INT count, status;
  STRING Section, Entry, Hold;
  REGISTRY *Defaults;
  struct stat info;
  char mypath[1024];

  strcpy(mypath,mp);

  Defaults = new REGISTRY("zcon");

  //  Filename = ZGATE_INI_NAME;
  //  Defaults->ProfileLoadFromFile(Filename, Position);
  Filename = ZGATE_INI_NAME;
  if (IsFile(Filename)) {
    Defaults->ProfileLoadFromFile(Filename, Position);
  } else {
    // Didn't find the compiled-in name - let's see if we can guess the name
    Filename = mypath;
    Filename.Cat("/gateway.ini");
    if (IsFile(Filename))
      Defaults->ProfileLoadFromFile(Filename, Position);
  }

  Section="config";
  /*
    Entry="WEBFORMS";
    Defaults->ProfileGetString(Section, Entry, "/usr/local/etc/httpd/htdocs", 
    &Hold);
    WEB_FORMS = Hold.NewCString();
  */

  Entry="SCRATCH_DIR";
  //  Defaults->ProfileGetString(Section, Entry, "/tmp", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_TMP_DIR, &Hold);
  SCRATCH = Hold.NewCString();

  Entry="SPATH";
  //  Defaults->ProfileGetString(Section, Entry, "/tmp/queries", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_SPATH, &Hold);
  SPATH = Hold.NewCString();

  if(stat(SPATH,&info) != 0) {
    // Directory doesn't exist - make it
    if (mkdir(SPATH, S_IROTH | S_IRUSR | S_IRGRP | S_IWUSR | S_IWGRP | S_IXUSR | S_IXGRP) != 0) {
      // The mkdir failed, so just use /tmp instead
      strcpy(SPATH,GATEWAY_TMP_DIR);
    }
  }
  if (S_ISDIR(info.st_mode) == 0) {
    // It exists, but it isn't a directory so just use /tmp instead
    strcpy(SPATH,GATEWAY_TMP_DIR);
  }

  Entry="TIMEOUT";
  Defaults->ProfileGetString(Section, Entry, DEFAULT_TIMEOUT, &Hold);
  LocalData.Timeout = Hold.GetInt();

  Entry="TRACKER_PATH";
  Defaults->ProfileGetString(Section, Entry, SCRATCH, &Hold);
  ZGATE_TRACKER_PATH = Hold.NewCString();

  Entry="TRACKER_NAME";
  //Defaults->ProfileGetString(Section, Entry, "recent", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_TRACKER, &Hold);
  ZGATE_TRACKER_NAME = Hold.NewCString();

  Entry="HISTORY_PATH";
  Defaults->ProfileGetString(Section, Entry, SCRATCH, &Hold);
  ZCON_HISTORY_PATH = Hold.NewCString();

  Entry="HISTORY_NAME";
  //Defaults->ProfileGetString(Section, Entry, "hx", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_HISTORY, &Hold);
  ZCON_HISTORY_NAME = Hold.NewCString();

  Entry="LOCK_PATH";
  Defaults->ProfileGetString(Section, Entry, SCRATCH, &Hold);
  ZCON_LOCK_PATH = Hold.NewCString();

  Entry="LOCK_NAME";
  //Defaults->ProfileGetString(Section, Entry, "lock", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_LOCK, &Hold);
  ZCON_LOCK_NAME = Hold.NewCString();


  // zgate configuration paths
  Section="gateway";
  Entry="GATEWAY_PATH";
  //Defaults->ProfileGetString(Section, Entry, "/usr/local/etc/httpd/htdocs", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_HTDOCS, &Hold);
  ZGATE_GATEWAY_PATH = Hold.NewCString();

  Entry="GATEWAY_HTML";
  //Defaults->ProfileGetString(Section, Entry, "gateway.html", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_FORM, &Hold);
  ZGATE_GATEWAY_HTML = Hold.NewCString();

  Entry="GATEWAY_URL";
  //Defaults->ProfileGetString(Section, Entry, "http://localhost/gateway.html", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_LINK, &Hold);
  ZGATE_GATEWAY_URL = Hold.NewCString();

  Entry="PROG_PATH";
  //Defaults->ProfileGetString(Section, Entry, "/usr/local/etc/httpd/cgi-bin", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_CGIDIR, &Hold);
  ZGATE_FILE_PATH = Hold.NewCString();

  Entry="PROG_NAME";
  //Defaults->ProfileGetString(Section, Entry, "zgate", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_ZGATE, &Hold);
  ZGATE_FILE_NAME = Hold.NewCString();

  Entry="PROG_URL";
  //Defaults->ProfileGetString(Section, Entry, "http://localhost/cgi-bin/zgate", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_GATELINK, &Hold);
  ZGATE_FILE_URL = Hold.NewCString();


  // zcon configuration paths
  Section="connection";
  Entry="FILE_PATH";
  //Defaults->ProfileGetString(Section, Entry, "cgi-bin", &Hold);
  Defaults->ProfileGetString(Section, Entry, CGIBIN, &Hold);
  ZCON_FILE_PATH = Hold.NewCString();

  Entry="FILE_NAME";
  //Defaults->ProfileGetString(Section, Entry, "zcon", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_ZCON, &Hold);
  ZCON_FILE_NAME = Hold.NewCString();

  Entry="PROG_PATH";
  //Defaults->ProfileGetString(Section, Entry, "/usr/local/etc/httpd/cgi-bin", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_CGIDIR, &Hold);
  ZCON_PROG_PATH = Hold.NewCString();

  Entry="PROG_NAME";
  //Defaults->ProfileGetString(Section, Entry, "zcon", &Hold);
  Defaults->ProfileGetString(Section, Entry, GATEWAY_ZCON, &Hold);
  ZCON_PROG_NAME = Hold.NewCString();

  delete Defaults;
}


//  Utility Functions
int 
ShowDisplayList(INT pos, CHR * recsyntax, FILE * fp, CHR * ln)
{
  pid_t  localPid = getpid();
  
  fprintf(fp,
	  "<A HREF=\"%s?present+%d+%s+%d+1+F+%s\">Full</A>", 
	  ZGATE_FILE_URL,localPid, ln, pos, recsyntax);
  return(0);
}


void 
ShowLogoutButton(FILE * fp)
{
#ifdef NO_LOGOUT
  fprintf(fp, "Return to the <a href=\"%s\">gateway.</a>\n",GATEWAY_LINK);
#else
  fprintf(fp, "<form method=\"POST\" action=\"%s\">\n",
	  ZGATE_FILE_URL);
  fprintf(fp, "<input name=\"SERVICE\" value=\"LOGOUT\" type=\"HIDDEN\">\n");
  fprintf(fp, "<INPUT NAME=\"SESSION_ID\" VALUE=\"%d\"", getpid());
  fprintf(fp, " TYPE=\"HIDDEN\">\n");
  fprintf(fp, "<INPUT TYPE=\"SUBMIT\" VALUE=\"Logout\">\n");
  fprintf(fp, "</FORM>\n");
#endif
}


void 
ShowGeneralButtons(FILE * fp)
{
  INT             localPid = getpid();
  fprintf(fp, "<a href=\"%s\">Help</a> ",ZCON_DETAILED_HELP);
  ShowCmtButtons(fp);
  fprintf(fp, "<a href=\"#0\">First Hit</a>");
}


void 
ShowScanButtons(FILE * fpc)
{
  fprintf(fpc, "<a href=\"%s\">Help</a> ",ZCON_SCAN_HELP);
  ShowCmtButtons(fpc);
}


void 
ShowHitsHlpButton(FILE * fpc)
{
  fprintf(fpc, "<a href=\"%s\">Help</a> ",ZCON_HITS_HELP);
}


void 
ShowCmtButtons(FILE * fpc)
{
  fprintf(fpc, "<a href=\"/%s\">Comments</a> ",ZCON_COMMENTS);
}


void 
ShowDocHeader(FILE * fpc)
{
  return;
}


void 
ShowDocTrailer(FILE * fpc)
{
//  fprintf(fpc, "\n<HR>\n<a href=\"http://www.fgdc.gov/\">FGDC</a> Geospatial Z39.50 ");
  fprintf(fpc, "Gateway Based on <a href=\"http://www.cnidr.org/\"><B>CNIDR</B></a> Isite\n<BR>");
  fprintf(fpc, "<I>Copyright &copy MCNC, Clearinghouse for Networked Information Discovery and Retrieval, 1994-1997.</I><P>\n");
  fprintf(fpc, "</BODY>\n</HTML>\n");
}


void
ShowGetMoreForm(FILE * fp, INT start, INT HitCount, INT Count,
		INT MaxRecords, char * recsyntax)
// fp = file pointer to the results page
// start = starting hit #
// HitCount = total number of hits from the search
// Count = # of hits returned to the user, = min(HitCount,MaxCount)
// MaxRecords = maximum # of hits requested by user
// recsyntax = Preferred Record Syntax OID
{
  INT NumLeft;
  NumLeft = (Count <= (HitCount-(start+Count-1)))
    ? Count:(HitCount-(start+Count-1));
  if (NumLeft >0) {
    fprintf(fp, 
	    "<form method=\"POST\" action=\"%s\">\n",
	    ZGATE_FILE_URL);
    fprintf(fp, 
	    "<input name=\"SERVICE\" value=\"PRESENT\" type=\"HIDDEN\">\n");
    fprintf(fp, 
	    "<input name=\"START\" value=\"%d\" type=\"HIDDEN\">\n", 
	    start+Count);
    fprintf(fp, 
	    "<input name=\"ESNAME\" value=\"B\" type=\"HIDDEN\">\n");
    fprintf(fp, 
	    "<input name=\"RECSYNTAX\" value=\"%s\" type=\"HIDDEN\">\n", 
	    recsyntax);
    fprintf(fp, 
	    "<input name=\"COUNT\" value=\"%d\" type=\"HIDDEN\">\n", 
	    NumLeft);
    fprintf(fp, 
	    "<input name=\"HITCOUNT\" value=\"%d\" type=\"HIDDEN\">\n", 
	    HitCount);
    fprintf(fp, 
	    "<input name=\"SESSION_ID\" value=\"%d\" type=\"HIDDEN\">\n", 
	    getpid());
    fprintf(fp, "<input type=\"SUBMIT\" value=\"Show More Records\">\n");
    fprintf(fp, "</form>\n");
  }
}


/*=========================================================
  
  CLASS:  CONMGR
  METHOD: ProcessRecord
  AUTHOR:	Jim Fullton
  
  Notes: Adds links between hits in a document
  
  =========================================================*/
char*
ProcessRecord(char *s, INT Link)
{
  
  int             len;
  char           *NewRec, *p;
  int             oldPtr, newPtr;
  int             mode = 0;
  int             Counter = 0;
  int             i;
  CHR             tmp[256];
  
  len = strlen(s);
  NewRec = new CHR[len + 5000];
  if (!NewRec)
    return (NULL);
  for (i = 0; i < len + 5000; i++)
    NewRec[i] = '\0';
  for (oldPtr = 0, newPtr = 0; oldPtr < len; oldPtr++) {
    if (s[oldPtr] == '>' || s[oldPtr] == '<')
      s[oldPtr] = ' ';
    if (s[oldPtr] == '*' && s[oldPtr + 1] == '*' && s[oldPtr + 2] == '*') {
      if (mode == 0) {
	
	p = &NewRec[newPtr];
	sprintf(tmp, "<a name=\"%d\">", Counter++);
	if (Link) {
	  strcpy(p, tmp);
	  newPtr += strlen(tmp);
	  NewRec[newPtr++] = '<';
	  NewRec[newPtr++] = '/';
	  NewRec[newPtr++] = 'A';
	  NewRec[newPtr++] = '>';
	}
	NewRec[newPtr++] = '<';
	NewRec[newPtr++] = 'B';
	NewRec[newPtr++] = '>';
	p = &NewRec[newPtr];
	sprintf(tmp, "<a href=\"#%d\">", Counter);
	if (Link) {
	  strcat(p, tmp);
	  newPtr += strlen(tmp);
	}
	oldPtr += 2;
	mode = 1;
      } else {
	if (Link) {
	  NewRec[newPtr++] = '<';
	  NewRec[newPtr++] = '/';
	  NewRec[newPtr++] = 'A';
	  NewRec[newPtr++] = '>';
	}
	NewRec[newPtr++] = '<';
	NewRec[newPtr++] = '/';
	NewRec[newPtr++] = 'B';
	NewRec[newPtr++] = '>';
	oldPtr += 2;
	mode = 0;
      }
      
    } else
      NewRec[newPtr++] = s[oldPtr];
    
  }
  NewRec[newPtr] = '\0';
  NewRec[newPtr++] = '\0';
  
  return (NewRec);
}


/*=========================================================
  
  CLASS:  	CONMGR
  METHOD: 	ExecutePresent
  AUTHOR:	Jim Fullton
  
  Notes:
  
  =========================================================*/
void 
ExecutePresent(FILE *fp,int lnum,char *resultset, int start, int count, 
	       char *esname,char *recsyntax)
{
  ZRECORDLIST RecordList;
  INT HitCount;
  CHR             LN[20];
  pid_t ThisPID = getpid();

/*
  if (resultset[0] != 'L') {
    sprintf(LN, "L%d", lnum);
  } else {
    strcpy(LN, resultset);
  }
  HitCount = LocalData.LHIT[atoi(&LN[1])]; // get hits for this L number

  if (start + count > HitCount)
    count = HitCount - start + 1;
  if (count <= 0)
    count = 1;

  resultset[0] = '\0';
*/

  if(!zclient->Present(start, count, recsyntax, LN, esname, 
		       &RecordList)) {
    fprintf(fp, "Failed to present\n");
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Failed to present " << endl;
    LogFile.flush();
    LogFile.close();
    exit(1);
  }
		   
  INT4 Count, i;
  Count = RecordList.GetRecordCount();
  HitCount = LocalData.MaxRecords;
  if (!StrCaseCmp(esname, "B")) {
    ShowLogoutButton(fp);
    ShowHitsHlpButton(fp);
    ShowCmtButtons(fp);
    //    fprintf(fp, "</CENTER>\n");
    fprintf(fp, "<hr><B>Search Results - </B>");
    if (Count > 0) {
      fprintf(fp, "<I>Records %d through %d of %d returned.</I><HR>\n", 
	      start, start+Count-1, HitCount);
    } else {
      fprintf(fp, "<I>No records returned.</I><HR>\n");
    }
  } else if (StrCaseCmp(esname, "F"))
    fprintf(fp, "<I>Records %d through %d of %d returned.</I><HR>\n", 
	    start, start+Count-1, HitCount);

  ZRECORDSCHOICE *Record;
  STRING Data, OID;

  for(i=0;i < Count;i++) {
    Record = RecordList.GetRecord(i);
    if(!Record)
      continue;
    Record->GetOID(&OID);
    Record->GetRecordData(&Data);
    if(OID == USMARC_OID) {
      MARC *m;
      m = new MARC(Data);
      fprintf(fp, "<PRE>\n");
      m->Print(fp);
      fprintf(fp, "</PRE>\n");
      delete m;

    } else if (OID == SUTRS_OID) {
      CHR *temp;
      fprintf(fp, "<PRE>\n");
      temp = Data.NewCString();
      fprintf(fp, "%s\n", temp);
      fprintf(fp, "</PRE>\n");
      fprintf(fp, "<hr>\n");
      delete [] temp;

    } else {
      CHR *temp;
      CHR *New;
      INT Link = StrCaseCmp(esname, "B");
      temp = Data.NewCString();

/*
      New = ProcessRecord(temp, Link);
      if (New) {
        delete[] temp;
        temp = New;
      }
*/

      fprintf(fp, "%s\n", temp);
      delete [] temp;
    }
    // This is showing up in the full record, and we don't need it there
    if(StrCaseCmp(esname, "F")) {
      // Add CallersPID to the end of the string to the URL looks different
      // to the browser each time it's constructed.
      fprintf(fp, 
	      "<BR><A HREF=\"%s?present+%d+Default+%d+1+F+%s+%d\">", 
	      ZGATE_FILE_URL,getpid(), start+i, 
	      recsyntax, LocalData.CallersPID);
      fprintf(fp, "More on this record</A>\n<P>\n");
      fprintf(fp, "<HR>\n");
    } else
      fprintf(fp, "\n<P>\n");
  }

  // Put a followup form at the bottom for brief records
  if(!(StrCaseCmp(esname, "B"))) {
    ShowGetMoreForm(fp,start,HitCount,Count,Count,recsyntax);
  }

}


/*=========================================================
  
  CLASS:  CONMGR
  METHOD: ExecuteLogOut
  AUTHOR:	Jim Fullton
  
  Notes:  LogOut from remote Z39.50 server.  Exits zcon
  session and returns indicator to calling process
  
  =========================================================*/
void 
ExecuteLogOut(FILE *fp, pid_t CallersPID, ZCLIENT_BASE *zclient)
{
  CHR HistoryFile[256];
  pid_t ThisPID=getpid();

  zclient->Close();

  /*
  fprintf(fp, "<HTML>\n<HEAD>\n");
  fprintf(fp, "<TITLE>Successful Logout</TITLE>\n");
  fprintf(fp, "</HEAD>\n<BODY>\n");
  fprintf(fp, "<h1>Logout Completed</h1>");
  fprintf(fp, "Successful Logout Completed<BR><HR>\n");
  fprintf(fp, "<A HREF=\"http://www.cnidr.org\">CNIDR's</A> Stateful Z39.50 Gateway\n");
  fprintf(fp, "</BODY>\n</HTML>\n");
  */

  CHR GatewayFile[256];
  sprintf(GatewayFile,"%s/%s",ZGATE_GATEWAY_PATH,ZGATE_GATEWAY_HTML);

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Writing final text to " << GatewayFile << endl;
  LogFile.flush();

  STRING HTMLgate;
  HTMLgate.ReadFile(GatewayFile);
  fprintf(fp,HTMLgate);
  fclose(fp);


  if (CallersPID>0)
    kill(CallersPID, SIGUSR1);
  sprintf(HistoryFile, "%s/%s.%d", 
	  ZCON_HISTORY_PATH, ZCON_HISTORY_NAME, getpid());
  unlink(HistoryFile);
  signal(SIGUSR1, USR1Handler);
  alarm(LocalData.Timeout);

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Exiting via ExecuteLogOut " << endl;
  LogFile.flush();
  LogFile.close();
  
  exit(0);
  
}


/*=========================================================
  
  CLASS: CONMGR
  METHOD: ExecuteScan
  AUTHOR:	Jim Fullton
  
  Notes:  Handles requests for Z39.50 SCAN service
  
  =========================================================*/
void 
ExecuteScan(FILE *fp, INT NumReq, char *db, char *term)
{
  ZRECORDLIST     RecordList;
  CHR            *temp;
  STRING          Data, OID;
  INT i;
  if (NumReq > 25)
    NumReq = 25;
  if (!zclient->Scan(db, term, 3, 0, NumReq, &RecordList)) {
    ShowLogoutButton(fp);
    ShowScanButtons(fp);
    fprintf(fp, "</FORM></CENTER>");
    fprintf(fp, "<hr><B>Scan failed - Check request syntax </B>");
    return;
  }
  ZRECORDSCHOICE *Record;
  
  INT4            Occ;
  
  ShowLogoutButton(fp);
  ShowScanButtons(fp);
  fprintf(fp, "</FORM></CENTER>");
  fprintf(fp, "<hr><B>Scan Results</B><HR><P ALIGN=CENTER><TABLE ALIGN=CENTER>");
  fprintf(fp, "<TH>List of Terms</TH>");
  fprintf(fp, "<TH>Number of Occurrences</TH>");
  
  for (i = 0; i < NumReq; i++) {
    Record = RecordList.GetRecord(i);
    if (!Record)
      continue;
    Record->GetOID(&OID);
    Record->GetRecordData(&Data);
    Record->GetRecordOcc(&Occ);
    temp = Data.NewCString();
    fprintf(fp, "<TR ALIGN=LEFT><TD>");
    fprintf(fp, "%s</TD><TD>%d</TD></TR>", temp, Occ);
    delete          temp;
    
  }
  fprintf(fp, "</TABLE></P><HR>");
  fprintf(fp, "<CENTER>");
  ShowScanButtons(fp);
  ShowLogoutButton(fp);
  fprintf(fp, "</FORM></CENTER>");
    
}


/*=========================================================
  
  CLASS:  CONMGR
  METHOD: ExecuteSearch
  AUTHOR:	Jim Fullton
  
  Notes:
  
  =========================================================*/
void 
ExecuteSearch(FILE *fp, char *db, char *term, int lnum, int start,
	      int MaxRecords,int HitsPerServer,char *esname, 
	      char *recsyntax)
{
  INT HitCount;
  CHR             LN[20];
  ZRECORDLIST     RecordList;
  pid_t           ThisPID=getpid();
  INT             n_recs;

  if (LocalData.lnum == 100)
    LocalData.lnum = 0;
  LocalData.lnum = (LocalData.lnum + 1) % 100;
  sprintf(LN, "L%d", LocalData.lnum);
  n_recs = MaxRecords;

  // Count how many databases?
  if ((HitsPerServer > 0) && (strlen(db)>0)) {
    INT n_dbs;
    CHR *ptr;
    CHR tmp[10240];

    strcpy(tmp,db);

    n_dbs = 0;
    ptr = strtok(tmp,",");
    while (ptr) {
      n_dbs++;
      ptr = strtok((CHR*)NULL,",");
    }

    n_recs = HitsPerServer * n_dbs;
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Requesting " << n_recs << " records from " << n_dbs 
	    << " databases." << endl;
    LogFile.flush();
  }

#ifdef DEBUG
  FILE *fy;
  fy=fopen("/tmp/search.txt","a");
  fprintf(fy,"%s\n",term);
  fclose(fy);
#endif /* DEBUG */

  if (!zclient->Search(db, term, LN, LocalData.attrset, &HitCount, 
		       LocalData.GlobalUserId, start, n_recs, &RecordList)) 
    {
      LogFile << "[" << ThisPID << "] ";
      LogFile << "Search in db " << db << " for " << term << " failed" << endl;
      LogFile.flush();
      LogFile.close();
      exit(1);
    }
  //  fprintf(fp, "<p align=\"center\">\n");
  ShowLogoutButton(fp);
  ShowHitsHlpButton(fp);
  ShowCmtButtons(fp);
  //  fprintf(fp, "</p>\n");

  fprintf(fp, "<H1>Query Results: Search %s for %s</H1>\n",db,term);
  if (HitCount == 0) {
    fprintf(fp, "<I>No records matched your query</I><BR>\n");
  } else if (HitCount == 1) {
    fprintf(fp, "<I>1 record matched your query</I><BR>\n");
  } else {
    fprintf(fp, "<I>%d records matched your query</I><BR>\n",HitCount);
  }

  // Fetch as many as they asked for (MaxRecords), if there are that many
  // Otherwise, get all of the hits (HitCount)
  INT4 FetchCount;
  if (HitCount <= n_recs)
    FetchCount = HitCount;
  else
    FetchCount = n_recs;

  RecordList.Clear();
  if(!zclient->Present(1, FetchCount, recsyntax, LN, esname, &RecordList)) {
    fprintf(fp, "Failed to present\n");
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Failed to present" << endl;
    LogFile.flush();
    LogFile.close();
    exit(1);
  }
    
  INT4 Count, i;
  Count = RecordList.GetRecordCount(); // Should equal FetchCount
  if (Count > 0) {
    fprintf(fp, "<I>Records 1 through %d of %d returned.</I><HR>\n", 
	    Count, HitCount);
  } else {
    fprintf(fp, "<I>No records returned.</I><HR>\n");
  }

  ZRECORDSCHOICE *Record;
  STRING Data, OID;
  for(i=0;i < Count;i++) {
    Record = RecordList.GetRecord(i);
    if(!Record)
      continue;
    Record->GetOID(&OID);
    Record->GetRecordData(&Data);
    
    if(OID == USMARC_OID) {
      MARC *m;
      m = new MARC(Data);
      fprintf(fp, "<PRE>\n"); 
      m->Print(fp);
      fprintf(fp, "</PRE>\n"); 
      delete m;
    } else if (OID == HTML_OID) {
      CHR *temp;
      temp = Data.NewCString();
      fprintf(fp, "%s\n", temp);
      delete [] temp;
    } else if (OID == SUTRS_OID) {
      CHR *temp;
      temp = Data.NewCString();
      fprintf(fp, "<pre>\n");
      fprintf(fp, "%s", temp);
      fprintf(fp, "</pre>\n");
      delete [] temp;
    } else {
      CHR *temp;
      temp = Data.NewCString();
      fprintf(fp, "<B>"); 
      fprintf(fp, "%s", temp);
      fprintf(fp, "</B>\n"); 
      delete [] temp;
    }
    fprintf(fp, 
	    "<BR><A HREF=\"%s?present+%d+Default+%d+1+F+%s\">More on this record</A>\n", 
	    ZGATE_FILE_URL,getpid(), start+i, recsyntax);
    fprintf(fp, "<P>\n");
    fprintf(fp, "<hr>\n");
  }

  if(HitCount > FetchCount) {
    ShowGetMoreForm(fp,start,HitCount,FetchCount,n_recs,recsyntax);
  }
    
}


void 
Process(struct _Data *l)
{
  FILE *fp=(FILE*)NULL; 
  FILE *flock=(FILE*)NULL;
  CHR LockName[256];
  pid_t ThisPID=getpid();

  sprintf(LockName,"%s/%s%d",
	  ZCON_LOCK_PATH,ZCON_LOCK_NAME,getpid());

  if ((fp = fopen(l->gCommandFile, "w")) == NULL) {
    perror(LocalData.gCommandFile);
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Failed to open outgoing command file " 
	    << LocalData.gCommandFile << endl;
    LogFile.flush();
    LogFile.close();
    exit(1);

  }
  
  static INT4     HitCount;

  switch (LocalData.command) {

  case CMD_SCAN:
    fprintf(fp, "<HTML>\n<HEAD>\n");
    fprintf(fp, "<TITLE>SCAN %s for %s</TITLE>",l->db, l->term);
    fprintf(fp,"<BODY BGCOLOR=\"#ffffff\" TEXT=\"#000000\"><CENTER>\n");
    ExecuteScan(fp,l->NumReq,l->db,l->term);
    break;
    
  case CMD_LOGOUT:
    ExecuteLogOut(fp,l->CallersPID,zclient);
    break;
    
  case CMD_SEARCH:
    fprintf(fp, "<HTML>\n<HEAD>\n");
    fprintf(fp, "<TITLE>Search %s for %s</TITLE>\n",l->db, l->term);
    fprintf(fp,"<BODY BGCOLOR=\"#ffffff\" TEXT=\"#000000\">\n<CENTER>\n");
    flock=fopen(LockName,"w");
    if (flock) {
      fprintf(flock,"SEARCH\n");
      fclose(flock);
    }
    ExecuteSearch(fp,l->db,l->term,l->lnum,l->start,l->MaxRecords,
		  l->RecsPerServer,l->esname,l->recsyntax);
    unlink(LockName);
    break;

  case CMD_PRESENT: 
    flock=fopen(LockName,"w");
    if (flock) {
      fprintf(flock,"PRESENT\n");
      fclose(flock);
    }
    if (!(StrCaseCmp(l->esname,"B"))) {
      fprintf(fp, "<HTML>\n<HEAD>\n");
      fprintf(fp, "<TITLE>Search %s for %s</TITLE>\n",l->db, l->term);
      fprintf(fp,"<BODY BGCOLOR=\"#ffffff\" TEXT=\"#000000\">\n");
      //      fprintf(fp,"<CENTER>\n");
    }
    ExecutePresent(fp,l->lnum,l->resultset,l->start,l->count,l->esname,
		   l->recsyntax);
    unlink(LockName);
    break;
    
  }
  ShowDocTrailer(fp);
  fflush(fp);
  fclose(fp);
  if(l->CallersPID>0)
    kill(l->CallersPID, SIGUSR1);
  signal(SIGUSR1, USR1Handler);
  alarm(LocalData.Timeout);
}


INT 
GetCurrentSNumber()
{
  CHR             UserDir[256];
  INT             Value=0;
  CHR             buf[256];
  FILE           *fu;
  
  sprintf(UserDir, "%s/%s/Count", SPATH,LocalData.GlobalUserId);
  fu = fopen(UserDir, "r");
  if (fu) {
    fgets(buf, 256, fu);
    fclose(fu);
    Value = atoi(buf);
    Value--;
  }
  return(Value);
}


void 
GetCurrentSQuery(char *s)
{
  CHR             UserDir[256];
  INT             Value;
  CHR             buf[1024];
  FILE           *fu;

  buf[0]='\0';
  sprintf(UserDir, "%s/%s/Count", SPATH,LocalData.GlobalUserId);
  fu = fopen(UserDir, "r");
  if(fu) {
    fgets(buf, 256, fu);
    fclose(fu);
    Value = atoi(buf);
    Value--;
    sprintf(UserDir,"%s/%s/%05d",SPATH,LocalData.GlobalUserId,Value);
    fu = fopen(UserDir, "r");
    if (fu) {
      fgets(buf, 1024, fu);
      fgets(buf, 1024, fu);
      fclose(fu);
    } else
      buf[0]='\0';
  }
  strcpy(s,buf);
}
