/* $Id: zgate.cxx,v 1.11 1998/01/30 19:34:10 cnidr Exp $ */
/************************************************************************
Copyright Notice
 
Copyright (c) MCNC, Clearinghouse for Networked Information Discovery
and Retrieval, 1994.
 
Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby granted
without fee, provided that
 
1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact.
 
2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of noteworthy
uses of this software.
 
3. The names of MCNC and Clearinghouse for Networked Information Discovery
and Retrieval may not be used in any advertising or publicity relating to
the software without the specific, prior written permission of MCNC/CNIDR.
 
THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY
OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 
IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
************************************************************************/

/*@@@
File:           zgate.cxx
Version:        2.00
$Revision: 1.11 $
Description:    http-to-Z39.50 gateway
Author:         Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#include <iostream.h>
#include <fstream.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#ifdef UNIX
#include <unistd.h>
#endif
#include <string.h>
#include <locale.h>

#include "config.hxx"
#include "ztags.hxx"
#include "string.hxx"
#include "vlist.hxx"
#include "strlist.hxx"
#include "tokengen.hxx"
#include "strstack.hxx"
#include "infix2rpn.hxx"
#include "gdt.h"
#include "cgi.hxx"
#include "defines.hxx"
#include "registry.hxx"
#include "gate.hxx"
#include "zutils.hxx"

volatile GDT_BOOLEAN WaitingForSignal;

char *TempFile;

CHR *Action;
CHR *Host, *Tmp;
UINT Port;
CHR *DBName, *DBInfo;
STRLIST List;
INT PID;
enum { FROM_COMMANDLINE, FROM_CGI } Source;
STRING TempString;
struct stat statbuf;
CGI *cgi;
INT NumDatabases;
CHR *database_names[MAX_SERVERS], *database_values[MAX_SERVERS];
CHR *QuerySource;
CHR *RecSyntax, *ESName, *RSName, *MaxRecords;
CHR *AttrSet,*RecsPerServer;
GATE *ThisGate;
ofstream LogFile;

int 
main(int argc, char **argv)
{
  CGI *cgi;
  INT i;
  INT max_open_files;
  INT Code=0;
  CHR mypath[1024];
  CHR ErrorBuf[512];
  CHR *Tmp2;
  INT slen;
  INT nDBs;

  STRING TempString;
  CHR LockName[1024];
  FILE *flock;

  GDT_BOOLEAN LockStatus;

  pid_t ThisPID = getpid();

  getcwd(mypath, sizeof(mypath));
  ios::sync_with_stdio();

  cout << "Content-type: text/html" << endl << endl;

  if ((max_open_files = (int)sysconf(_SC_OPEN_MAX)) < 0) {
    cerr << "I can't determine the maximum number of open files!" << endl;
    LogFile.open("/tmp/zgate.log");
    LogFile << "[" << ThisPID << "] ";
    LogFile << "I can't determine the maximum number of open files!" << endl;
    LogFile.flush();
    MyExit(1);
  } else {
    for (i = 3; i < max_open_files; i++)
      close(i);
  }

  //  fclose(stderr);
  //  freopen("/tmp/zgate.err","a",stderr);

  LogFile.open("/tmp/zgate.log",ios::app | ios::ate | ios::out,0664);
  LogFile << "[" << ThisPID << "] ";
  LogFile << "Session started" << endl;
  LogFile.flush();
  if (!setlocale(LC_CTYPE,"")) {
    //    cerr << "Warning: Failed to set the locale!<BR>" << endl;
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Warning: Failed to set the locale!" << endl;
    LogFile.flush();
  }

  // Install the interrupt handler
  signal(SIGINT, ZgateStopHandler);

  // Install the user handler to talk to zcon
  signal(SIGUSR1, USR1Handler);

  // Install a segfault handler so we can figure out why we crashed
  signal(SIGSEGV, SegvHandler);

  cgi = cgi_Create();

  if(argc > 1) {
    // Read from command-line instead of CGI
    TempString = argv[1];
    Action = TempString.NewCString();
    if(!StrCaseCmp(Action,"PRESENT")){
      cgi_AddEntry(cgi,Action,"SERVICE");   
      cgi_AddEntry(cgi,argv[2],"SESSION_ID");
      
      TempString = argv[3];
      RSName = TempString.NewCString();
      cgi_AddEntry(cgi,RSName,"RSNAME");
      cgi_AddEntry(cgi,argv[4],"START");
      cgi_AddEntry(cgi,argv[5],"COUNT");
      TempString = argv[6];
      ESName = TempString.NewCString();
      cgi_AddEntry(cgi,ESName,"ESNAME");
      TempString = argv[7];
      RecSyntax = TempString.NewCString();
      cgi_AddEntry(cgi,RecSyntax,"RECSYNTAX");
    }
  } else {
    cgi_GetInput(cgi);
  }

#ifdef DEBUG
  cgi_DisplayContents(cgi);
#endif

  ThisGate=new GATE(mypath,cgi); 

  // Install an interrupt handler for an alarm signal
  signal(SIGALRM, ZgateAlarmHandler);
  alarm(ThisGate->GetGatewayTimeout());

  if((Action = cgi_GetValueByName(cgi, "SERVICE")) == NULL) {
    cout << "<I>Your form didn't include an SERVICE!</I><P>" << endl;
    MyExit(1);
  }

  LogFile << "[" << ThisPID << "] ";
  LogFile << "Got SERVICE=" << Action << endl;
  LogFile.flush();

  NumDatabases = ThisGate->LoadDBNames(database_names,database_values);
  nDBs = NumDatabases;

  if (StrCaseCmp(Action,"INIT")) {
    PCHR LockPath;
    PCHR LockName;
    INT SessionID;

    LogFile << "[" << ThisPID << "] ";
    LogFile << "Creating lock file" << endl;
    LogFile.flush();

    LockPath = ThisGate->GetLockPath();
    LockName = ThisGate->GetLockName();
    SessionID = ThisGate->CheckSession();

    PCHR LockNameFull;
    LockNameFull = new CHR[ strlen(LockPath)
                            +strlen(LockName)
                            +2 // two slashes
                            +7 // larger than largest PID
                            +1 // ASCIIZ terminator
                           ];

    if (SessionID != -1) {
      sprintf(LockNameFull,"/%s/%s%d", LockPath,LockName,SessionID);
      flock=fopen(LockName,"r");

      if (flock) {
	LockStatus=GDT_TRUE;
      } else {
	LockStatus=GDT_FALSE;
      }

      if (LockStatus) {
	fclose(flock);
	if (SessionID>0) {
	  kill(SessionID,SIGUSR2);
	}
	unlink(LockName);
	sleep(5);
      }
    }
    delete LockNameFull;
  }

#ifdef DEBUG
  LogFile << "[" << ThisPID << "] ";
  LogFile << "Main branch" << endl;
  LogFile.flush();
#endif

  if (!StrCaseCmp(Action, "INIT")) {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Calling GenerateInitRequest" << endl;
    LogFile.flush();
    Code=ThisGate->GenerateInitRequest();
  }

  else if (!StrCaseCmp(Action, "LOGOUT")) {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Calling GenerateLogOutRequest" << endl;
    LogFile.flush();
    Code=ThisGate->GenerateLogOutRequest();
  }

  else if (!StrCaseCmp(Action, "SCAN")) {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Calling GenerateScanRequest" << endl;
    LogFile.flush();
    Code=ThisGate->GenerateScanRequest();
  }

  else if(!StrCaseCmp(Action, "SEARCH")) {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Calling GenerateSearchRequest" << endl;
    LogFile.flush();
    Code=ThisGate->GenerateSearchRequest(nDBs);
  }

  else if(!StrCaseCmp(Action, "PRESENT")) {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Calling GeneratePresentRequest" << endl;
    LogFile.flush();
    Code=ThisGate->GeneratePresentRequest();
  }

  // trap for map click
  else if (((Tmp = cgi_GetValueByName(cgi, "map.x")) != NULL) &&
      ((Tmp2= cgi_GetValueByName(cgi, "map.y")) != NULL)) {

    LogFile << "[" << ThisPID << "] ";
    LogFile << "Got map request" << endl;
    LogFile.flush();

    //  there has been a mouse click in the map
    
    if ((Tmp = cgi_GetValueByName(cgi, "MAP_CLICKED")) != NULL) {
      // this was second mouse click
      // collect the state variables needed for query form
      // calculate new bounding box coordinates
      // display query form
    } else {
      // this is the first map click
      // generate new html page
    }
  }

  else if(!StrCaseCmp(Action, "MAP_WORLD")) {

    LogFile << "[" << ThisPID << "] ";
    LogFile << "Got MAP_WORLD request" << endl;
    LogFile.flush();

    CHR North[] =  "90", 
	South[] = "-90",
	East[]  = "180",
	West[]  = "-180";
	
    if((Tmp=cgi_GetValueByName(cgi,"SESSION_ID")) == NULL) {
      cout << "Form didn't include a SESSION_ID!";
      cout << endl;
      MyExit(1);
    }
    PID = atoi(Tmp);
    
    CHR MapString[150];
    CHR FormName[150];
    PCHR Motd=(PCHR)NULL;
    strcpy(FormName,FORMNAME);
    
    // redisplay query form with map of world, region of interest unchanged
    DisplaySearchForm(FormName, 
		      MapString,
		      South, North, West, East,
		      NumDatabases,
		      database_names, database_values,
		      PID, Motd);
  }

  else if(!StrCaseCmp(Action, "MAP_RESET")) {
    // redisplay query form with map & region of interest of default bounds
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Got MAP_RESET" << endl;
    LogFile.flush();
  }

  else if(!StrCaseCmp(Action, "MAP_REDRAW")) {
    // check for legal bounds
    // redisplay query form with map of world	
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Got MAP_REDRAW request" << endl;
    LogFile.flush();
  }

  else {
    LogFile << "[" << ThisPID << "] ";
    LogFile << "Unknown Request " << Action << endl;
    LogFile.flush();
  }

  if(Code<1) {			// error condition
    ThisGate->GetErrorMessage(ErrorBuf);

    ShowDocHeader();
    cout << "<p><em>Error:</em> "
         << ErrorBuf << "</p>"<< endl;
    cout << "<a href=\"" << GATEWAY_LINK
         << "\">Return</a> to the gateway." << endl;
    ShowDocTrailer();

    LogFile << "[" << ThisPID << "] ";
    LogFile << "Error: ";
    LogFile << ErrorBuf << endl;
    LogFile.flush();

  }
  
  LogFile << "[" << ThisPID << "] ";
  LogFile << "Exiting from main" << endl;
  LogFile.flush();
  MyExit(1);
  exit(1);
}
