#include "zs-classNT.hxx"

#define ZSERVER_INETD_OUTPUT "c:\\zserver.cerr"

extern volatile  INT   ChildCount;
extern           INT   GlobalDebugLevel;


CHR *
ResultSet_GetRecord(SAPI_RSET *c_result_set, INT i, LONG c_max_rec_size, 
		    LONG *rec_length, CHR *es_cname, 
		    CHR *c_rec_syntax,
		    STRING *actual_rec_syntax);


ZSERVER::ZSERVER(int argc, char **argv)
	: ZSESSION()
{
  if(StartedByInetd()) {
    //
    // Route all console output to a file.
    //
    c_output_file.open(ZSERVER_INETD_OUTPUT,ios::app);
    
    cerr = cout = c_output_file;
    // for now, disallow this because of incompatibilities between
    // the relevant iostream classes.....
    c_debuglevel = 0;
  }
  c_group = "Default";
  c_inifile = "zserver.ini";
  c_sapi = (SAPI *)NULL;
  c_default_db = "";
  STRING msg;
  
  cerr << endl << "CNIDR zserver, Version " << VERS;
  cerr << ", Copyright (c) 1995,1996,1997 MCNC/CNIDR\n" << endl;
  
  // Set defaults from zserver.ini in CWD (if available)
  LoadDefaultsFromFile(c_inifile, c_group);
  
  // Set defaults from command line
  STRING g, file;
  g = c_group;
  file = c_inifile;
  LoadDefaultsFromCommandLine(argc, argv, c_inifile, c_group,
			      &g, &file);
  
  c_group = g;
  c_inifile = file;
  ExpandFileSpec(&c_inifile);
  
  c_inipath = c_inifile;
  RemoveFileName(&c_inipath);
  
  //
  // Now that we've loaded all possible combination of user variables,
  // lets copy those variables from the generic REGISTRY structure
  // into private class variables for ease of use.  If any values
  // were not explicity set by the user, hard-coded defaults will
  // be inserted.
  //
  StoreDefaults(c_group);
  
  // Read list of databases to serve from zserver configuration file
  STRLIST Position, DBList;
  Position.AddEntry("Default");
  Position.AddEntry("DBList");
  c_defaults->GetData(Position, &DBList);
  c_database_count = DBList.GetTotalEntries();
  if(c_database_count <= 0) {
    cerr << "No databases configured for zserver [" <<
      c_inifile;
    cerr << "]." << endl;
    exit(1);
  }
  
  // Load SAPI configuration file
  c_defaults->ProfileGetString(c_group, "SAPI", "sapi.ini",
			       &c_sapifile);
  
  if(c_debuglevel >= 2) {
    cerr << "zserver.ini Path = " << c_inifile << endl;
    cerr << "SAPI File = " << c_sapifile << endl;
  }
  
  //
  // If the filename has a leading slash (Unix), assume it's a full path
  // or if the filename has a leading forward or backward slash, or a colon
  // for the second character (Windows-like c:\...), assume it's a full path
  // If neither, append it to the inipath to get a fully-qualified path
  //
  if ( !((c_sapifile.GetChr(2) == ':')
	  || (c_sapifile.GetChr(1) == '\\')
	  || (c_sapifile.GetChr(1) == '/')) )
	  c_sapifile.Insert(1,c_inipath);
  c_sapi = new SAPI(c_sapifile,DBList);
  
  if(!c_sapi) {
    cerr << "Failed to load CNIDR Search API [" <<
      c_sapifile;
    cerr << "]" << endl;
    exit(1);
  }
  c_sapipath = c_sapifile;
  RemoveFileName(&c_sapipath);
  
  if(c_debuglevel >= 2)
    cerr << "Looking for map files in " << c_sapipath <<
      endl;
  
  if(c_database_count <= 0) {
    cerr << "There are no databases to mount!  ";
    cerr << "Edit DBList in " << c_inifile << endl;
    exit(1);
  }
  
  // Load the default field mapping table, bib1;
  c_mapping_table = new REGISTRY("map");
  // Added for testing NT - aw3
  c_dblist = new PLIST();

  // Attempt mount the databases
  INT actual_db_count=0;
  STRING DBName;
  SAPI_DB *TempDbPtr;
  STRING Filename;
  REGISTRY *sapireg;
  STRING MapFile;
  STRLIST ResultList;
  STRINGINDEX count, j;

  for(INT i=1;i <= c_database_count;i++) {
    DBList.GetEntry(i, &DBName);
    cerr << endl;
    cerr << "Attempting to mount database " << DBName << endl;
    TempDbPtr = c_sapi->GetDatabasePtr(DBName);

    if (TempDbPtr == (SAPI_DB *)NULL) {
      cerr << "Database " << DBName << " not available!" << endl;
      break;
    }

    TempDbPtr->SetDebugLevel(c_debuglevel);
    if(!TempDbPtr->Initialize(c_password)) {
      cerr << "Failed to initialize " << DBName << ":";
      TempDbPtr->GetLastErrorString(&msg);
      cerr << msg << endl;
      continue;
    }
    c_dblist->AddTail(TempDbPtr);
    cerr << "Successfully mounted." << endl;
    if(i == 1)
      c_default_db = DBName;
    actual_db_count++;
    
    // Load the field mapping for this database
    /*
      // Store the default mapping table, bib1.map, for
      // this database.
      Filename = c_sapipath;
      if(c_sapipath.GetLength() > 0) {
      // does the path end in a forward or backward slash?
      char lastchar;
      lastchar = c_sapipath.GetChr(c_sapipath.GetLength());
      if ( !((lastchar == '/') 
      || (lastchar == '\\') 
      || (lastchar == ':')) )
      Filename.Cat("\\");
      }
      Filename.Cat("bib1.map");
      AddToMappingTable(Filename, DBName);
    */

    // Now step through any mapping files the user
    // has specified, overlaying the default bib1
    // mapping.
    
    // Get the (optional) list of mapping files
    // for this database from the sapi file.
    ResultList.Clear();
    sapireg = new REGISTRY("sapi");
    sapireg->ProfileLoadFromFile(c_sapifile, ResultList);
    Position.Clear();
    Position.AddEntry(DBName);
    Position.AddEntry("FieldMaps");
    sapireg->GetData(Position, &ResultList);
    count = ResultList.GetTotalEntries();

    if (count == 0) {
      // Store the default mapping table, bib1.map, for
      // this database.
      Filename = c_sapipath;
      if(c_sapipath.GetLength() > 0) {
	// does the path end in a forward or backward slash?
	char lastchar;
	lastchar = c_sapipath.GetChr(c_sapipath.GetLength());
	if ( !((lastchar == '/') 
	       || (lastchar == '\\') 
	       || (lastchar == ':')) )
	  Filename.Cat("\\");
      }
      Filename.Cat("bib1.map");
      AddToMappingTable(Filename, DBName);
    } else {
      for(j=1;j <= count;j++) {
	ResultList.GetEntry(j, &MapFile);
	MapFile.TrimLeading();
	if ( !((MapFile.GetChr(2) == ':')
	       || (MapFile.GetChr(1) == '\\')
	       || (MapFile.GetChr(1) == '/')) ) {
	  Filename = c_sapipath;
	  if (c_sapipath.GetLength() > 0) {
	    // does the path end in a forward or backward slash?
	    char lastchar;
	    lastchar = c_sapipath.GetChr(c_sapipath.GetLength());
	    if ( !((lastchar == '/') 
		   || (lastchar == '\\') 
		   || (lastchar == ':')) ) {
	      Filename.Cat("\\");
	    }
	    Filename.Cat(MapFile);
	  }
	} else {
	  Filename = MapFile;
	}
		AddToMappingTable(Filename, DBName);
      }
    }
    delete sapireg;
  }
  c_database_count = actual_db_count;
  TmpSapi=(SAPI *)NULL;
  if (c_database_count < 1) {
    cerr << "No local databases available!" << endl;
  }
}


ZSERVER::ZSERVER(const ZSERVER& OtherServer)
	: ZSESSION()
{
	Copy(OtherServer);
}


ZSERVER::ZSERVER()
	: ZSESSION()
{
}


void
ZSERVER::Copy(const ZSERVER& OtherServer)
{
  c_output_file = OtherServer.c_output_file;
  c_max_sessions = OtherServer.c_max_sessions;
  c_preferred_msg_size = OtherServer.c_preferred_msg_size;
  c_database_count = OtherServer.c_database_count;
  c_hitcount = OtherServer.c_hitcount;
  c_max_rec_size = OtherServer.c_max_rec_size;
  c_server_type = OtherServer.c_server_type;
  c_trace = OtherServer.c_trace;
  c_tracelog = OtherServer.c_tracelog;
  c_accesslog = OtherServer.c_accesslog;
  c_sapifile = OtherServer.c_sapifile;
  c_sapipath = OtherServer.c_sapipath;
  c_diagsetid = OtherServer.c_diagsetid;
  c_imp_id = OtherServer.c_imp_id;
  c_imp_name = OtherServer.c_imp_name;
  c_imp_version = OtherServer.c_imp_version;
  c_reverse_name_lookup = OtherServer.c_reverse_name_lookup;
  c_client_hostname = OtherServer.c_client_hostname;
  c_client_ipaddress = OtherServer.c_client_ipaddress;
  c_groupid = OtherServer.c_groupid;
  c_userid = OtherServer.c_userid;
  c_password = OtherServer.c_password;
  c_default_db = OtherServer.c_default_db;
  c_inipath = OtherServer.c_inipath;
  c_init_req = OtherServer.c_init_req;
  c_init_res = OtherServer.c_init_res;
  c_forking = OtherServer.c_forking;
  c_server_tcp = OtherServer.c_server_tcp;

  c_result_set = OtherServer.c_result_set;
  c_sapi = OtherServer.c_sapi;
  c_dblist = OtherServer.c_dblist;
  c_mapping_table = OtherServer.c_mapping_table;
  c_argc = OtherServer.c_argc;
  c_argv = OtherServer.c_argv;
  TmpSapi = OtherServer.TmpSapi;
  c_tcp = new TCPSOCK();
  c_tcp = OtherServer.c_tcp;
  c_debuglevel = OtherServer.c_debuglevel;
  c_state = OtherServer.c_state;
  c_prefmsgsize = OtherServer.c_prefmsgsize;
  c_maxrecordsize = OtherServer.c_maxrecordsize;
  c_timeout = OtherServer.c_timeout;
  c_port = OtherServer.c_port;
  c_protocol_version = OtherServer.c_protocol_version;
  c_options = OtherServer.c_options;
  c_refid = OtherServer.c_refid;
  c_group = OtherServer.c_group;
  c_inifile = OtherServer.c_inifile;
  c_dbname = OtherServer.c_dbname;
  c_defaults = OtherServer.c_defaults;
}

ZSERVER::~ZSERVER()
{
  if (c_sapi != NULL)
    delete c_sapi;
  if (c_tcp != NULL)
	delete c_tcp;
}


void 
ZSERVER::StartServer()
{
  STRING StrVal;
  
  //
  // Were we started by inetd or from the command line?
  //
  if (StartedByInetd()) {
    INT err;
    CHR msg[128];
    /* Can't do any debugging output until I fix the
       hexdir stuff to use streams instead of stderr.
       
       c_defaults->ProfileGetString(c_group, "DebugLevel", "1", &StrVal);
       c_debuglevel = StrVal.GetInt();
       */
    c_debuglevel = 0;
    c_tcp = new TCPSOCK();
    if((err = c_tcp->LastError())) {
      c_tcp->ErrorMessage(err, msg, (sizeof(msg) - 1));
      cerr << msg << endl;
      return;
    }
    c_tcp->SetSocket((TSOCKET)fileno(stdout));
    StartSession();
  } else {
    c_defaults->ProfileGetString(c_group, "DebugLevel", "1", &StrVal);
    c_debuglevel = StrVal.GetInt();
    if(c_debuglevel >= 2)
      cerr << endl << "DebugLevel=" << c_debuglevel << "." << endl;
    
#if defined (UNIX) || defined (_WIN32)
    if(c_server_type.CaseEquals("STANDALONE"))
      StartForkingServer(c_port);
    else
#endif
      StartServer(c_port);
  }
}


GDT_BOOLEAN
ZSERVER::ForkingInit()
{
  // Start the parent thread listening for a connection on the port
  if(!Listen(c_port)) {
    return GDT_FALSE;
  }
  return GDT_TRUE;
}


// This is the NT version of the forking server
INT 
ZSERVER::StartForkingServer(UINT Port)
{
  if(c_debuglevel >= 1)
    cerr << "Received connection - threaded server is responding..." << endl;
  StartSession();
  delete c_tcp; // allocated in AcceptClient()
  return 1;
}


INT 
ZSERVER::StartServer(UINT Port)
{
  if(Listen(Port)) {
    if(c_debuglevel >= 1) {
      cerr << "Single Connection Server Started, Port ";
      cerr << Port << "..." << endl;
    }
    if(!AcceptClient())
      return 0;
    
    // We don't need the main socket anymore, so kill it.
    delete c_server_tcp; // allocated in Listen()
    
    if(c_debuglevel >= 1)
      cerr << "Received connection..." << endl;
    StartSession();
    delete c_tcp; // allocated in AcceptClient()
    return 1;
  } else
    return 0;
}


INT 
ZSERVER::Listen(UINT Port)
{
  INT err;
  CHR msg[128];
  
  c_server_tcp = new TCPSOCK();
  if((err = c_server_tcp->LastError())) {
    c_server_tcp->ErrorMessage(err, msg, (sizeof(msg) - 1));
    cerr << msg << "\n";
    return 0;
  }
  
  c_server_tcp->Listen(Port, 5);
  if((err = c_server_tcp->LastError())) {
    c_server_tcp->ErrorMessage(err, msg, (sizeof(msg) - 1));
    cerr << "Listen failed: " << msg << endl;
    return 0;
  }
  if(c_debuglevel >= 2)
    cerr << "Listening on port " << Port << "..." << endl;
  if(c_debuglevel >= 7) {
    cerr << "Listening to socket # " <<
      c_server_tcp->Socket();
    cerr << "..." << endl;
  }
  
  c_server_tcp->BlockingModeON();
  return 1;
}


// 1 if connection, 0 otherwise.
//
// Blocks until connection arrives
//
INT 
ZSERVER::AcceptClient()
{
  INT err;
  CHR msg[128];
  c_tcp = new TCPSOCK();  

  if (c_debuglevel >= 5)
    cerr << endl << "Waiting to Accept client..." << endl;
  
  if(c_server_tcp->Accept(c_tcp) > 0) {
    if(c_debuglevel >= 1)
      cerr << "Connection has arrived..." << endl;
    
    c_tcp->GetHostname(&c_client_hostname);
    c_tcp->GetIPAddress(&c_client_ipaddress);
    
    // Allocate the tcp stream to be used for this client
    // connection.
    if((err = c_tcp->LastError())) {
      c_tcp->ErrorMessage(err, msg,
			  (sizeof(msg) - 1));
      cerr << msg << endl;
      return 0;
    }
    
    return 1;
  }
  if((err = c_tcp->LastError())) {
    c_tcp->ErrorMessage(err, msg, (sizeof(msg) - 1));
    cerr << msg << endl;
    return 0;
  }
  return 0;
}


void 
ZSERVER::StartSession()
{
  unsigned short type;
  INT4 len, BytesSent;
  CHR TempBuf[64];
  UCHR *buf;
  CHR *TempString;
  INT Finished = 0;
  time_t Start, Done;
  
  time(&Start);
  do {
    if((buf = ReadPDU(&len, &type)) == NULL)  {
      // Client closed on us without a proper CLOSE
      time(&Done);
      sprintf(TempBuf, "%ld", Done - Start);
      Log("DISCONNECT", TempBuf, 1, 0);
      c_tcp->Close();
      FreeSapi();
      return;  // rs leaks here, allocated near line 1108
    }

    switch(type) {
    case INITREQUEST_TAG:
      {
	ZINITREQUEST *rq;		// Init Request PDU
	INT4 Result = 1;	// Assume success
	  
	// Cast buffer to PDU
	rq = new ZINITREQUEST(buf, len);
	
	// Negotiate based on request parameters
	
	//
	// Grab the Reference ID
	//
	c_refid = "";
	rq->GetReferenceId(&c_refid);
	
	//
	// We support versions 2 and 3 (default=2)
	//
	rq->GetProtocolVersion(&c_protocol_version);
	c_protocol_version.UpperCase();
	if((!c_protocol_version.Equals("YY")) &&
	   (!c_protocol_version.Equals("YYY")))
	  c_protocol_version = "YY";
	
	//
	// We support Init, Search and Present.
	// If you need their options, here they
	// are.  Ours were set in our  constructor.
	//
	STRING Value;
	rq->GetOptions(&Value);
	
	//
	// Negotiate preferred message size
	//
	rq->GetPreferredMessageSize(&c_prefmsgsize);
	if(c_prefmsgsize > ZPREFMSGSIZE)
	  c_prefmsgsize = ZPREFMSGSIZE;
	if(c_prefmsgsize < 1024)
	  c_prefmsgsize = 1024;
	
	//
	// Negotiate maximum message size
	//
	rq->GetExceptionalRecordSize(&c_maxrecordsize);
	if(c_maxrecordsize > ZMAXRECORDSIZE)
	  c_maxrecordsize = ZMAXRECORDSIZE;
	if(c_maxrecordsize < 1024)
	  c_maxrecordsize = 1024;
	
	//
	// GroupId/UserId/Password?
	//
	rq->GetGroupId(&c_groupid);
	rq->GetUserId(&c_userid);
	rq->GetPassword(&c_password);
	
	//
	// If you want to do authentication,
	// do it here.  We'll add support later.
	//
	
	rq->GetAuthentication(&c_password);
	delete rq;
	
	//
	// Build the response
	//
	ZINITRESPONSE *rs;
	rs = new ZINITRESPONSE(c_refid,
			       c_protocol_version,
			       c_options,
			       c_prefmsgsize,
			       c_maxrecordsize,
			       Result,
			       CNIDR_IMP_ID,
			       ZSERVER_NAME,
			       IsiteVersion);
	
	if((BytesSent = SendPDU(rs)) == 0)
	  Finished = 1;
	delete rs;
	
	Log("INIT", "-", 1, BytesSent);
	break;
      }
    case SEARCHREQUEST_TAG:
      {
	ZSEARCHREQUEST *rq;		// Search Request PDU
	  
	rq = new ZSEARCHREQUEST(buf, len);
	
	// Perform the search
	ZSEARCHRESPONSE *rs; // Search Response PDU
	rs = Search(*rq);
	delete rq;
	
	if((BytesSent = SendPDU(rs)) == 0)
	  Finished = 1;
	
	// Get result of search for logging purposes
	INT4 Result;
	rs->GetResultCount(&Result);
	TempString = c_dbname.NewCString();
	
	Log("SEARCH", TempString, Result, BytesSent);
	
	delete [] TempString;
	delete rs;
	
	break;
      }
    case PRESENTREQUEST_TAG:
      {
	ZPRESENTREQUEST *rq;	// Present Request PDU
	  
	rq = new ZPRESENTREQUEST(buf, len);
	
	ZPRESENTRESPONSE *rs;	// Present Response PDU
	rs = Present(*rq);
	if(!rs) {
	  cerr << "Fatal Error. No PRESENTRESPONSE generated" << endl;
	  exit(1);
	}
	delete rq;
	
	if((BytesSent = SendPDU(rs)) == 0)
	  Finished = 1;
	
	INT4 Status;
	rs->GetPresentStatus(&Status);
	TempString = c_dbname.NewCString();
	
	Log("PRESENT", TempString, Status, BytesSent);
        
	delete [] TempString;
	delete rs;
	
	break;
      }
    case EXTENDEDSERVICESREQUEST_TAG:
      {
	ZESREQUEST *rq;	// Extended Service Request PDU
	
	rq = new ZESREQUEST(buf, len);
	
	INT4 Function;
	STRING PackageType;
	STRING PackageName;
	STRING UserId;
	INT4 RetentionTime;
	STRING Description;
	INT4 WaitAction;

	rq->GetFunction(&Function);
	
	rq->GetPackageType(&PackageType);
	rq->GetPackageName(&PackageName);
	rq->GetUserId(&UserId);
	rq->GetRetentionTime(&RetentionTime);
	rq->GetDescription(&Description);
	rq->GetWaitAction(&WaitAction);

	delete rq;
	
	ZESTASKPACKAGE *TaskPackage;
	
	RetentionTime = 0;

	TaskPackage = (ZESTASKPACKAGE *)NULL;
	if ( WaitAction != ES_DONT_RETURN_PACKAGE_WAITACTION )
	  TaskPackage = new ZESTASKPACKAGE(
					   PackageType,
					   PackageName,
					   UserId,
					   RetentionTime,
					   Description,
					   ES_TP_COMPLETE_TASKSTATUS);
	
	ZESRESPONSE *rs;// Extended Service Response PDU
	/*	
	TempString = "UNSUPPORTED";
	
	ZDEFAULTDIAGFORMAT *diagrec;	
	
	diagrec = new ZDEFAULTDIAGFORMAT(221, PackageType, BIB1_DIAG_OID);
	rs = new ZESRESPONSE("", ES_FAILURE_OPERATIONSTATUS, diagrec, NULL);
	delete diagrec;
	*/

       if ( PackageType.Equals(ES_UPDATE_OID) ) {
         TempString = "UPDATE";
  
         if ( Function == ES_CREATE_FUNCTION ) {
           ZESUPDATEREQUEST *request;
 
           request = new ZESUPDATEREQUEST(buf,len);
 
           rs = ESUpdate(TaskPackage, *request);
 
           delete request;
         } else {
           ZDEFAULTDIAGFORMAT *diagrec;        
 
           diagrec = new ZDEFAULTDIAGFORMAT(219, PackageName, BIB1_DIAG_OID);
           rs = new ZESRESPONSE("", ES_FAILURE_OPERATIONSTATUS, diagrec,NULL);
           delete diagrec;
         }
       } else {
         TempString = "UNSUPPORTED";
 
         ZDEFAULTDIAGFORMAT *diagrec;  
 
         diagrec = new ZDEFAULTDIAGFORMAT(221, PackageType, BIB1_DIAG_OID);
         rs = new ZESRESPONSE("", ES_FAILURE_OPERATIONSTATUS, diagrec, NULL);
         delete diagrec;
       }
 
	if ( TaskPackage )
	  delete TaskPackage;

	if((BytesSent = SendPDU(rs)) == 0)
	  Finished = 1;

	INT4 OperationStatus;

	rs->GetOperationStatus(&OperationStatus);

	Log("EXTENDSERVICE", TempString, OperationStatus, BytesSent);

	delete rs;
	
	break;
      }
    case CLOSE_TAG:
      {
	// Cast buffer to PDU
	ZCLOSE *pdu;
	pdu = new ZCLOSE(buf, len);
	
	// Why did they close?
	INT4 Reason;
	pdu->GetCloseReason(&Reason);
	if(c_debuglevel >= 1) {
	  cerr << "Client has closed. Reason: ";
	  switch(Reason) {
	  case 0:
	    cerr << "Finished";
	    break;
	  case 1:
	    cerr << "Shutdown";
	    break;
	  case 2:
	    cerr << "System Problem";
	    break;
	  case 3:
	    cerr << "Cost Limit";
	    break;
	  case 4:
	    cerr << "Resources";
	    break;
	  case 5:
	    cerr << "Security Violation";
	    break;
	  case 6:
	    cerr << "ProtocolError";
	    break;
	  case 7:
	    cerr << "Lack of Activity";
	    break;
	  case 8:
	    cerr << "Peer Abort";
	    break;
	  case 9:
	    cerr << "Unspecified";
	    break;
	  default:
	    cerr << "Unknown reason";
	    break;
	  }
	  cerr << endl;
	}
	
	// We're done!
	Finished = 1;
	time(&Done);
	sprintf(TempBuf, "%ld", Done - Start);
	Log("CLOSE", TempBuf, 1, 0);
	FreeSapi();
	delete pdu;
	break;
      }
    default:
      cerr << "Unsupported PDU of " << type <<
	endl;
    }
    delete [] buf;
  } while(!Finished);
}


ZSEARCHRESPONSE *
ZSERVER::Search(ZSEARCHREQUEST & Request)
{
  ZSEARCHRESPONSE *rs;
  INT hitcount;
  //INT i;
  //CHR *p;
  BERBROWSER *b;
  ZRECORDLIST *Records;
  ZDEFAULTDIAGFORMAT *diagrec;
  SAPI_DB *Database;
  
  // We only support getting a single database name
  Request.GetDatabaseName(&c_dbname);
  if (c_dbname.CaseEquals("xxdefault"))
    c_dbname = c_default_db;
  
  // Get the list of local databases already mounted
  IPOSITION Pos;
  Pos = c_dblist->GetHeadPosition();
  
  // here is where we fiddle with databases....  
  // split the list into a set of databases, and load a fake sapi.

  // if it has a plus, do MetaSearch, and Set Database
  if (c_dbname.Search(':') >= 1) {
    Database=MetaSearch(c_dbname);
  } else if (c_dbname.Search('+') >= 1) {
    Database=MetaSearch(c_dbname);
  } else if (c_dbname.Search(',') >= 1) {
    Database=MetaSearch(c_dbname);
  } else {
    // Otherwise, it's a local index, of which we can search just one.
    while((Database = (SAPI_DB *)c_dblist->GetNext(&Pos))) {
      STRING name;
      Database->GetName(&name);
      if(c_debuglevel > 5) 
	cout << "checking against " << name << endl;
      if(c_dbname.CaseEquals(name))
	break;
      //      delete Database;
    }
  }
  
  if(Database == NULL) {
    // illegal database name requested
    if(c_debuglevel >= 2) {
      cerr << "Illegal database name requested: " <<
	c_dbname;
      cerr << endl;
    }
    // Construct a non-surrogate diagnostic record and return
    // with response
    Records = new ZRECORDLIST(NONSURROGATEDIAGNOSTIC_TAG);
    diagrec = new ZDEFAULTDIAGFORMAT(109, c_dbname,
				     BIB1_DIAG_OID);
    Records->AddRecord(diagrec);
    rs = new ZSEARCHRESPONSE(c_refid, 0, 0, 0, GDT_FALSE, 3,
			     5,Records);
    delete diagrec;
    //    delete Records;    // segfaults on bad name in destructor
    return rs;
  }
  
  // If this database supports OPSTACK queries, then perform custom query :-)
  STRING DbType;
  ZSQUERY *Query;
  Database->GetType(&DbType);

  if(c_debuglevel>=2)
    cerr << "Searching " << DbType << " Database " << c_dbname << endl;

  if (DbType.CaseEquals("ISEARCH")) {
    if(c_debuglevel >= 2) {
      cerr << "Searching using a OPSTACK based query" << endl;
    }

    STRLIST FieldsAvailable;

    // Figure out which fields _are_ available
    Database->GetFieldNames(&FieldsAvailable);

    if(FieldsAvailable.GetTotalEntries() == 0)
      cerr << "No fields available!!" << endl;

    // ConvertQuery puts BER into a stack 
    
    b = new BERBROWSER(&Request);
    Query = new ZSQUERY();
    Query->SetDebugLevel(c_debuglevel);
    Query->ConvertQuery(b, c_mapping_table, FieldsAvailable, c_dbname);
    delete b;
    
    if(Query->Error()) {
      INT code;
      STRING addinfo;
      Query->GetErrorInfo(&code, &addinfo);
      // build diag rec here
      Records = new
	ZRECORDLIST(NONSURROGATEDIAGNOSTIC_TAG);
      diagrec = new ZDEFAULTDIAGFORMAT(code, addinfo, BIB1_DIAG_OID);
      Records->AddRecord(diagrec);
      rs = new ZSEARCHRESPONSE(c_refid, 0, 0, 0, GDT_FALSE,
			       3, 5, Records);
      delete diagrec;
      delete Query;
//      delete Records;
      return rs;
    }

  } else {
    if(c_debuglevel >= 2) {
      cerr << "Searching using a KWAQS based query" << endl;
    }
    
    // Convert from Z39.50 structure to KWAQS
    CHR *query_str;
    b = new BERBROWSER(&Request);
    KWAQS_STRING *q;
    q = new KWAQS_STRING;
    q->SetDebugLevel(c_debuglevel);
    q->Convert(b, Request);
    if(q->Error()) {
      INT code;
      STRING addinfo;
      q->GetErrorInfo(&code, &addinfo);
      // build diag rec here
      Records = new
	ZRECORDLIST(NONSURROGATEDIAGNOSTIC_TAG);
      diagrec = new ZDEFAULTDIAGFORMAT(code, addinfo, BIB1_DIAG_OID);
      Records->AddRecord(diagrec);
      rs = new ZSEARCHRESPONSE(c_refid, 0, 0, 0, GDT_FALSE,
			       3, 5, Records);
      delete diagrec;
      //      delete Records;
      return rs;
    }
    
    // put KWAQS_STRING q into a query.  Should put it in a stack
    
    query_str = q->NewCString();
    STRING t=query_str;
    delete query_str;
    Query = new ZSQUERY();
    
    Query->SetKWAQSTerm(t);
  }
  
  INT4 SmallUpper,LargeLower,MediumCount;
  STRING rec_syntax,es_name;
  INT start_point=0,num_requested;
  
  Request.GetMediumSetPresentNumber(&MediumCount);
  Query->SetHitsRequested(MediumCount);
  c_result_set = Database->Search(*Query);

  if ( !c_result_set ) {
    STRING addinfo;
    
    Database->GetLastErrorString(&addinfo);
 
    Records = new
      ZRECORDLIST(NONSURROGATEDIAGNOSTIC_TAG);
    diagrec = new ZDEFAULTDIAGFORMAT(100, addinfo, BIB1_DIAG_OID);
    Records->AddRecord(diagrec);
    rs = new ZSEARCHRESPONSE(c_refid, 0, 0, 0, GDT_FALSE,
			     3, 5, Records);
    delete diagrec;
    delete Query;
    return rs;
  }
 
  hitcount = c_result_set->GetHitCount();
  c_hitcount = hitcount;
  
  if(c_debuglevel >= 5)
    cerr << "There were " << hitcount << " hits." << endl;
  
  Request.GetSmallSetUpperBound(&SmallUpper);
  Request.GetLargeSetLowerBound(&LargeLower);
  Request.GetMediumSetPresentNumber(&MediumCount);
  Request.GetPreferredRecordSyntax(&rec_syntax);
  
  es_name="B";

  if (hitcount<=SmallUpper) {
    start_point=1;
    num_requested=hitcount;
  } else if(hitcount<LargeLower) {
    start_point=1;
    num_requested=((hitcount < MediumCount) ? hitcount : MediumCount);
  } else
    num_requested=0;

  /*  
      ZRESPONSERECORDS records;
      BuildReturnRecordList(&records,es_name,rec_syntax,start_point,
      num_requested);
  
      rs = new ZSEARCHRESPONSE(c_refid, num_requested, 0, 1, GDT_TRUE,
      1,0, &records);
  */

  ZRESPONSERECORDS *records;

  records = new ZRESPONSERECORDS;

  BuildReturnRecordList(records,es_name,rec_syntax,start_point,
			num_requested);
  
  rs = new ZSEARCHRESPONSE(c_refid, hitcount, num_requested, start_point, GDT_TRUE,
			   1,0, records);

  delete records;
  delete Query;
  return rs;
}

//
// After we've loaded all possible combinations of user variables from files
// and the command line, copy them from the generic REGISTRY structure
// into private class variables for ease of use.  If an attribute doesn't
// have a value, a hard-coded default is added.
//
void 
ZSERVER::StoreDefaults(const STRING & Group)
{
  //STRLIST Pos, ValList;
  STRING StrVal;
  
  c_defaults->ProfileGetString(Group, "Port", "210", &StrVal);
  c_port = StrVal.GetInt();
  
  c_defaults->ProfileGetString(Group, "DebugLevel", "1", &StrVal);
  c_debuglevel = StrVal.GetInt();
  
  c_defaults->ProfileGetString(Group, "ServerType", "STANDALONE",
			       &c_server_type);
  
  c_defaults->ProfileGetString(Group, "MaxSessions", "10",
			       &StrVal);
  c_max_sessions = StrVal.GetInt();
  
  c_defaults->ProfileGetString(Group, "TimeOut", "3600", &StrVal);
  c_timeout = StrVal.GetInt();
  
  c_defaults->ProfileGetString(Group, "Trace", "OFF", &c_trace);
  
  c_defaults->ProfileGetString(Group,
			       "TraceLog","/tmp/zserver_trace.log",
			       &c_tracelog);
  
  c_defaults->ProfileGetString(Group, "AccessLog",
			       "/tmp/zserver_access.log", &c_accesslog);
  
  c_defaults->ProfileGetString(Group, "ReverseNameLookup", "ON",
			       &c_reverse_name_lookup);
  
  c_defaults->ProfileGetString(Group, "SAPI", "sapi.ini",
			       &c_sapifile);
  
  c_defaults->ProfileGetString(Group, "DiagSetId",
			       "1.2.840.10003.3.1",
			       &c_diagsetid);
  
  c_defaults->ProfileGetString(Group, "iImpID", "34", &c_imp_id);
  c_defaults->ProfileGetString(Group, "iImpName", "CNIDR zserver",
			       &c_imp_name);
  c_defaults->ProfileGetString(Group, "iImpVersion",
			       IsiteVersion,
			       &c_imp_version);
  
  c_defaults->ProfileGetString(Group, "iPreferredMsgSize",
			       "32768",
			       &StrVal);
  c_preferred_msg_size = StrVal.GetInt();
  
  c_defaults->ProfileGetString(Group, "iMaxRecSize", "8388608",
			       &StrVal);
  c_max_rec_size = StrVal.GetLong();
  
}


CHR *
ResultSet_GetRecord( SAPI_RSET *c_result_set, INT i, LONG c_max_rec_size, 
		     LONG *rec_length, CHR *es_cname, 
		     CHR *c_rec_syntax,
		     STRING *actual_rec_syntax)
{
  
  CHR *rec=new CHR[1024];
  PLIST RecordList;
  rec[0]='\0';
  STRING esn;
  
//  cout << "Get Record es: " << es_cname<<endl;
  esn=es_cname;

  if (!c_result_set->Present(i, 1, esn, c_rec_syntax, &RecordList)) {
    sprintf(rec,"Record %d failed es = %s  MaxSize = %ld",i,es_cname,
	    c_max_rec_size);

  } else {
    IPOSITION p;
    p = RecordList.GetHeadPosition();
    MIME_RECORD *r;
    STRING value;
    delete [] rec;
    while((r = (MIME_RECORD *)RecordList.GetNext(&p))) {
      r->GetData(&value);
      rec=value.NewCString();
      if((INT)strlen(rec)>c_max_rec_size){
	delete rec;
	rec=new CHR[c_max_rec_size];
	value.GetCString(rec,c_max_rec_size-1);
      }
      
    }
    
  }
  *rec_length=strlen(rec);
  *actual_rec_syntax=c_rec_syntax;
  return rec;
}


ZPRESENTRESPONSE *
ZSERVER::Present(ZPRESENTREQUEST & Request)
{
  ZRESPONSERECORDS *records;
  records = new ZRESPONSERECORDS();
  
  //
  // Version 3 conformance states we must recognize
  // AdditionalRanges.
  //
  if(Request.HasAdditionalRanges())
    return PresentResponseWithDiagnostic(BIB1_DIAG_OID, 243, "");
  
  //
  // What range of records being requested?
  INT4 start_point, num_requested;
  Request.GetResultSetStartPoint(&start_point);
  Request.GetNumberOfRecordsRequested(&num_requested);
  
  if(c_debuglevel >= 5) {
    cerr << "Client has asked for records " << start_point;
    cerr << " through " << (start_point + num_requested-1)
      << endl;
  }
  
  if(start_point < 1 || (start_point + (num_requested - 1)) > c_hitcount)
    return PresentResponseWithDiagnostic(BIB1_DIAG_OID, 13, "");

//  INT i;
//  CHR *record;
//  INT4 rec_length;
//  CHR *es_cname;
  CHR *c_rec_syntax;
  STRING es_name, rec_syntax;
  
  //
  // Element Set Name?
  //
  // We must recognize COMPLEX type
  //
  if(Request.HasCompSpec())
    return PresentResponseWithDiagnostic(BIB1_DIAG_OID, 244, "");
  
  //
  // Default to "B"
  //
  Request.GetElementSetName(&es_name);
  if(es_name.Equals(""))
    es_name = "B";
  if(c_debuglevel >= 5) {
    cerr << "Client has asked for element set name of ";
    cerr << es_name << endl;
  }
  
  // Preferred Record Syntax?
  Request.GetPreferredRecordSyntax(&rec_syntax);
  if(c_debuglevel >= 5) {
    cerr << "Client has asked for record syntax of ";
    cerr << rec_syntax << endl;
  }
  
  // Default to SUTRS if client didn't specify
  if(rec_syntax.Equals(""))
    rec_syntax = SUTRS_OID;
  c_rec_syntax = rec_syntax.NewCString();
  
  
  BuildReturnRecordList(records,es_name,rec_syntax,start_point,
			num_requested);
  
  return new ZPRESENTRESPONSE(c_refid, records->GetRecordCount(),
			      (start_point + records->GetRecordCount()), 0, 
			      records);
}


void 
ZSERVER::BuildReturnRecordList(ZRESPONSERECORDS *records, STRING es_name, 
			       STRING rec_syntax,INT start_point, 
			       INT num_requested)
{
  STRING temprecord;
  CHR *es_cname,*c_rec_syntax;
  INT i;
  CHR *record;
  LONG rec_length;
  STRING actual_rec_syntax = rec_syntax;
  c_rec_syntax = rec_syntax.NewCString();
  for(i=start_point;i < (start_point+num_requested);i++) {
    es_cname = es_name.NewCString();
    //    cout << "es: " << es_cname <<" "<<c_rec_syntax<<" "
    //<<c_result_set<<" " << c_max_rec_size<<endl;
    ZNAMEPLUSRECORD *npr;

    // make sure we haven't exceeded the maxrecsize for complete pdu
    record = (CHR *)ResultSet_GetRecord(c_result_set, i,
					c_max_rec_size, &rec_length, 
					(CHR *)es_cname,
					(CHR *)c_rec_syntax, 
					&actual_rec_syntax);

    if((!record)||(!*record)) {
      // Hmm.  If they asked for something other than
      // "B" or "F", lets try a default for them
      if((!es_name.CaseEquals("B"))&&
	 (!es_name.CaseEquals("F"))) {
	delete [] es_cname;
	es_cname = new CHR[2];
	strcpy(es_cname, "F");
	
	record = (CHR *)ResultSet_GetRecord(c_result_set, i, 
					    c_max_rec_size,
					    &rec_length, 
					    (CHR *)es_cname,
					    (CHR *)c_rec_syntax,
					    &actual_rec_syntax);

	//strcpy(record, "test");
	if(!record) {
	  // I give up
	  record = new CHR[80];
	  sprintf(record, "Failed to retrieve record #%i", i);
	} else {
	  // Should I warn them???
	}
      } else {
	record = new CHR[80];
	sprintf(record, "Failed to retrieve record #%i", i);
      }
    }
    
    if(actual_rec_syntax.Equals(USMARC_OID)) {
      ZUSMARCRECORD *usm;
      temprecord = record;
      usm = new ZUSMARCRECORD(temprecord);
      npr = new ZNAMEPLUSRECORD(c_dbname, usm);
    } else {
      // Default to SUTRS (??)
      ZSUTRSRECORD *s;
      if(c_debuglevel >= 7) {
	cout << "SUTRS Requested "<<endl;
      }
      
      s = new ZSUTRSRECORD(record);
      npr = new ZNAMEPLUSRECORD(c_dbname, s);
    }
    records->AddRecord(npr);
    delete [] record;
    delete [] es_cname;
  }
  delete [] c_rec_syntax;
}


ZESUPDATERESPONSE *
ZSERVER::ESUpdate(ZESTASKPACKAGE *TaskPackage, ZESUPDATEREQUEST &Request)
{
  INT4 action;
  SAPI_DB *Database;
  ZDEFAULTDIAGFORMAT *diagrec;
  ZESUPDATERESPONSE *rs;
  
  // We only support getting a single database name
  
  Request.GetAction(&action);
  Request.GetDatabaseName(&c_dbname);

  if ( c_dbname.CaseEquals("xxdefault") )
    c_dbname = c_default_db;
  
  // Get the list of local databases already mounted
  IPOSITION Pos;
  Pos = c_dblist->GetHeadPosition();
   
  // here is where we fiddle with databases....  
  // split the list into a set of databases, and load a fake sapi.
 
  // if it has a plus, do MetaSearch, and Set Database
  if (c_dbname.Search(':') >= 1) {
    Database = MetaSearch(c_dbname);
  }
  else if (c_dbname.Search('+') >= 1) {
    Database = MetaSearch(c_dbname);
  }
  else {
    // Otherwise, it's a local index, of which we can search just one.
    while ( (Database = (SAPI_DB *)c_dblist->GetNext(&Pos)) ) {
      STRING name;
      Database->GetName(&name);
      cout << "checking against " << name << endl;
      if ( c_dbname.CaseEquals(name) )
	break;
    }
  }
   
  if (Database == NULL) {
    cerr << "Update request specified an unknown database name: ";
    cerr << c_dbname << endl;
 
    diagrec = new ZDEFAULTDIAGFORMAT(109, c_dbname, BIB1_DIAG_OID);
    rs = new ZESUPDATERESPONSE("", TaskPackage, action, c_dbname,
			       ES_UPDATE_FAILURE_UPDATESTATUS, diagrec);
    delete diagrec;
    return rs;
  }
 
  // How many records to be process are included with the request?
  ASN1TAGU exttag(ASN1_EXTERNAL);
  ASN1TAGU oidtag(ASN1_OBJECTIDENTIFIER);
  BERBROWSER ber(&Request);
  INT4 count=0;
  STRING msg;
  STRING OID;
 
  if ( !ber.GetSubdirectory(ES_TASKSPECIFIC_TAG, &ber) ||
       !ber.GetSubdirectory(exttag, &ber) ||
       !ber.GetOID(oidtag, &OID) || OID != ES_UPDATE_OID ||
       !ber.GetSubdirectory(ES_TS_ESREQUEST_TAG, &ber) ||
       !ber.GetSubdirectory(ES_TS_NOTTOKEEP_TAG, &ber) ||
       (count = ber.GetSubdirectoryCount()) == 0) {
    msg = "Update request didn't include any records.";
 
    cerr << msg << endl;
    diagrec = new ZDEFAULTDIAGFORMAT(226, msg, BIB1_DIAG_OID);
    rs = new ZESUPDATERESPONSE("", TaskPackage, action, c_dbname,
			       ES_UPDATE_FAILURE_UPDATESTATUS, diagrec);
    delete diagrec;
    return rs;
  }
 
  // Process the records.
  BERBROWSER rec(ber);
  BERBROWSER sub(ber);
  INT4 nupdated;
  STRING Data;
  STRING Key;
 
#if !defined(_MSDOS) && !defined(_WIN32)
  sigset_t original_signal_set;
  sigset_t signal_set;

  sigemptyset(&signal_set);
  sigaddset(&signal_set, SIGHUP);
  sigaddset(&signal_set, SIGINT);
  sigaddset(&signal_set, SIGTERM);
#endif

  nupdated = 0;
 
  diagrec = (ZDEFAULTDIAGFORMAT*)NULL;
 
  for (INT4 i = 0; i < count; i++) {
    if ( !ber.GetSubdirectoryX(i, &rec) )
      continue;
 
    Key = "";
    Data = "";
 
    if ( rec.GetSubdirectory(ES_UPDATE_RECORDID_TAG, &sub) ) {
      if ( !sub.HasTag(ES_UPDATE_STRING_RECORDID_TAG) ||
	   !sub.GetChar(ES_UPDATE_STRING_RECORDID_TAG, &Key) ) {
	msg = "Update request included an unusable record key.";
 
	cerr << msg << endl;
	diagrec = new ZDEFAULTDIAGFORMAT(226, msg, BIB1_DIAG_OID);
	break;
      }
    }
 
    if ( rec.GetSubdirectory(ES_UPDATE_RECORD_TAG, &sub) ) {
      if ( !sub.GetSubdirectory(exttag, &sub) ||
	   !sub.GetOID(oidtag, &OID) ) {
	msg = "Update request included an unusable record.";
 
	cerr << msg << endl;
	diagrec = new ZDEFAULTDIAGFORMAT(226, msg, BIB1_DIAG_OID);
	break;
      }
 
      if (OID == SUTRS_OID) {
	//
	// SUTRS isn't encoded very
	// consistently across servers,
	// unfortunately.  I have to test
	// for every case here.
	//
	if ( sub.HasTag(0) ) {
	  if ( !sub.GetSubdirectory(0, &sub) ) {
	    msg = "Update request included an unusable SUTRS record.";
 
	    cerr << msg << endl;
	    diagrec = new ZDEFAULTDIAGFORMAT(226, msg, BIB1_DIAG_OID);
	    break;
	  }
	  ASN1TAGU IntTag(ASN1_GENERALSTRING);
	  sub.GetChar(IntTag, &Data);
	}
	else if ( sub.HasTag(1) ) {
	  sub.GetChar(1, &Data);
	}
	else {
	  msg = "Update request included an unusable SUTRS record.";
 
	  cerr << msg << endl;
	  diagrec = new ZDEFAULTDIAGFORMAT(226, msg, BIB1_DIAG_OID);
	  break;
	}
      }
      else {
	msg = "Update request included an unsupported record type.";
 
	cerr << msg << endl;
	diagrec = new ZDEFAULTDIAGFORMAT(226, msg, BIB1_DIAG_OID);
	break;
      }
    }
 
    if (action == ES_UPDATE_RECORD_INSERT_ACTION) {
      if (Database->Add_Record(Key, Data) != GDT_TRUE) {

	Database->GetLastErrorString(&msg);
	diagrec = new ZDEFAULTDIAGFORMAT(224, msg, BIB1_DIAG_OID);
	break;
      }
    }
    else if (action == ES_UPDATE_RECORD_DELETE_ACTION) {
      if (Key == "") {
	msg = "Update request didn't include key to delete.";
 
	cerr << msg << endl;
	diagrec = new ZDEFAULTDIAGFORMAT(224, msg, BIB1_DIAG_OID);
	break;
      }
      if (Database->Delete_Record(Key) != GDT_TRUE) {
        Database->GetLastErrorString(&msg);
        diagrec = new ZDEFAULTDIAGFORMAT(224, msg, BIB1_DIAG_OID);
        break;
      }
    }
    else {
      msg = "Update request included an unsupported action.";

      cerr << msg << endl;
      diagrec = new ZDEFAULTDIAGFORMAT(226, msg, BIB1_DIAG_OID);
      break;
    }

    nupdated++;
  }

  INT4 update_status;

  update_status = (nupdated == count) ? ES_UPDATE_SUCCESS_UPDATESTATUS :
    nupdated ? ES_UPDATE_PARTIAL_UPDATESTATUS : ES_UPDATE_FAILURE_UPDATESTATUS;

  rs = new ZESUPDATERESPONSE("", TaskPackage, action, c_dbname,
			     update_status, diagrec);

  if ( diagrec )
    delete diagrec;

  return rs;
}


KWAQS_STRING::KWAQS_STRING() : STRING()
{
  c_debuglevel = 0;
  c_errorcode = 0;
  c_addinfo = "";
}


GDT_BOOLEAN 
KWAQS_STRING::OpConvert(BERBROWSER *ber)
{
  BERBROWSER sub = *ber;
  CHR COperator[256],Query[2048];
  INT operator_value;
  ASN1TAG Tag;
  ber->GetTag(&Tag);
  if(Tag.Number==RPNRPNOP_TAG) { // it's RPN
        
    cout << "Found RPN query OK " << endl;
    BERBROWSER rpn1 = sub;
    BERBROWSER rpn2 = sub;
    BERBROWSER Operator = sub;
    
    sub.GetSubdirectoryX(0, &rpn1);
    sub.GetSubdirectoryX(1, &rpn2);
    sub.GetSubdirectoryX(2, &Operator);
    if(Operator.HasTag(AND_TAG)) {
      if(c_debuglevel >= 5)
	cerr << "Operator=AND" << endl;
      operator_value=0;
    } else if(Operator.HasTag(OR_TAG)) {
      if(c_debuglevel >= 5)
	cerr << "Operator=OR" << endl;
      operator_value=1;
    } else if(Operator.HasTag(ANDNOT_TAG)) {
      if(c_debuglevel >= 5)
	cerr << "Operator=ANDNOT" << endl;
      operator_value=2; 
    } else if(Operator.HasTag(PROX_TAG)) {
      if(c_debuglevel >= 5)
	cerr << "Operator=PROX" << endl;
      // We don't support this yet.
      c_errorcode = 129;
      c_addinfo = "Proximity not supported";
      return GDT_FALSE;
    } else {
      if(c_debuglevel >= 1) 
	cerr << "Unsupported operator in Boolean query" << endl;
      c_errorcode = 110;
      c_addinfo = "Unknown OPERATOR";
      return GDT_FALSE;
    }
    KWAQS_STRING Term1,Term2;
    
    switch(operator_value){
    case 0:			// and
      cout << "Got AND " << endl;
      strcpy(COperator,"AND");
      break;
    case 1:			// OR
      cout << "Got OR "<< endl;
      strcpy(COperator,"OR");
      break;
    case 2:			// and-not
      cout << "Got ANDNOT "<< endl;
      strcpy(COperator,"ANDNOT");
      break;
    }
    Term1.OpConvert(&rpn1);
    Term2.OpConvert(&rpn2);
    CHR *t1,*t2;
    t1=Term1.NewCString();
    t2=Term2.NewCString();
    sprintf(Query,"%s(%s,%s)",COperator,t1,t2);
    delete t1;
    delete t2;
    cout << Query << endl;
    Set((UCHR *)Query,strlen(Query));
    return(GDT_TRUE);
  }else{
    
    if(!sub.HasTag(ATTRIBUTESPLUSTERM_TAG)) {
      // We only support attrplusterm
      c_errorcode = 3;
      c_addinfo = "DB only supports attributes plus term";
      return GDT_FALSE;
    }
    sub.GetSubdirectory(ATTRIBUTESPLUSTERM_TAG, &sub);
    
    if(!sub.HasTag(TERMGENERAL_TAG)) {
      // We only support attrplusterm
      c_errorcode = 229;
      c_addinfo = "Only general term type allowed";
      return GDT_FALSE;
    }
    STRING query_string;
    sub.GetChar(TERMGENERAL_TAG, &query_string);
    cout << "Query: "<<query_string << endl;

    // We now have the term.  We need to add the attribute list.
    // Is there an attribute list present?  We won't require one.
    if(!sub.HasTag(ATTRIBUTELIST_TAG)) {
      // Nope.  We're done.
      UCHR *t;
      t = (UCHR *)query_string.NewCString();
      Set(t, strlen((CHR *)t));
      delete [] t;
      return GDT_TRUE;
    }
    // Add the opening '['
    query_string.Cat("[");
    
    sub.GetSubdirectory(ATTRIBUTELIST_TAG, &sub);
    BERBROWSER seq = sub;
    INT4 attr_count, attr_type, attr_value, i;
    CHR buf[32];
    attr_count = sub.GetSubdirectoryCount();
    
    for(i = 0;i < attr_count;i++) {
      sub.GetSubdirectoryX(i, &seq);
      // We ignore per/term attribute set ids
      // Get the attribute type
      seq.GetInt(ATTRIBUTETYPE_TAG, &attr_type);
      // We only support numeric for attribute value
      if(seq.HasTag(NUMERIC_TAG)) {
	seq.GetInt(NUMERIC_TAG, &attr_value);
	// append these to the query term
	sprintf(buf, "%d,%d", attr_type,attr_value);
	query_string.Cat(buf);
	if(i+1 < attr_count)
	  query_string.Cat(",");
      }
    }
    query_string.Cat("]");
    UCHR *t;
    t = (UCHR *)query_string.NewCString();
    Set(t, strlen((CHR *)t));
    delete [] t;
    
    return GDT_TRUE;
  }
} 


GDT_BOOLEAN 
KWAQS_STRING::Convert(BERBROWSER *ber, 
		      ZSEARCHREQUEST & Request)
{
  if(c_debuglevel >= 9)
    ber->HexDir();
  
  if(!ber->HasTag(QUERY_TAG)) {
    cerr << "PROTOCOL ERROR: QUERY_TAG not present" << endl;
    c_errorcode = 108;
    c_addinfo = "No query present in search request PDU!";
    return GDT_FALSE;
  }
  
  BERBROWSER sub = *ber;
  ber->GetSubdirectory(QUERY_TAG, &sub);
  
  // What type of query?
  INT query_type = -1;
  if(sub.HasTag(TYPE0_TAG))
    query_type = 0;
  if(sub.HasTag(TYPE1_TAG))
    query_type = 1;
  if(sub.HasTag(TYPE2_TAG))
    query_type = 2;
  if(sub.HasTag(TYPE100_TAG))
    query_type = 100;
  if(sub.HasTag(TYPE101_TAG))
    query_type = 101;
  if(sub.HasTag(TYPE102_TAG))
    query_type = 102;
  
  if(c_debuglevel >= 5)
    cerr << "Query type is " << query_type << endl;
  
  if(query_type == 0) {
    STRING query_string;
    
    sub.GetSubdirectory(TYPE0_TAG, &sub);
    ASN1TAGU tag(ASN1_GENERALSTRING);
    sub.GetChar(tag, &query_string);
    
    if(c_debuglevel >= 5) {
      cerr << "Type 0 query string is \"";
      cerr << query_string << "\"" << endl;
    }
    UCHR *t;
    t = (UCHR *)query_string.NewCString();
    Set(t, strlen((CHR *)t));
    delete [] t;
    return GDT_FALSE;
  } else if ((query_type == 1) || (query_type == 101)) {
    sub.GetSubdirectory(TYPE1_TAG, &sub);
    
    // Get the query-global attribute set id
    STRING attr_set;
    sub.GetOID(ATTRIBUTESETID_TAG, &attr_set);
    if(c_debuglevel >= 5)
      cerr << "Attribute Set ID = " << attr_set <<
	endl;
    
    // this should be a different function for recursion
    // I've been Kevin'd
    sub.GetSubdirectoryX(1, &sub);
    return  OpConvert(&sub);
    
  } else {
    // We don't support the requested query type
    c_errorcode = 107;
    return GDT_FALSE;
  }
  
}


GDT_BOOLEAN 
KWAQS_STRING::Error()
{
  if(c_errorcode == 0)
    return GDT_FALSE;
  return GDT_TRUE;
}


void 
KWAQS_STRING::GetErrorInfo(INT *ErrorCode, STRING *AddInfo)
{
  if(c_errorcode == 0)
    return;
  
  *AddInfo = c_addinfo;
  *ErrorCode = c_errorcode;
}


// Log file entry format (all on single line):
//
// host pid groupid/userid [datetime gmt_offset] method extra_data
// protocol/version status bytes_sent
//
void 
ZSERVER::Log(CHR *Method, CHR *Extra, INT Status, INT4 BytesSent)
{
  INT fd;
  FILE *fp;
  CHR *lbuf;
  STRING temp;
  pid_t pid;
  CHR TempString[128];
  
  pid = getpid();
  if((fp = fopen(c_accesslog, "a")) == NULL) {
    perror(c_accesslog);
    return;
  }
  fd = fileno(fp);
  
  // If another process has the file locked already, we'll block
  // here until the file is unlocked.
  lseek(fd, SEEK_SET, 0);
  if (locking(fd, LK_LOCK, LONG_MAX) == -1) {
    perror(c_accesslog);
    fclose(fp);
    return;
  }
  lseek(fd, SEEK_END, 0);
  
  // Hostname
  temp = c_client_hostname;
  
  // pid
  sprintf(TempString, " %d ", pid);
  temp.Cat(TempString);
  
  // Z39.50 GroupId/UserId
  if(c_groupid.Equals(""))
    temp.Cat("-/");
  else {
    temp.Cat(c_groupid);
    temp.Cat("/");
  }
  if(c_userid.Equals(""))
    temp.Cat("-");
  else
    temp.Cat(c_userid);
  temp.Cat(" [");
  
  // Date and time
  struct tm *Time;
  time_t time_clock;
  
  time(&time_clock);
  Time = localtime(&time_clock);
  strftime(TempString, sizeof(TempString), "%d/%b/%Y:%H:%M:%S] ",
	   Time);
  temp.Cat(TempString);
  
  // Method
  temp.Cat(Method);
  temp.Cat(" ");
  
  // Extra stuff
  temp.Cat(Extra);
  
  // Protocol/Version
  // Very peculiar to Z39.50 ProtocolVersion bitstring
  temp.Cat(" Z39.50/");
  sprintf(TempString, "%d ", c_protocol_version.GetLength());
  temp.Cat(TempString);
  
  // Status and BytesSent.
  // Status doesn't really apply to us.  I'll make one up for now.
  // We'll say 1 means good, 0 bad:-)
  sprintf(TempString, "%d %d ", Status, BytesSent);
  temp.Cat(TempString);
  
  temp.Cat("\n");
  
  lbuf = temp.NewCString();
  write(fd, lbuf, strlen(lbuf));
  delete lbuf;
  
  lseek(fd, SEEK_SET, 0);
  if (locking(fd, LK_UNLCK, LONG_MAX) == -1) {
    perror(c_accesslog);
    fclose(fp);
    return;
  }
  
  fclose(fp);
}


void 
ZSERVER::AddToMappingTable(const STRING &FileToAdd,
			   const STRING &DatabaseName)
{
  STRLIST FromPos, Position, ResultList, ToPos, TempList;
  
  // Build entry into mapping table for the database
  Position.Clear();
  Position.AddEntry(DatabaseName);
  
  REGISTRY *FromFile;
  FromPos.Clear();
  FromFile = new REGISTRY("blah");
  FromFile->ProfileLoadFromFile(FileToAdd, FromPos);
  FromPos.AddEntry("Default");
  FromFile->GetData(FromPos, &ResultList);
  
  // Now have a list of all mapping directives ("bib1/4")
  // Step through each one, build a position list, get the value
  // of that directive and SetData on the internal mapping table
  // for the current database.
  STRINGINDEX count, i;
  STRING Entry;
  count = ResultList.GetTotalEntries();
  if(c_debuglevel >= 5) {
    cerr << "Attempting to add " << count << " fields from ";
    cerr << FileToAdd;
    cerr << " to internal mapping tables..." << endl;
  }
  
  ToPos.AddEntry(DatabaseName);
  for(i=1;i <= count;i++) {
    ResultList.GetEntry(i, &Entry);
    FromPos.SetEntry(2, Entry);
    FromFile->GetData(FromPos, &TempList);
    ToPos.SetEntry(2, Entry);
    c_mapping_table->SetData(ToPos, TempList);
  }
  delete FromFile;
}


ZPRESENTRESPONSE *
ZSERVER::PresentResponseWithDiagnostic(
				       const STRING & DiagSetOID, 
				       const INT4 ErrorCode,
				       const STRING & AddInfo)
{
  ZPRESENTRESPONSE *rs;
  
  ZRECORDLIST *RecordList;
  RecordList = new ZRECORDLIST(NONSURROGATEDIAGNOSTIC_TAG);
  ZDEFAULTDIAGFORMAT *diagrec;
  diagrec = new ZDEFAULTDIAGFORMAT(ErrorCode, AddInfo,
				   DiagSetOID);
  RecordList->AddRecord(diagrec);
  rs = new ZPRESENTRESPONSE(c_refid, 1, 1, 5, RecordList);
  delete diagrec;
  delete RecordList;
  return rs;
}


void 
ZSERVER::FreeSapi()
{
  delete TmpSapi;
  TmpSapi=(SAPI *)NULL;
}


SAPI_DB *
ZSERVER::MetaSearch(STRING DBN)
{
  
  //char *name;
  INT LocationCount = 0,Count,i;
  STRING LocationString;
  LocationString = "Location=";
  STRING tempName,a;
  STRLIST Dbs;
  char Host[256];
  //char Port[256];
  char dbn[256];
  char buf[1024],*tempFile;
  INT iPort;
  
  /* 
   * get formatted list in locationstring
   * It'll be a "+" delimited list with entries like
   *
   * host:port/dbname+host:port/dbname 
   *
   * for remote dbs and no punctuation for local ones
   */
  
  Dbs.Split("+",DBN);
  Count=Dbs.GetTotalEntries();

/*  
  for(i=1; i<=Count; i+=3){
    if (i>1)
      LocationString.Cat(",");
    Dbs.GetEntry(i+1,&a);
    a.GetCString(Host,256);
    Dbs.GetEntry(i+2,&a);
    a.GetCString(Port,256);
    Dbs.GetEntry(i,&a);
    a.GetCString(dbn,256);
    sprintf(buf,"%s:%s/%s",Host,Port,dbn);
    printf("buf: %s\n",buf);
    LocationString.Cat(buf);
  }
  STRLIST LocalDbs, RemoteDbs, DBList;
  STRING ThisDBName,RemoteDBs,LocalDBs;
  INT Count, i;
  INT RemoteCount=0,LocalCount=0;

  DBList.Split("+",c_dbname);
  Count = DBList.GetTotalEntries();

  for (i=1;i<=Count;i++) {
    DBlist.GetEntry(i,&ThisDBName);
    if (ThisDBName.Search(':')) {
      if (RemoteCount>0)
	RemoteDBs.Cat("+");
      RemoteDBs.Cat(ThisDBName);
      RemoteCount++;
    } else {
      if (LocalCount>0)
	LocalDBs.Cat(",");
      LocalDBs.Cat(ThisDBName);
      LocalCount++;
    }
  }

*/

  INT DBCount=0;
  for (i=1; i<=Count; i++) {
    Dbs.GetEntry(i,&a);
    if(a.Search(':')) {
      // It's a remote database, add to the list of METABASE dbs
      if (DBCount>0)
	LocationString.Cat(",");
      LocationString.Cat(a);
      DBCount++;
    } else {
      // It's a local database
      strcpy(Host,"localhost");
      iPort = c_port+1;
      a.GetCString(dbn,256);
      sprintf(buf,"%s:%d/%s",Host,iPort,dbn);
      if (DBCount>0)
	LocationString.Cat(",");
      LocationString.Cat(buf);
      DBCount++;
    }
  }

  STRING SapiIniFile;
  SapiIniFile="[Default]\nDBList=METABASE\n\n[METABASE]\nType=ZCLIENT\n";
  SapiIniFile.Cat(LocationString);
  
  
  tempFile = tempnam("/tmp", "newsapi");
  SapiIniFile.WriteFile(tempFile);
  
  INT4 HitCount = 0;
  TmpSapi = new SAPI(tempFile);
  SAPI_DB *db;
  
  cout << "Getting pointer to database..." << endl;
  db = TmpSapi->GetDatabasePtr("Metabase");
  if(db) {
    cout << "Initializing connections to database(s)..." << endl;
    db->SetDebugLevel(9);
    if(!db->Initialize(c_password)) {// send password
      cout << "Failed to initialize" << endl;
      
      return NULL;
    }
  }
  unlink(tempFile);
  return(db);
}
