#ifndef _ZSCLASS_NT_H_
#define _ZSCLASS_NT_H_

#include <limits.h>     // for INT_MAX
#include <sys/types.h>
#include <sys/stat.h>
#include <fstream.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>

#if defined(_MSDOS) || defined(_WIN32)
#include <direct.h>
#include <limits.h>
#include <io.h>
#include <process.h>
#include <sys/locking.h>
#include <windows.h>
#define MAXPATHLEN _MAX_PATH
#define pid_t int
#else
#include <unistd.h>
#include <sys/wait.h>
#include <sys/file.h>
#endif

#ifdef _WIN32
#include <winbase.h>
#endif

#include "isearch.hxx"

#include "dtreg.hxx"
#include "rcache.hxx"
#include "index.hxx"
#include "fprec.hxx"
#include "fpt.hxx"
#include "registry.hxx"
#include "idb.hxx"
#include "operator.hxx"

#include "config.hxx"
#include "ber.hxx"
#include "zsquery.hxx"
#include "zrecords.hxx"
#include "zsession.hxx"

#include "plist.hxx"
#include "mime_record.hxx"
#include "sapi_db.hxx"
#include "db_Isearch.hxx"
#include "sapi.hxx"

#define ZSERVER_NAME "CNIDR-NT zserver"

/// Basic Z39.50 Server Class - Handles all Server functionality
class ZSERVER 
  : public ZSESSION 
{
private:
  ofstream        c_output_file;
  INT             c_max_sessions,
                  c_preferred_msg_size,
                  c_database_count,
                  c_hitcount;
  LONG            c_max_rec_size;
  STRING          c_server_type,
                  c_trace,
                  c_tracelog,
                  c_accesslog,
                  c_sapifile,
                  c_sapipath,
                  c_diagsetid,
                  c_imp_id,
                  c_imp_name,
                  c_imp_version,
                  c_reverse_name_lookup,
                  c_client_hostname,
                  c_client_ipaddress,
                  c_groupid,
                  c_userid,
                  c_password,
                  c_default_db,
                  c_inipath;
  ZINITREQUEST   *c_init_req;
  ZINITRESPONSE  *c_init_res;
  INT             c_forking;
  TCPSOCK        *c_server_tcp;

  SAPI_RSET      *c_result_set;
  SAPI           *c_sapi;
//  PLIST           c_dblist;
  PLIST          *c_dblist;
  REGISTRY       *c_mapping_table;
  INT             c_argc;
  CHR           **c_argv;
  SAPI           *TmpSapi;

  /// Create a new instance
  void Copy(const ZSERVER& OtherServer);
  /// listen for connection on a given port
  INT Listen(UINT Port);
  /// perform a search of user-selected multiple databases
  SAPI_DB * MetaSearch(STRING DBN);
  /// clean up the sapi
  void FreeSapi();
  /// perform a search in the databases
  ZSEARCHRESPONSE *Search(ZSEARCHREQUEST & Request);
  /// retrieve a record from the databases
  ZPRESENTRESPONSE *Present(ZPRESENTREQUEST & Request);
  /// update records in the databases
  ZESUPDATERESPONSE *ESUpdate(ZESTASKPACKAGE *TaskPackage,
			      ZESUPDATEREQUEST & Request);
  void SetDefaults();
  void StoreDefaults(const STRING & Group);
  /// write logging data
  void Log(CHR *Method, CHR *Extra, INT Status, INT4 BytesSent);
  /// ad field maps to the master mapping table
  void ZSERVER::AddToMappingTable(const STRING &FileToAdd,
				  const STRING &DatabaseName);
  /// Initialize the server
  void Init();
  /// Generate Present Response
  ZPRESENTRESPONSE *PresentResponseWithDiagnostic(const STRING & DiagSetOID, 
						  const INT4 ErrorCode,
						  const STRING & AddInfo);
  ///Construct a set of records
  void BuildReturnRecordList(ZRESPONSERECORDS *Records, STRING es_name, 
			     STRING rec_syntax,
			     INT start_point, INT num_requested);
public:
  /// Create a Z39.50 server object
  ZSERVER(int argc, char **argv);
  /// Copy constructor for a Z39.50 server object
  ZSERVER(const ZSERVER& OtherServer);
  /// Generic constructor
  ZSERVER();
  /// Destroy a server and close all connections
  ~ZSERVER();
#if !defined(_WIN16) && !defined(_MSDOS)
  INT StartForkingServer(UINT Port);
#endif
  /// Initialize the INETD server
  void SingleInit() {};
  /// Start forking server listening
  GDT_BOOLEAN ForkingInit();
  /// Clean up after running the INETD server
  void SingleCleanup() {};
  /// start a single Z39.50 session
  void StartSession();
  /// start a server on a given port
  INT StartServer(UINT Port);
  /// Get the current database name for this server
  void DatabaseName(STRING *Value) { *Value = c_dbname; }
  /// Run the server.
  void StartServer();
  /// What kind of server are we running?
  STRING GetServerType() { return c_server_type; }
  /// How many children can we spawn?
  INT GetMaxSessions() { return c_max_sessions; }
  /// wait for client connection
  INT AcceptClient();
};


#if !defined(_MSDOS) && !defined(_WIN32) && !defined (NO_SIGINTERRUPT)
void exit_status(pid_t child_pid, INT status);
#ifndef __SUNPRO_CC
void ChildHandler(int t);
#else
void ChildHandler(int t,...);
#endif
#endif

#if defined(_WIN32)
extern DWORD WorkerThread(ZSERVER *ThisServer);
#endif

/// Provides management functions for KWAQS query strings
class KWAQS_STRING 
  : public STRING 
{
private:
  INT     c_errorcode,
          c_debuglevel;
  STRING  c_addinfo;
public:
  KWAQS_STRING();
  GDT_BOOLEAN Error();
  void GetErrorInfo(INT *ErrorCode, STRING *AddInfo);
  GDT_BOOLEAN Convert(BERBROWSER *b, ZSEARCHREQUEST & Request);
  void SetDebugLevel(INT level) { c_debuglevel = level; }
  GDT_BOOLEAN OpConvert(BERBROWSER *ber);
};

#endif


