/* $Id: zserver.hxx,v 1.10 1998/02/21 00:35:14 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery
and Retrieval, 1994.

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby
granted without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact.

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of
noteworthy uses of this software.

3. The names of MCNC and Clearinghouse for Networked Information
Discovery and Retrieval may not be used in any advertising or publicity
relating to the software without the specific, prior written permission
of MCNC/CNIDR.

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF
THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
************************************************************************/

/*@@@
File:           zserver.hxx
Version:        2.00
$Revision: 1.10 $
Description:    Z39.50 server
Author:         Kevin Gamiel, kevin.gamiel@cnidr.org
@@@*/

#ifndef _ZSERVER_H_
#define _ZSERVER_H_

//#include <values.h>     // for MAXINT
#include <limits.h>     // for INT_MAX
/*
#include <fstream.h>
#include "sapi.hxx"
#include "zsession.hxx"
#include "registry.hxx"
*/
#define ZSERVER_NAME "CNIDR zserver"

/// Basic Z39.50 Server Class - Handles all Server functionality
class ZSERVER 
  : public ZSESSION 
{
private:
  ofstream        c_output_file;
  INT             c_max_sessions,
                  c_preferred_msg_size,
                  c_database_count,
                  c_hitcount;
  LONG            c_max_rec_size;
  STRING          c_server_type,
                  c_trace,
                  c_tracelog,
                  c_accesslog,
                  c_sapifile,
                  c_sapipath,
                  c_diagsetid,
                  c_imp_id,
                  c_imp_name,
                  c_imp_version,
                  c_reverse_name_lookup,
                  c_client_hostname,
                  c_client_ipaddress,
                  c_groupid,
                  c_userid,
                  c_password,
                  c_default_db,
                  c_inipath;
  ZINITREQUEST   *c_init_req;
  ZINITRESPONSE  *c_init_res;
  INT             c_forking;
  TCPSOCK        *c_server_tcp;

  SAPI_RSET      *c_result_set;
  SAPI           *c_sapi;
  PLIST           c_dblist;
  REGISTRY       *c_mapping_table;
  INT             c_argc;
  CHR           **c_argv;
  SAPI           *TmpSapi;

  /// wait for client connection
  INT AcceptClient();
  /// listen for connection on a given port
  INT Listen(UINT Port);
  /// start a single Z39.50 session
  void StartSession();
  /// perform a search of user-selected multiple databases
  SAPI_DB * MetaSearch(STRING DBN);
  /// clean up the sapi
  void FreeSapi();
  /// perform a search in the databases
  ZSEARCHRESPONSE *Search(ZSEARCHREQUEST & Request);
  /// retrieve a record from the databases
  ZPRESENTRESPONSE *Present(ZPRESENTREQUEST & Request);
  /// update records in the databases
  ZESUPDATERESPONSE *ESUpdate(ZESTASKPACKAGE *TaskPackage,
			      ZESUPDATEREQUEST & Request);
  void SetDefaults();
  void StoreDefaults(const STRING & Group);
#if !defined(_MSDOS) && !defined(_WIN32)
  INT StartForkingServer(UINT Port);
#endif
  /// start a server on a given port
  INT StartServer(UINT Port);
  /// write logging data
  void Log(CHR *Method, CHR *Extra, INT Status, INT4 BytesSent);
  /// ad field maps to the master mapping table
  void ZSERVER::AddToMappingTable(const STRING &FileToAdd,
				  const STRING &DatabaseName);
  /// Initialize the server
  void Init();
  /// Generate Present Response
  ZPRESENTRESPONSE *PresentResponseWithDiagnostic(const STRING & DiagSetOID, 
						  const INT4 ErrorCode,
						  const STRING & AddInfo);
  ///Construct a set of records
  void BuildReturnRecordList(ZRESPONSERECORDS *Records, STRING es_name, 
			     STRING rec_syntax,
			     INT start_point, INT num_requested);
public:
  /// Create a Z39.50 server object
  ZSERVER(int argc, char **argv);
  /// Destroy a server and close all connections
  ~ZSERVER();
  /// Get the current database name for this server
  void DatabaseName(STRING *Value) { *Value = c_dbname; }
  /// Run the server.
  void StartServer();
};

#if !defined(_MSDOS) && !defined(_WIN32)
void exit_status(pid_t child_pid, INT status);

#if defined(HAVE_SIGINTERRUPT) || defined(USE_OLDSIGNAL)
#ifndef __SUNPRO_CC
void ChildHandler(int t);
#else
void ChildHandler(int t,...);
#endif
#else
//void ChildHandler(int sig, siginfo_t *sip, void *uap);
void ChildHandler(int sig,...);
#endif
#endif

/// Provides management functions for KWAQS query strings
class KWAQS_STRING 
  : public STRING 
{
private:
  INT     c_errorcode,
          c_debuglevel;
  STRING  c_addinfo;
public:
  KWAQS_STRING();
  GDT_BOOLEAN Error();
  void GetErrorInfo(INT *ErrorCode, STRING *AddInfo);
  GDT_BOOLEAN Convert(BERBROWSER *b, ZSEARCHREQUEST & Request);
  void SetDebugLevel(INT level) { c_debuglevel = level; }
  GDT_BOOLEAN OpConvert(BERBROWSER *ber);
};

#endif


