#include <stdlib.h>
#include <iostream.h>

#include "zserverNT.hxx"

volatile  INT   ChildCount;
          INT   GlobalDebugLevel;


int 
main(int argc, char **argv)
{
  ZSERVER *server;

  STRING Flag;
  INT x = 0;
  while (x < argc) {
    if (argv[x][0] == '-') {
      Flag = argv[x];
      if (Flag.Equals("-V")) {
	cout << "CNIDR zserver for NT, Version " << VERS << endl;
	cout << "Copyright (c) 1995,1996,1997 MCNC/CNIDR" << endl;
	return 0;
      }
    }
    x++;
  }

  if (!setlocale(LC_CTYPE,"")) {
    cout << "Warning: Failed to set the locale!" << endl;
  }

  ios::sync_with_stdio();

  server = new ZSERVER(argc, argv);

  StartServer(server);
//  delete server;

  return 0;
}


void 
StartServer(ZSERVER *RootServer)
{
  STRING ServerType;
  
  //
  // Were we started by inetd or from the command line?
  //
  //    if (StartedByInetd()) {
  ServerType = RootServer->GetServerType();
  if (ServerType.CaseEquals("SERIAL")) {
	  if (RootServer->GetDebugLevel() >= 1)
		  cerr << "NT Serial Connection Server Started" << endl;
	  RunSingleServer(RootServer);
  } else if (ServerType.CaseEquals("INETD")) {
	  if (RootServer->GetDebugLevel() >= 1)
		  cerr << "NT Single/INETD Connection Server Started" << endl;
	  RunInetdServer(RootServer);
  } else {
	  if (RootServer->GetDebugLevel() >= 1)
		  cerr << "NT Threaded Server Started" << endl;
	  StartForkingServer(RootServer);
  }
}


void 
RunSingleServer(ZSERVER *RootServer)
{ 
	for(;;) {
		RootServer->SingleInit();
		RootServer->StartServer();
		RootServer->SingleCleanup();
	}
	return;
}


void 
RunInetdServer(ZSERVER *RootServer)
{ 
	RootServer->SingleInit();
	RootServer->StartServer();
	RootServer->SingleCleanup();
	return;
}


// This is the NT version of the forking server
void
StartForkingServer(ZSERVER* RootServer)
{
	ZSERVER *pTServer;
	ZSERVER  tServer;
	INT      MaxChildren;
	DWORD   *pThreadId;
	DWORD    ThreadParam;
	HANDLE  *pThreadHandle;
	HANDLE  *pDupThreadHandle;
	BOOL	*bThreadStatus;
	INT      i, j, nClosed;

	GlobalDebugLevel=RootServer->GetDebugLevel();
	ChildCount=0;
	MaxChildren = RootServer->GetMaxSessions();

	// Technically, these should get deleted, but since the forking server
	// only passes through here once, the only time they'd need to be deleted
	// is when the server terminates, and that only happens when the entire 
	// process terminates - so we don't need to worry about them.  Besides, 
	// they're not that big.
	pThreadId        = new DWORD[MaxChildren];
	pThreadHandle    = new HANDLE[MaxChildren];
	pDupThreadHandle = new HANDLE[MaxChildren];
	bThreadStatus    = new BOOL[MaxChildren];

	// Start listening on the port
	if (RootServer->ForkingInit()) { // allocates c_server_tcp
		// The port is set up - start listening for a client
		if (GlobalDebugLevel >= 1)
			cerr << "NT Forking Server Started" << endl;
		for (;;) {
			// Accept a connection and start the thread to handle it
			if (!RootServer->AcceptClient()) { // allocates c_tcp
				// For some reason, there was a problem talking to the client
				return;
			}

			// Are there too many children already?
			if (ChildCount >= MaxChildren) {
				// Send init response set to denied
				cerr << "Session limit exceeded" << endl;
				// TO DO - SEND INIT RESPONSE THAT INIT FAILED
			} else {
				// Allocate a new ZSERVER object to copy this one into				
				pTServer = new ZSERVER();	// TO DO - figure out where to delete this
				*pTServer = *RootServer;

				// Start a new thread with the server object
				pThreadHandle[ChildCount] 
					= CreateThread (
						NULL,
						0,
						(LPTHREAD_START_ROUTINE)ServerThread,
						pTServer,
//						ThreadParam,
						CREATE_SUSPENDED,
						&pThreadId[ChildCount]);

				// Check to see if the thread was created successfully
				if (!pThreadHandle[ChildCount]) {
					perror("Create Worker Thread Failed\n");
				} else {
					// Copy the thread handle so we can kill it later
					::DuplicateHandle (
						GetCurrentProcess(),
						pThreadHandle[ChildCount],
						GetCurrentProcess(),
						&pDupThreadHandle[ChildCount],
						0, FALSE, DUPLICATE_SAME_ACCESS);

					// Start the thread executing
					ResumeThread(pThreadHandle[ChildCount]);
					bThreadStatus[ChildCount] = TRUE;
	
					// Add the client to the counter so we can keep track
					ChildCount++;
					if (GlobalDebugLevel >= 5) {
						cerr << "Now up to " << ChildCount 
							<< " child process(es)" << endl;
						cerr.flush();
					}

					DWORD ThreadStatus;
					BOOL bStatus;

					// While we're freed from looking for connections, take
					// the opportunity to clean up the threads that have finished

					// Loop over all the current threads
					for (i=0;i<ChildCount;i++) {
						// Retrieve the exit status for each one
						// Use the duplicate, because this will fail if the thread has
						// already finished and deleted itself
						bStatus = ::GetExitCodeThread(pDupThreadHandle[i],&ThreadStatus);
	
						// OK - got the status and the thread isn't still running,
						// so stuff FALSE into the status array
						if ((!bStatus) || (ThreadStatus != STILL_ACTIVE)) {
							bThreadStatus[i] = FALSE;
						}
					}

					nClosed = 0;
					j = 0;

					// Now, loop again, checking the status array
					for (i=0;i<ChildCount;i++) {
						if (bThreadStatus[i]) {
							// The thread was still running, so keep its info
							if (j != i) {
								pThreadHandle[j]=pThreadHandle[i];
								pDupThreadHandle[j]=pDupThreadHandle[i];
								pThreadId[j]=pThreadId[i];
								bThreadStatus[j]=bThreadStatus[i];
							}
							j++;
						} else {
							// The thread was done, so do some cleanup
							// First, decrement the child counter
							nClosed++;
			
							// Then close the duplicate thread handle
							::CloseHandle(pDupThreadHandle[i]);
						}
					}
					// Adjust ChildCount for the number of completed threads
					ChildCount = ChildCount - nClosed;
					if (GlobalDebugLevel >= 5) {
						cerr << "    Back to " << ChildCount
							<< " child process(es)" << endl;
						cerr.flush();
					}
				}	// thread created
			}	// have fewer that MaxChildren threads
		}	// for(;;)
	}
}

DWORD
ServerThread (ZSERVER *ThisServer)
{
	ThisServer->StartServer();
	return 0;
}

