/* $Id: zutils.cxx,v 1.9 1998/02/26 01:53:02 cnidr Exp $ */
/************************************************************************
Copyright Notice

Copyright (c) MCNC, Clearinghouse for Networked Information Discovery
and Retrieval, 1994.

Permission to use, copy, modify, distribute, and sell this software and
its documentation, in whole or in part, for any purpose is hereby
granted without fee, provided that

1. The above copyright notice and this permission notice appear in all
copies of the software and related documentation. Notices of copyright
and/or attribution which appear at the beginning of any file included in
this distribution must remain intact.

2. Users of this software agree to make their best efforts (a) to return
to MCNC any improvements or extensions that they make, so that these may
be included in future releases; and (b) to inform MCNC/CNIDR of
noteworthy uses of this software.

3. The names of MCNC and Clearinghouse for Networked Information
Discovery and Retrieval may not be used in any advertising or publicity
relating to the software without the specific, prior written permission
of MCNC/CNIDR.

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

IN NO EVENT SHALL MCNC/CNIDR BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF
THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
************************************************************************/

#include <iostream.h>
#include <fstream.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#ifdef UNIX
#include <unistd.h>
#endif
#include <string.h>
//#include "config.hxx"
#include "ztags.hxx"
#include "string.hxx"
#include "vlist.hxx"
#include "strlist.hxx"
#include "tokengen.hxx"
#include "strstack.hxx"
#include "infix2rpn.hxx"
#include "gdt.h"
#include "cgi.hxx"
#include "registry.hxx"
#include "defines.hxx"
#include "gate.hxx"
#include "zutils.hxx"

extern volatile GDT_BOOLEAN WaitingForSignal;

extern char *TempFile;

extern CHR *Action;
extern CHR *Host, *Tmp;
extern UINT Port;
extern CHR *DBName, *DBInfo;
extern STRLIST List;
extern INT PID;
extern enum { FROM_COMMANDLINE, FROM_CGI } Source;
extern STRING TempString;
extern struct stat statbuf;
extern CGI *cgi;
//extern INT NumDatabases;
//extern CHR *database_names[MAX_SERVERS], *database_values[MAX_SERVERS];
extern CHR *QuerySource;
extern GATE *ThisGate;
extern ofstream LogFile;
extern char* IsiteVersion;

void 
MyExit(INT Status)
{
  fclose(stdout);
  fclose(stdin);
  fclose(stderr);
  LogFile << "Exit via MyExit" << endl;
  LogFile.flush();
  LogFile.close();
  
  exit(Status);
}


CHR *
BuildKWAQS(CHR *Term, CHR *Use, CHR* StructureTag, CHR *RelationTag)
{
  CHR *Result;
  INT UseAttr, StructAttr, RelationAttr,TruncAttr;
  INT TermLen=strlen(Term);
  
  if(Term[TermLen-1]=='*'){
    Term[TermLen-1]='\0';
    TruncAttr=1; // right truncation
  }
  else
    TruncAttr=100;		// do not truncate
  
  if(!strcmp(StructureTag,"WORD_LIST"))
    StructAttr=6;
  else if (!strcmp(StructureTag,"DATE"))
    StructAttr=5;
  else if (!strcmp(StructureTag,"DATERANGE"))
    StructAttr=115;
  else
    StructAttr=1;

  if (strlen(Use) == 0)
    UseAttr=1025;
  else
    UseAttr=atoi(Use);

  if (strlen(RelationTag) == 0)
    RelationAttr = 3;
  else if (StrCaseCmp(RelationTag,"EQUAL") == 0)
    RelationAttr = 3;
  else
    RelationAttr=atoi(RelationTag);

  if (RelationAttr == 0)
    RelationAttr = 3;
  
  //  if( (Result=(CHR *)malloc(2000)) ) {
  Result = new CHR[2000];
  if (Result) {
    sprintf(Result,"%s[1,%d,2,%d,4,%d,5,%d]",Term,UseAttr,
	    RelationAttr,StructAttr,TruncAttr);
    return(Result);
  } else {
    cout << "new failed in BuildKWAQS" << endl;
    MyExit(1);
  }
  return(NULL);
}


CHR *
ProcessDateTerms(CHR *Term, CHR *Use, CHR* StructureTag, CHR *Relation)
{
  CHR *Result;
  INT UseAttr, StructAttr, RelationAttr,TruncAttr;
  INT TermLen=strlen(Term);
  
  if(Term[TermLen-1]=='*'){
    Term[TermLen-1]='\0';
    TruncAttr=1; // right truncation
  }
  else
    TruncAttr=100;		// do not truncate
  
  if(!strcmp(StructureTag,"WORD_LIST"))
    StructAttr=6;
  else if (!strcmp(StructureTag,"DATE"))
    StructAttr=5;
  else if (!strcmp(StructureTag,"DATERANGE"))
    StructAttr=115;
  else
    StructAttr=1;
  RelationAttr=atoi(Relation);
  UseAttr=atoi(Use);
  
  //  if( (Result=(CHR *)malloc(2000)) ) {
  Result = new CHR[2000];
  if (Result) {
    sprintf(Result,"%s[1,%d,2,%d,4,%d,5,%d]",Term,UseAttr,
	    RelationAttr,StructAttr,TruncAttr);
    return(Result);
  } else {
    cout << "new failed in ProcessDateTerms" << endl;
    MyExit(1);
  }
  return(NULL);
}


CHR *
BuildFGDCDate(CHR *Year, CHR *Month, INT Day)
{
  CHR *FGDCDate;
  CHR Hold[3];

#ifdef DEBUG
  cout << "Year=" << Year << "<BR>" << endl;
  cout << "Month=" << Month << "<BR>" << endl;
  cout << "Day=" << Day << "<BR>" << endl;
#endif

  sprintf(Hold,"%02d",Day);

  //  if( (FGDCDate=(CHR*)malloc(16)) ) {
  FGDCDate = new CHR[16];
  if (FGDCDate) {
    if(strlen(Year) > 0)
      strcpy(FGDCDate,Year);
    else {
      strcpy(FGDCDate,"");
      return(FGDCDate);
    }
    if(strlen(Month) > 0) {
      strcat(FGDCDate,Month);
      if(Day > 0)
	strcat(FGDCDate,Hold);
    }
  } else {
    cout << "new failed in BuildFGDCDate" << endl;
    MyExit(1);
  }
  return(FGDCDate);
}


void 
SetOption(CGI *cgi, STRING& Buf, CHR *field, CHR *default_string) 
{
  int i;
  CHR field2[100], *value, search[100], replace[100];

  // get existing value of set default
  if ((value = cgi_GetValueByName(cgi, field)) == NULL) {
    //    value = (CHR *)malloc(strlen(default_string)+1);
    value = new CHR [strlen(default_string)+1];
    if (value)
      strcpy(value, default_string);
    else {
      return;
    }
  }

  // construct the search string
  sprintf(search, "$%s$ VALUE=\"%s\"", field, value);

  // construct the replacement string
  sprintf(replace, "SELECTED VALUE=\"%s\"", value);

  // replace search string in Buffer
  Buf.Replace(search, replace);

  // now replace all other occurances of field with nothing
  //  *** if this doesn't replace all occurances, change to a while ***
  sprintf(field2, "$%s$", field);
  Buf.Replace(field2, "");
  delete value;
}

GDT_BOOLEAN 
SetValue(CGI *cgi, STRING& Buf, CHR *field, CHR *default_string)
{
  CHR *Hold;
  CHR search[100], replace[100];
  INT Count;

  if ((Hold = cgi_GetValueByName(cgi, field)) == NULL) {
    //    Hold = (CHR *)malloc(strlen(default_string)+1);
    Hold = new CHR[strlen(default_string)+1];
    if (Hold)
      strcpy(Hold, default_string); 
    else
      return GDT_FALSE;
  }
  sprintf(search, "$%s$", field);
  if(strlen(Hold) > 0) {
    sprintf(replace, "\"%s\"", Hold);
  } else {
    strcpy(replace,"");
  }
  Count = Buf.Replace(search, replace);

  delete Hold;
  
  if (Count >0)
    return GDT_TRUE;
  else
    return GDT_FALSE;
}


void 
SetRadio(CGI *cgi, STRING& Buf, CHR *field, CHR *default_string) 
{
  int i;
  CHR field2[100], *value, search[100], replace[100];

  // get existing value of set default
  if ((value = cgi_GetValueByName(cgi, field)) == NULL) {
    //    value = (CHR *)malloc(strlen(default_string)+1);
    value = new CHR[strlen(default_string)+1];
    if (value)
      strcpy(value, default_string);
    else
      return;
  }

  // construct the search string
  sprintf(search, "$%s$ VALUE=\"%s\"", field, value);

  // construct the replacement string
  sprintf(replace, "CHECKED VALUE=\"%s\"", value);

  // replace search string in Buffer
  Buf.Replace(search, replace);

  // now replace all other occurances of field with nothing
  //  *** if this doesn't replace all occurances, change to a while ***
  sprintf(field2, "$%s$", field);
  Buf.Replace(field2, "");
  delete value;
  return;
}

/*
   Reads the HTML form Filename and adds a CGI variable named SESSION_ID
   with the SessionId value to the form.  It assumes that the SESSION_ID
   should be placed within the first <FORM></FORM> pair it finds.  It
   then prints the form to standard output.
   */


void 
DisplaySearchForm(CHR *cFilename, 
		  CHR *MapString,
		  CHR *min_lat, CHR *max_lat, 
		  CHR *min_lon, CHR *max_lon,
		  INT NumDatabases,
		  CHR *database_names[], CHR *database_values[],
		  INT SessionId,
		  char *Motd)
{
  STRING Buf, FileName, Temp;
  STRING MonthSelect[12];
  STRINGINDEX p;
  INT i;

  for(i=0;i<12;i++)
    MonthSelect[i] = "";

  CHR Hold[1024];
  CHR *Hold1;
  time_t today;
  struct tm    *t;

  //////////////////
  //  FileName = cFilename;
  //  Buf.ReadFile(FileName);
  //////////////////
  LONG lRecStart, lRecEnd, ActualLength;
  LONG filelen;
  PFILE fp;
  PCHR b;
  INT n;
  lRecStart = 0L;

  fp = fopen(cFilename, "r");
  if (!fp) {
    return;
  }

  if ((n=fseek(fp, 0L, SEEK_END)) != 0) {
    return;
  }

  lRecEnd = ftell(fp);

  if ((n=fseek(fp, lRecStart, SEEK_SET)) != 0) {
    return;
  }

  filelen = lRecEnd - lRecStart;
  b = new CHR[filelen + 1];
  ActualLength = fread(b, 1, filelen, fp);
  fclose(fp);
  b[ActualLength] = '\0';
  Buf = b;

  if (Buf.GetLength() == 0) {
    cout << "Your gateway doesn't include a search form ";
    cout << "named <B>" << FileName << "</B>.  Please ";
    cout << "read the Isite documentation for more ";
    cout << "<A HREF=\"http://vinca.cnidr.org/software/Isite";
    cout << "/Isite.html\">information</A>." << endl;
    return;
  }

  STRING LogOut;
  CHR cLogOut[255];

#ifdef NO_LOGOUT
   LogOut = "<p align=\"center\">\n";
   LogOut.Cat("Return to the ");
   sprintf(cLogOut,"<a href=\"%s\">",GATEWAY_LINK);
   LogOut.Cat(cLogOut);
   LogOut.Cat("gateway.</a>\n");
#else
  LogOut = "<CENTER>\n";
  sprintf(cLogOut,"<FORM METHOD=\"POST\" ACTION=\"%s\">\n",
	  ThisGate->GetGatewayURL());
  LogOut.Cat(cLogOut);
  LogOut.Cat("<INPUT NAME=\"SERVICE\" VALUE=\"LOGOUT\" TYPE=\"HIDDEN\">\n");
  sprintf(cLogOut,"<INPUT NAME=\"SESSION_ID\" VALUE=\"%d\"", SessionId);
  LogOut.Cat(cLogOut);
  LogOut.Cat(" TYPE=\"HIDDEN\">\n");
  LogOut.Cat("<INPUT TYPE=\"SUBMIT\" VALUE=\"Logout\">\n");
  LogOut.Cat("</FORM>\n");
  LogOut.Cat("</CENTER>\n");
#endif

  Buf.Replace("$log_out$",LogOut);

  // Replace the variables with the args passed down from above
  Buf.Replace("$map_string$",MapString);
  Buf.Replace("$max_lat$",max_lat);
  Buf.Replace("$min_lat$",min_lat);
  Buf.Replace("$max_lon$",max_lon);
  Buf.Replace("$min_lon$",min_lon);

  today = time((time_t *)NULL);
  t = localtime(&today);

  strftime(Hold,5,"%Y",t);
  Buf.Replace("$sng_year$",Hold);
  Buf.Replace("$enddate_year$",Hold);

  strftime(Hold,4,"%d",t);
  Buf.Replace("$sng_day$",Hold);
  Buf.Replace("$enddate_day$",Hold);

  // Handle months - they're weird
  MonthSelect[t->tm_mon] = " SELECTED ";
  Buf.Replace("$sng_mon_jan$",MonthSelect[0]);
  Buf.Replace("$sng_mon_feb$",MonthSelect[1]);
  Buf.Replace("$sng_mon_mar$",MonthSelect[2]);
  Buf.Replace("$sng_mon_apr$",MonthSelect[3]);
  Buf.Replace("$sng_mon_may$",MonthSelect[4]);
  Buf.Replace("$sng_mon_jun$",MonthSelect[5]);
  Buf.Replace("$sng_mon_jul$",MonthSelect[6]);
  Buf.Replace("$sng_mon_aug$",MonthSelect[7]);
  Buf.Replace("$sng_mon_sep$",MonthSelect[8]);
  Buf.Replace("$sng_mon_oct$",MonthSelect[9]);
  Buf.Replace("$sng_mon_nov$",MonthSelect[10]);
  Buf.Replace("$sng_mon_dec$",MonthSelect[11]);
  Buf.Replace("$enddate_mon_jan$",MonthSelect[0]);
  Buf.Replace("$enddate_mon_feb$",MonthSelect[1]);
  Buf.Replace("$enddate_mon_mar$",MonthSelect[2]);
  Buf.Replace("$enddate_mon_apr$",MonthSelect[3]);
  Buf.Replace("$enddate_mon_may$",MonthSelect[4]);
  Buf.Replace("$enddate_mon_jun$",MonthSelect[5]);
  Buf.Replace("$enddate_mon_jul$",MonthSelect[6]);
  Buf.Replace("$enddate_mon_aug$",MonthSelect[7]);
  Buf.Replace("$enddate_mon_sep$",MonthSelect[8]);
  Buf.Replace("$enddate_mon_oct$",MonthSelect[9]);
  Buf.Replace("$enddate_mon_nov$",MonthSelect[10]);
  Buf.Replace("$enddate_mon_dec$",MonthSelect[11]);

  // if first time, set begining date to Jan 1 1980
  // else set it to whatever it was coming in
  SetValue (cgi, Buf, "begdate-year", "1980");		// 1980
  SetValue (cgi, Buf, "begdate-day", "1");		// 1

  for(i=0;i<12;i++)
    MonthSelect[i] = "";
  if ((Hold1 = cgi_GetValueByName(cgi, "begdate-month")) == NULL) {
    //    Hold1 = (PCHR)malloc(2);
    Hold1 = new CHR[2];
    strcpy(Hold1, "1");
  }
  i = atoi(Hold1);
  delete Hold1;

  MonthSelect[i-1] = " SELECTED ";

  Buf.Replace("$begdate_mon_jan$",MonthSelect[0]);
  Buf.Replace("$begdate_mon_feb$",MonthSelect[1]);
  Buf.Replace("$begdate_mon_mar$",MonthSelect[2]);
  Buf.Replace("$begdate_mon_apr$",MonthSelect[3]);
  Buf.Replace("$begdate_mon_may$",MonthSelect[4]);
  Buf.Replace("$begdate_mon_jun$",MonthSelect[5]);
  Buf.Replace("$begdate_mon_jul$",MonthSelect[6]);
  Buf.Replace("$begdate_mon_aug$",MonthSelect[7]);
  Buf.Replace("$begdate_mon_sep$",MonthSelect[8]);
  Buf.Replace("$begdate_mon_oct$",MonthSelect[9]);
  Buf.Replace("$begdate_mon_nov$",MonthSelect[10]);
  Buf.Replace("$begdate_mon_dec$",MonthSelect[11]);
  
  // relative-date
  SetValue (cgi, Buf, "relative-days", "1");		// 1

  // text search
   
  SetOption(cgi, Buf, "use_1", "1016"); 	// FullText
  SetOption(cgi, Buf, "rel_1", "3"); 		// contains
  SetValue (cgi, Buf, "term_1", "");		// blank
  SetOption(cgi, Buf, "BOOLEAN_OP", ""); 	// blank
  SetOption(cgi, Buf, "use_2", "1016"); 	// FullText
  SetOption(cgi, Buf, "rel_2", "3"); 		// contains
  SetValue (cgi, Buf, "term_2", "");		// blank
  SetOption(cgi, Buf, "use_3", "1016"); 	// FullText
  SetOption(cgi, Buf, "rel_3", "3"); 		// contains
  SetValue (cgi, Buf, "term_3", "");		// blank
  SetOption(cgi, Buf, "num_use_1", "1024"); 	// Scale of Source
  SetOption(cgi, Buf, "num_rel_1", "3"); 	// Equals
  SetValue (cgi, Buf, "num_term_1", "");	// blank
  SetOption(cgi, Buf, "NUM_BOOLEAN_OP", ""); 	// blank
  SetOption(cgi, Buf, "num_use_2", "1024"); 	// Scale of Source
  SetOption(cgi, Buf, "num_rel_2", "3"); 	// Equals
  SetValue (cgi, Buf, "num_term_2", "");	// blank
  SetValue (cgi, Buf, "add-geo", "");		// unchecked
//  SetRadio (cgi, Buf, "add-date", "range"); 	// Range
//  SetRadio (cgi, Buf, "database-type", "all"); 	// all

  // compose a string of HTML multiple select list for databases
  STRING db_str;
  INT ndb;
  CHR sdb[10];

  db_str = "";
  if (NumDatabases > 0) {
    for (i=0; i<NumDatabases; i++) {
      db_str.Cat("        <OPTION VALUE=\"");
      db_str.Cat(database_names[i]);
      // *** add check for selected database here ***
      if (i==0) db_str.Cat("\" SELECTED> "); else
	db_str.Cat("\"> ");
      db_str.Cat(database_values[i]);
      db_str.Cat("\n");
    }
  } else {
      db_str.Cat("        <OPTION VALUE=\"");
      db_str.Cat("xxdefault");
      db_str.Cat("\"> ");
      db_str.Cat("Z39.50 Default");
      db_str.Cat("\n");
      NumDatabases = 1;
  }
  ndb = min(NumDatabases,5);
  sprintf(sdb,"%d",ndb);
  Buf.Replace("$num_db$",sdb);

  Buf.Replace("$databases$",db_str);

  // maxrecords
  SetValue (cgi, Buf, "maxrecords", "20");		// 20

  // SESSION_ID
  sprintf(Hold,"%d",SessionId);
  INT nrep;

  if(!(SetValue (cgi, Buf, "SESSION_ID", Hold))) {		// ID
    sprintf(Hold,"<INPUT NAME=\"SESSION_ID\" VALUE=\"%d\" TYPE=\"HIDDEN\">\n</form>",SessionId);
    nrep = Buf.Replace("</FORM>",Hold);
    if (nrep <= 0)
      nrep = Buf.Replace("</form>",Hold);
  }

  // Stick the PID in before the end of the form
  p = Buf.Search("</FORM>");
  if (p == 0)
    p = Buf.Search("</form>");
  if (p == 0) {
    cout << "Your form ";
    cout << "named <B>" << FileName << "</B> doesn't include ";
    cout << "a proper FORM tag.  Please ";
    cout << "read the Isite documentation for more ";
    cout << "<A HREF=\"http: //vinca.cnidr.org/software/Isite";
    cout << "/Isite.html\">information</A>." << endl;
    MyExit(1);
  }

  // And send the form on its way
  cout << Buf;
  cout.flush();

}

void
ShowDocHeader()
{
  cout << "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">" << endl;
  cout << "<HTML>" << endl;
  cout << "<HEAD>" << endl;
  cout << "   <TITLE>Isite Gateway - Error</TITLE>" << endl;
  //cout << "   <base href=\"http://\">" << endl;
  cout << "</HEAD>" << endl;

  cout << "<BODY BGCOLOR=\"#ffffff\" TEXT=\"#000000\">" << endl;

  cout << "<p align=\"center\">" << endl;
  return;
}


void
ShowDocTrailer()
{
  cout << "</p>\n<hr>" << endl;
  cout << "Gateway Based on <a href=\"http://www.cnidr.org/\"><B>CNIDR</B></a> Isite\n<BR>" << endl;
  cout << "<I>Copyright &copy MCNC, Clearinghouse for Networked Information Discovery and Retrieval, 1994-1997.</I><P>" << endl;
  cout << "</BODY>\n</HTML>" << endl;
}
